#ifndef lint
static char SccsDispmaskId[] = "@(#)Dispmask.h	1.1  9/1/93";
#endif

/* Module:	dispmask.h
 *
 * Purpose:	Define the structures for display mask control parameters and
 *		some related constants.
 *
 * Modified:	{0} Tom Nicinski	initial version		   1 Sep 1993
 *		{n} <who> -- <does what> -- <when>
 */

/******************************************************************************
 *                                                                            *
 * The goal of dispmask.h (and its corresponding code in dispmask.c) is to    *
 * reduce the number of places in SAO that require modification for its use.  *
 * Thus, all display mask related information is maintained in one structure  *
 * (rather than, for example, the mask buffer being maintained in buffer (as  *
 * defined by buffer.h, etc.).                                                *
 *                                                                            *
 * The following header files need to be modified to provide access to the    *
 * display mask software:                                                     *
 *                                                                            *
 *   struct.h	Add <#include "dispmask.h">                                   *
 *   extern.h	Add <struct dispmaskRec  dispmask;>                           *
 *                                                                            *
 * The following header files must be included prior to the inclusion of this *
 * header file:                                                               *
 *                                                                            *
 *   imtool.h	(under ${DERVISH_DIR}/include)                                  *
 *                                                                            *
 ******************************************************************************/

/******************************************************************************/
/*
 * dispmaskRec.
 *
 * This structure contains the complete mask image, the mask lookup table, and
 * the X image data buffer (only).  The X image data buffer mimics the dispbox
 * X image data buffer in size.  Thus, the dispbox X image structure can be
 * used with this X image data buffer.
 *
 * The complete mask image is retained as a byte array, even in a bitmap envi-
 * ronment.  When a mask is displayed in a monochrome environment, the byte
 * values are translated to 0 (transparent) or 1 (set) with the lookup table;
 * in a greyscale environment, the byte values are translated accordingly to
 * a color cell value.
 *
 *   o	coord.img information (width and height) is used to describe the com-
 *	plete mask).
 *
 * Fields should be used as follows:
 *
 *   valid	Indicates that a valid mask has been placed into the buffer.  If
 *		valid is not set, the mask will not be displayed.
 *
 *   visible	Indicates whether the masked image should be visible (that is,
 *		it should be displayed in the display box).
 *
 *   glyph	The type of glyph used to indicate a masked pixel when zoomed.
 *
 *   update.hard	Indicates that the masked image needs to be updated/
 *		regenerated. The image is only updated whenever it is to be
 *		displayed.  Thus, updates occur when "valid" and "visible" are
 *		true and, "update.hard" is true or "update.XXX" have changed.
 *		This is meant to cut down on the number of masked image updates
 *		(for example, when panning the image while the masked image is
 *		not visible).
 *
 *   update.XXX	Saved context of different SAO modes at the last update of the
 *		mask.  If these change when the next update call is made, the
 *		update will occur.
 *
 *   buf	The mask buffer contained in a byte array.  After being read, it
 *		will contain the color values for the mask with zero (0) being
 *		transparent (the mask is translated in-place).  As mentioned
 *		above, coord.img.height/width indicates the mask size.
 *
 *		If buf is null (0), then it indicates no mask is present.  This
 *		should take precedence over visible and update.
 *
 *   buf_sz	The size of the mask buf.  It is in units which are equivalent
 *		to bytes (this assumption is built into the code).
 *
 *   lut	The mask lookup table. It is used to translate original mask
 *		values, [0,255], to color cell values.
 *
 *   color.def	Default values for masking pixels.  This is mainly used for
 *		original mask values beyond color.cnt or for those whose color
 *		is not available:
 *
 *		   .mono	Under monochrome X visuals (screens).  Use only
 *				"white" or "black".
 *
 *		   .grey	The greyscale name (such as "grey13") as a null-
 *				terminated ASCII string.
 *
 *		   .color	The color name (such as "chartreuse") as a null-
 *				terminated string.
 *
 *   color.cnt	The number of color slots used in the color lookup table.
 *
 *   color.pixDefault	The default masking pixel value.  This is taken from
 *		appropriate color.def based on the X visual (screen) type.
 *
 *   color.pixTransparent	The pixel value used to indicate transparency.
 *		This is necessary, for example, under monochrome systems where
 *		masks are to be displayed in white, a value of 0;  while 0 in
 *		the original mask indicated transparency.
 *
 *   color.pix	The color lookup table.  This is used to convert the original
 *		mask (buf) values to pixel (colorcell index) values.  The table
 *		is indexed by the original mask value, with zero (0) being a
 *		placeholder (as a zero (0) mask value indicates transparency).
 *
 *   color.name	The name of the corresponding color in the color lookup table.
 *		The index is the original mask (buf) value prior to being fixed
 *		to use pixel values.
 *
 *   color.names	All the names. These are filled from the DERVISH message
 *		used to download the display mask.
 */

#define	dispmaskColorMax	(NUMCOLOR)	/* Max # colors allowed       */

#if ((BOTTOMCOLOR) != 1)
#error	Code expects "bottom" color index to be 1 (see ${DERVISH_DIR}/include/imtool.h)
#endif

struct dispmaskRec
   {
   unsigned long  int	 valid;			/* Bool: mask valid?          */
   unsigned long  int	 visible;		/* Bool: mask is visible?     */
                  int	 glyph;			/* Zoomed masked pixel glyph  */
   struct					/* Update structure           */
      {
      unsigned long  int	 hard;		/* Bool: force update?        */
                     int	 scaleMode;	/* color.scale.mode           */
                     int	 halfMode;	/* color.halftone.mode        */
                     int	 halfInverse;	/* color.halftone.inverse     */
                     int	 colormapMode;	/* color.colormap_mode        */
      }			 update;
                  unsigned char	*buf;			/* Complete mask              */
   unsigned long  int	 buf_sz;		/* Buf size (units (now bytes)*/
                  char	 lut[MASKLUTTOP];	/* Lookup table               */
   struct					/* Mimics portions of XImage  */
      {						/* ... struct from dispbox.   */
                     char	*data;		/* Image data                 */
      }			 image;
   struct					/* Color lookup table         */
      {
      struct					/* Default values             */
         {
                        char		*mono;	/* Monochrome color name      */
                        char		*grey;	/* Greyscale  color name      */
                        char		*color;	/* Color      color name      */
         }			 def;
      unsigned long  int	 cnt;		/* Number of colors in table  */
                     char	 pixDefault;	/* Default masking pixel value*/
                     char	 pixTransparent;/* Transparent     pixel value*/
                     char	 pix   [(dispmaskColorMax) + 1]; /*CLUT       */
                     char	*name  [(dispmaskColorMax) + 1]; /*Color names*/
                     char	 names [(NUMCOLOR) * (CHARSPERCOLOR)]; /*  "  */
      }			 color;
   };

/******************************************************************************/
