#Create a file currentStripeStatus.par that contains all good segments
#from a set of runs.
#This proc must run from within opdb
#Must run runInventory first, then unset undesired runs.
#
#Commands
#       stripePlan <stripes> -mjd <mjd>
#               Read todo.par and echo which segments
#               still need to be observed.  Estimates ut start, end times
#		for observing and creates imgPlan-mjd.par file.
#		-mjd is optional
#	imgStrategy [mjd]
#		Top level proc to run all imaging strategy procedures.
#	obsPlan
#		Convert imgPlan.par to plan.par

###########################################################################
#This is a top level procedure to run all image status jobs (including
#making plots).

proc imgStrategy {{mjd ""}} {
   if {[inSession]} {opdbDisconnect}
   opdbConnect

#Fetch run.par from mopdb - now done in runStatus
#   imgFetch

#List runs with genuine quality - now done in runStatus
#   imgQuality

#Status
   imgStatus
   stripeStatus
   if {$mjd == ""} {
	stripePlan
   } else {
	stripePlan -mjd $mjd
	}
   obsPlan
   chunkStatus
   segueChunk
   stripePlotGif

#Sanity check.  I shouldn't need the pgEnd, but just in case ...
   pgEnd
   imgPlotSummary
   pgEnd
   imgPlotEq
   pgEnd
   seguePlotEq
   pgEnd

#Copy imgSummary.gif to sdssdp area
   set dir /home/s1/sdssdp/WWW/status/imaging
   if {[file exists $dir]} {
	exec cp imgSummary.gif $dir
	}
   seeingHtml

#History of the world
   imgHistory
   imgHistorySummary imgHistory.txt
   imgHistoryHtml

#This takes a long time to run.
   runHistory
   runHistoryHtml

#Compute metrics
   idMetrics
   if {[inSession]} {opdbDisconnect}
   return
   }

#Helper procs
####################################################################
#Convert ha, dec to az, el
proc azel {ha dec} {
   global at_APO_25_latitude
   set rphi [expr $at_APO_25_latitude/57.2958]
   set rha [expr $ha/57.2958]
   set rdec [expr $dec/57.2958]
   set x1 [expr sin($rha) * cos($rdec)]
   set y1 [expr cos($rha) * cos($rdec)]
   set z1 [expr sin($rdec)]
   set x $x1
   set y [expr $y1*sin($rphi) - $z1*cos($rphi)]
   set z [expr $z1*sin($rphi) + $y1*cos($rphi)]
   set raz [expr atan2(-$x, -$y)]
   set rel [expr asin($z)]
   set az [expr $raz*57.2958]
   if {$az < 0.} {set az [expr $az+360.]}
   set el [expr $rel*57.2958]
   set list ""
   keylset list az $az el $el
   return $list
   }

####################################################################
proc halim {dec} {
   set list [azel 0 $dec]
   set za [expr 90.-[keylget list el]]
   set secz0 [expr 1./cos($za/57.2958)]
   set hamax 0
   loop ha 1 91 {
	set list [azel $ha $dec]
	set za [expr 90.-[keylget list el]]
	if {$za <= 0.} break
	set secz [expr 1./cos($za/57.2958)]
	if {($secz > $secz0 + .2) && ($secz > 1.3)} break
	set hamax $ha
	}
   return $hamax
   return
   }

#######################################################################
#Read the mt standards and compile a list of mean positions
proc mtstdRead {} {
   global env patch
   if {![info exists env(MTSTDS_DIR)]} {
	error "Must set up mtstds product first!"
	return
	}
   if {[info exists patch]} {unset patch}
   set mtchain [param2Chain $env(MTSTDS_DIR)/secondary/mt_patch.par hdr]
   set stripes ""
   loop stripe 1 46 {
	lappend stripes $stripe
	}
   lappend stripes 72 76 79 82 86 90
   foreach stripe $stripes {
	set subchain [chainSearch $mtchain "{stripe = $stripe}"]
	set patch($stripe) ""
	chainEach hndl $subchain {
	   set name [exprGet $hndl.name]
	   set lambda [exprGet $hndl.lambda]

#In the south, we add 360 to negative lambdas to avoid wrap
	   if {$lambda < -90} {set lambda [format %.2f [expr $lambda+360]]}
	   if {![regexp "[format %02d $stripe]\[ABCD\](\[0-9\])" $name \
		all field]} continue
	   if {[lsearch $patch($stripe) $field] < 0} {
		lappend patch($stripe) $field
		set patch($stripe,$field,min) 270.
		set patch($stripe,$field,max) -180.
		}
	   set patch($stripe,$field,min) [min $patch($stripe,$field,min) \
		$lambda]
	   set patch($stripe,$field,max) [max $patch($stripe,$field,max) \
		$lambda]
	   }
	chainDel $subchain
	}
   genericChainDel $mtchain
   return
   }

#######################################################################
#Get ut limits for observing a given segment.  Algorithm: compute
#utmin, utmax for start, end; take smallest common window.

proc utLimits {start end stripe stmid} {

#Start time - (this is UT for observing at Greenwich, and not too far
#off local time anywhere).
   set list [imgStripeLimits $stripe]
   set eta [keylget list eta]

#Compute ut limits for start
   set list [surveyToEq $start $eta]
   set ra [keylget list ra]
   set dec [keylget list dec]

#UT at which this ra transits.  stmid is APO sidtim for 0H UT
   set ut [expr $ra - $stmid]
   set hamax [halim $dec]
   set utmin [expr fmod($ut - $hamax + 720, 360)]
   set utmax [expr fmod($ut + $hamax + 720, 360)]
   set startut1 $utmin
   set startut2 $utmax
   if {$end < $start} {set end [expr $end+360]}
   set len [expr $end - $start]

#Compute ut limits for end.  What I actually compute is time at end minus
#length of scan
   set list [surveyToEq $end $eta]
   set ra [keylget list ra]
   set dec [keylget list dec]
   set ut [expr $ra - $stmid - $len]
   set hamax [halim $dec]
   set utmin [expr fmod($ut - $hamax + 720, 360)]
   set utmax [expr fmod($ut + $hamax + 720, 360)]
   set endut1 $utmin
   set endut2 $utmax

#Merge range.  First put all uts on the modularity of startut1
#The 1st 3 tests may be unnecessary and may cause harm, since all UT's
#should be modulo 360 deg and positive.
   if {$startut2-$startut1 > 180} {set startut2 [expr $startut2-360]}
   if {$endut1-$startut1 > 180} {set endut1 [expr $endut1-360]}

#Change this test from using startut1 to startut2.  It is possible for
#endut2 to be more than 12 hrs ahead of startut1.
   if {$endut2-$startut2 > 180} {set endut2 [expr $endut2-360]}
   if {$startut2-$startut1 < -180} {set startut2 [expr $startut2+360]}
   if {$endut1-$startut1 < -180} {set endut1 [expr $endut1+360]}
   if {$endut2-$startut1 < -180} {set endut2 [expr $endut2+360]}

   set utmin [max $startut1 $endut1]
   set utmax [min $startut2 $endut2]
   if {$utmax < $utmin} return
   set utmin [expr fmod($utmin+360,360)]
   set utmax [expr fmod($utmax+360,360)]
   return [list $utmin $utmax]
   }

#######################################################################
#Get ut limits for observing a given segment of an arbitrary great circle.
#Adapted from utLimits above.

#start and end are GC longitude

proc utGCLimits {start end node incl stmid} {

#Start time - (this is UT for observing at Greenwich, and not too far
#off local time anywhere).

#Compute ut limits for start
   set list [GCToEq $start 0. -node $node -inclination $incl]
   set ra [keylget list ra]
   set dec [keylget list dec]

#UT at which this ra transits.  stmid is APO sidtim for 0H UT
   set ut [expr $ra - $stmid]
   set hamax [halim $dec]
   set utmin [expr fmod($ut - $hamax + 720, 360)]
   set utmax [expr fmod($ut + $hamax + 720, 360)]
   set startut1 $utmin
   set startut2 $utmax
   if {$end < $start} {set end [expr $end+360]}
   set len [expr $end - $start]

#Compute ut limits for end.  What I actually compute is time at end minus
#length of scan
   set list [GCToEq $end 0. -node $node -inclination $incl]
   set ra [keylget list ra]
   set dec [keylget list dec]
   set ut [expr $ra - $stmid - $len]
   set hamax [halim $dec]
   set utmin [expr fmod($ut - $hamax + 720, 360)]
   set utmax [expr fmod($ut + $hamax + 720, 360)]
   set endut1 $utmin
   set endut2 $utmax

#Merge range.  First put all uts on the modularity of startut1
   if {$startut2-$startut1 > 180} {set startut2 [expr $startut2-360]}
   if {$endut1-$startut1 > 180} {set endut1 [expr $endut1-360]}

#Change startut1 to startut2, following same logic as in utLimits.
   if {$endut2-$startut2 > 180} {set endut2 [expr $endut2-360]}
   if {$startut2-$startut1 < -180} {set startut2 [expr $startut2+360]}
   if {$endut1-$startut1 < -180} {set endut1 [expr $endut1+360]}
   if {$endut2-$startut1 < -180} {set endut2 [expr $endut2+360]}

   set utmin [max $startut1 $endut1]
   set utmax [min $startut2 $endut2]
   if {$utmax < $utmin} return
   set utmin [expr fmod($utmin+360,360)]
   set utmax [expr fmod($utmax+360,360)]
   return [list $utmin $utmax]
   }

####################################################################
#Convert hrs to hh:mm
proc hm {hrs} {
   set min [expr round($hrs*60.)]
   if {$min >= 60*24} {set min 0}
   set hh [expr $min/60]
   set mm [expr $min-$hh*60]
   return [format $hh:%02d $mm]
   }
####################################################################
#Compute status of a set of stripes.  Read todo.par,

proc stripePlan {args} {
   global patch
   set opts { \
        {stripeStatus "Compute remaining area to scan"} \
        {{[stripes]} STRING "" stripes "List of stripes (blank = all)"} \
	{-mjd INTEGER -1 mjd "Plan mjd"} \
    }

   if {[shTclParseArg $args $opts stripeStatus] == 0} {
        return 
	}

   if {$mjd < 0} {set mjd [expr [getclock]/86400 + 40587]}

   mtstdRead
   global todo

   if {[info exists todo]} {
	genericChainDel $todo
	unset todo
	}

#Big oops here - todo-<mjd>.par is regenerated each time I run stripeStatus,
#but I run that multiple time in generating imgHistory with different mjds.
#I will rerun here with the input mjd.
   stripeStatus -mjdmax $mjd

   if {![info exists todo]} {
	set file todo-$mjd.par
	if {![file exists $file]} {error "No $file found"}
	echo Input file is $file
	set todo [param2Chain $file hdr]
	}

#Read holes
   if {[file exists holes.par]} {
	set holes [param2Chain holes.par hdr]
	chainJoin $todo $holes
	}

#Create a plan file
   set outfile imgPlan-$mjd.par
   echo Output file is $outfile
   set fid [open $outfile w]
   puts $fid "#Created [fmtclock [getclock]]"
   puts $fid "dateDesign [fmtclock [expr ($mjd-40587)*86400] "%d-%h-%Y"]"
   puts $fid "mjdDesign $mjd"
   puts $fid "typedef struct {"
   puts $fid "   int stripe;"
   puts $fid "   char strip<2>;"
   puts $fid "   float slongStart;		# Degrees;"
   puts $fid "   float slongEnd;		# Degrees;"
   puts $fid "   char mstEarlyStart<6>;	# HH:MM;"
   puts $fid "   char mstLateStart<6>;	# HH:MM"
   puts $fid "   float duration;		# Hours"
   puts $fid "   } IMGPLAN;"

   puts $fid "#    Stripe Strip Start End  StartUT-Range  Length(hr)"

#Approximate sidereal time in deg. at 0hr UT
   set yr [fmtclock [getclock] %Y]
   set mjdsep21 [expr [convertclock "21 sep $yr"]/86400 + 40587]
   set stmid [expr 360.*($mjd-$mjdsep21)/365.25]

#Sidereal time at APO corresponding to 0H UT
#(note that 7 hrs is good approx to true longitude)
   set stmid [expr $stmid - 7.*15.]
   set zone UT
   puts $fid "zone $zone"
   if {$stmid < 0} {set stmid [expr $stmid+360.]}
   echo APO Sidereal Time at 0H UT = \
	[format %.0f $stmid] (deg).

   if {$stripes == ""} {
	loop i 1 46 {lappend stripes $i}
	lappend stripes 72 76 79 82 86 90
	}
  
   foreach stripe $stripes {
	echo Stripe $stripe
	foreach strip "S N" {
	   set list [imgStripeLimits $stripe]
	   set lmin [keylget list lambdaMin]
	   set lmax [keylget list lambdaMax]

#In the south, we add 360 to negative lambdas to avoid wrap
	   if {$lmin < -90} {set lmin [expr $lmin+360.]}
	   if {$lmax < -90} {set lmax [expr $lmax+360.]}
	   set lmin [expr round($lmin)]
	   set lmax [expr round($lmax)]

#Total survey stripe length
	   set stripeLen [expr $lmax - $lmin]

	   chainEach hndl $todo {
		foreach var "stripe strip slongStart slongEnd" {
		   set c$var [exprGet $hndl.$var]
		   }
		if {$stripe != $cstripe || $strip != $cstrip} continue
		set start $cslongStart
		set end $cslongEnd
		if {$start < -90} {set start [expr $start+360.]}
		if {$end < -90} {set end [expr $end+360.]}
		set len [expr $end-$start]

#Break up long segments.
#If stripe is too long, split into 2
#For July, even 4 hrs is too long.  Try 3 hr segments.
		if {$len <= 45.} {
		   set list [list [list $start $end]]
		} elseif {$len <= 90.} {
		   set list [list [list $start [expr $start + $len/2.]]]
		   lappend list [list [expr $start + $len/2.] $end]
		} else {
		   set list [list [list $start [expr $start + $len/3.]]]
		   lappend list [list [expr $start + $len/3.] [expr $start + \
			2.*$len/3.]]
		   lappend list [list [expr $start + 2.*$len/3.] $end]
		   }
		foreach pair $list {
		   set start [format %.0f [lindex $pair 0]]
		   set end [format %.0f [lindex $pair 1]]

#Make sure we have a patch within range
		   set flag 0
		   foreach field $patch($stripe) {
			if {$patch($stripe,$field,min) >= $start && \
			   $patch($stripe,$field,max) <= $end} {
			   incr flag
			   }
			}

#If flag is 0, extend either start or end of run
		   if {$flag < 2} {
			set extend1 90
			set extend2 90
			foreach field $patch($stripe) {
			   set extend [expr $start - \
				$patch($stripe,$field,min)]
			   if {$extend > 0} {
				set extend1 [min $extend1 $extend]
				}
			   set extend [expr $patch($stripe,$field,max) - \
				$end]
			   if {$extend > 0} {
				set extend2 [min $extend2 $extend]
				}
			   }
#Don't extend if no patches found
			if {($extend1 < $extend2 || $flag < 1) && \
			   $extend1 < 90} {
echo Segment $start to $end, extending start by $extend1
			   set start [expr $start-$extend1]
			   incr flag
			   }
			if {($extend1 >= $extend2 || $flag < 2) && \
			   $extend2 < 90} {
echo Segment $start to $end, extending end by $extend2
			   set end [expr $end+$extend2]
			   incr flag
			   }
		   } else {
			}

#Get ut limits for observing this pair.
		   set len [expr $end - $start]
		   set utlimits [utLimits $start $end $stripe $stmid]
		   if {[string length $utlimits] == 0} {
			echo limits not optimal for stripe \
			   $stripe strip $strip
			continue
			}
		   set ut1 [hm [expr [lindex $utlimits 0]/15.]]
		   set ut2 [hm [expr [lindex $utlimits 1]/15.]]
		   set len [format %.1f [expr $len/15.]]
		   if {$len == 0.} continue

#Final formatting:
		   if {$start > 180} {set start [expr $start-360]}
		   if {$end > 180} {set end [expr $end-360]}
		   set start [format %.0f $start]
		   set end [format %.0f $end]
		   puts $fid [format "IMGPLAN %3s %1s %6s %6s %6s %6s \
			%6s" $stripe $strip $start $end $ut1 \
			$ut2 $len]
		   }
		}
	   }
	}
   close $fid
   return
   }

##########################################################################
#Compute visibility of stripes for a given sidereal time, and whether
#we look east or west.

proc visibility {st dir} {
   global halim

#Build up a table of halim vs. dec

   loop dec -20 90 {
	set halim($dec) [halim $dec]
	if {$dir == "east"} {
	   set ra($dec) [expr $st + $halim($dec)]
	} else {
	   set ra($dec) [expr $st - $halim($dec)]
	   }
	set list [eqToSurvey $ra($dec) $dec]
	set lambda [keylget list lambda]
	set eta [keylget list eta]
	if {$ra($dec) > 275. || $ra($dec) < 95.} {
	   set eta [expr $eta+180.]
	   }
	set stripe [expr round(($eta+32.5)/2.5 + 10)]
	if {$stripe < 1 || ($stripe > 45 && $stripe != 72 && \
	   $stripe != 76 && $stripe != 79 && $stripe != 82 \
	   && $stripe != 86 && $stripe != 90)} continue
	set limits [imgStripeLimits $stripe]
	set slmin [keylget limits lambdaMin]
	set slmax [keylget limits lambdaMax]

#Make stripes > 45 look like northern stripes
	if {$stripe > 45} {
	   set slmin [expr $slmin - 180.]
	   set slmax [expr $slmax + 180.]
	   set lambda [expr $lambda - 180.]
	   if {$lambda < -180.} {
		set lambda [expr $lambda+360.]
		}
	   }
	set slim ""
	if {$lambda >= $slmin && $lambda <=$slmax} {
	   set slim [format %.0f $lambda]
	} elseif {$dir == "east"} {
	   if {$lambda > $slmax && $slmax > [expr $st - $halim($dec)]} {
		set slim [format %.0f $slmax]
		}
	} elseif {$dir == "west"} {
	   if {$lambda < $slmin && $slmin < [expr $st + $halim($dec)]} {
		set slim [format %.0f $slmin]
		}
	   }
	if {$stripe > 45 && $slim != ""} {
	   set slim [expr $slim+180.]
	   if {$slim > 180.} {
		set slim [expr $slim-360.]}
		}
	echo stripe $stripe limit $slim
	}

   return
   }

###########################################################################
#Convert imgPlan file to a more readable format.  Editing will still be
#needed
proc obsPlan {} {
   set files [lsort [exec find . -name imgPlan-*.par]]
   set file [lindex $files end]
   set file [file tail $file]
   if {![regexp {imgPlan-([0-9]+).par} $file all mjd]} {
	echo Can't decode $file
	return
	}

#Read stripeExclude, if it exists.  By now I have several stripes with
#little bits not yet finished that are not worth chasing after.
#It is easiest just to wipe them out by hand.
   set exclude ""
   if {[file exists stripeExclude]} {
	param2Chain stripeExclude hdr
	set exclude [keylget hdr exclude]
	}
   set outfile plan-$mjd.par
   echo Creating $outfile
   set outid [open $outfile w]
   set chain [param2Chain $file hdr]
   set zone [keylget hdr zone]
   puts $outid "#mjdDesign $mjd"
   puts $outid ""

#No more south area except the repeats.
#   puts $outid "Strip  Start  End  Start-Time-Range  Length(hr)"
#   puts $outid "       Longitude  (Local $zone Time)"
#   puts $outid ""
#   set subchain [chainSearch $chain "{stripe > 45}"]
#   chainEach hndl $subchain {
#	foreach var "stripe strip slongStart slongEnd mstEarlyStart \
#		mstLateStart duration" {
#	   set $var [exprGet $hndl.$var]
#	   }
#	if {[lsearch $exclude $stripe] >= 0} continue
#	set format "%2s%2s%7d%7d%7s%7s%8.1f"
#	puts $outid [format $format $stripe $strip $slongStart $slongEnd \
#	   $mstEarlyStart $mstLateStart $duration]
#	}
#   chainDel $subchain
#   puts $outid ""
#   southPlan $mjd $outid
   puts $outid "" 
   seguePlan $mjd $outid
#   puts $outid "" 
#   GCPlan $mjd $outid
#   puts $outid "" 
#   trackPlan $mjd $outid
   puts $outid ""
   if {[file exists holes.par]} {
	holePlan $mjd $outid
	puts $outid ""
	}
   puts $outid "-----------"
   puts $outid "North Area"
   puts $outid "                                      Code for North Scans:"
   puts $outid "Rank ordered                          X Pre-existing data"
   puts $outid "by priority                           E Stripe endpoint"
   puts $outid "                                      * Area is to be scanned,\
 but no data yet"
   puts $outid "                                      1st symbol is start of\
 scan; 2nd is"
   puts $outid "                                        end of scan"
   puts $outid ""
   puts $outid "Strip  Start  End  Start-Time-Range  Length(hr)"
   puts $outid "       Longitude        ($zone Time)"
   puts $outid ""
   set subchain [chainSearch $chain "{stripe <= 90}"]
   chainEach hndl $subchain {
	foreach var "stripe strip slongStart slongEnd mstEarlyStart \
		mstLateStart duration" {
	   set $var [exprGet $hndl.$var]
	   }
	if {[lsearch $exclude $stripe] >= 0} continue
	set format "%4s%2s%7d%7d%7s%7s%8.1f"
	puts $outid [format $format $stripe $strip $slongStart $slongEnd \
	   $mstEarlyStart $mstLateStart $duration]
	}
   chainDel $subchain
   puts $outid ""

#Supernova plan will run Oct 1 to Nov 30 of every year.  Let me get the
#min, max MJD for the current year and see if our planning window is
#included.  If so, output a set of tables.
   set unix [expr 86400*($mjd - 40587)]
   set year [fmtclock $unix %Y]

#For Fall, 2006, SN are interleaved with SEGUE in Aug and Dec.
   set mjd1 [expr [convertclock "Aug 15, $year"]/86400+40587]
   set mjd2 [expr [convertclock "Dec 30, $year"]/86400+40587]
   if {$mjd >= $mjd1 && $mjd <= $mjd2} {
	loop i [expr $mjd1+3] [expr $mjd2-2+7] 7 {
	   snPlan $i $outid
	   }
	}
   close $outid
   genericChainDel $chain
   return
   }

#####################################################################
#Compute total area of a range of stripes to see how much we cover in 5 yrs.
#I don't think this proc is used in routine survey planning.

proc imgSum {stripe1 stripe2} {
   set sum 0
   for {set stripe $stripe1} {$stripe <= $stripe2} {incr stripe} {
	set line [imgStripeLimits $stripe]
	set lambdaMin [keylget line lambdaMin]
	set lambdaMax [keylget line lambdaMax]
	set area [expr 2.5*($lambdaMax-$lambdaMin)]
	set sum [expr $sum+$area]
	}
   return [format %.0f $sum]

#Ignore south for now.
#Southern part of main survey
   foreach stripe "76 82 86" {
	set line [stripeLimits $stripe]
	set lambdaMin [keylget line lambdaMin]
	set lambdaMax [keylget line lambdaMax]
	set area [expr 2.5*($lambdaMax-$lambdaMin+360.)]
	set sum [expr $sum+$area]
	}
   return [format %.0f $sum]
   }

########################################################################
#Compute remaining area to scan in the gap.
#This routine is based on stripePlan
proc gapArea {{stripes ""}} {
   set files [lsort [glob -nocomplain todo*.par]]
   set file [lindex $files end]
   if {[string length $file] == 0} {
	echo No todo file found
	return
	}
   echo Input file is $file
   set todo [param2Chain $file hdr]
   set area 0
   set footarea 0.
   if {$stripes == ""} {
	for {set stripe 10} {$stripe <= 37} {incr stripe} {
	   lappend stripes $stripe
	   }
	}
   chainEach hndl $todo {
	foreach var "stripe strip muStart muEnd slongStart slongEnd" {
	   set $var [exprGet $hndl.$var]
	   }
	if {[lsearch $stripes $stripe] < 0} continue
	set len [expr $muEnd - $muStart]
	if {$len < 0} {set len [expr $len + 360.]}
	set area [expr $area + $len*1.25]
	set footarea [expr $footarea + 1.25*57.3*(sin($slongEnd/57.3) - \
	   sin($slongStart/57.3))]
	}
   genericChainDel $todo
   echo To do (stripes $stripes):
   echo "   " Unique Area [format %.0f $area]
   echo "   " Footprint Area [format %.0f $footarea]

#Rich wants the total footprint
   set totarea 0
   set totfoot 0
   foreach stripe $stripes {
	set list [stripeLimits $stripe]
	set slongStart [keylget list lambdaMin]
	set slongEnd [keylget list lambdaMax]
	set totarea [expr $totarea + 2.5*($slongEnd - $slongStart)]
	set totfoot [expr $totfoot + 2.5*57.3*(sin($slongEnd/57.3) - \
	   sin($slongStart/57.3))]
	}
   echo Total (stripes $stripes):
   echo "   " Unique Area [format %.0f $totarea]
   echo "   " Footprint Area [format %.0f $totfoot]
   return
   }
