#Compute bias vector, given a directory of bias frames.

#Top level functions:

#   biasGenJob plan.par	Create a new set of biases.

#   biasQCJob plan.par	Compare biases in new run.

global dataRoot
set dataRoot /data/dp30.a/data

##########################################################################
#Get directories, min, max field numbers
#biasCheck is a global array with returned info
#Subdirectories are Fermi style

proc biasCheckSetup {run} {
   global biasCheck
   global dataDir outDir logDir dplogDir startField endField startCol endCol
   global sigmax1 sigmax2 refDir

   if {![info exists outDir]} {
	set outDir .
	}
   if {![file exists $outDir]} {
	echo Creating directory $outDir
	exec mkdir -p $outDir
	}

   if {![info exists refDir]} {
	set refDir .
	}

#Look for the top level directory
   if {![file exists $dataDir]} {
	set dataDir .
	}

   if {![info exists logDir]} {
	set logDir .
	}
   if {![file exists $logDir]} {
	echo Creating directory $logDir
	exec mkdir -p $logDir
	}

   if {![info exists dplogDir]} {
	set dplogDir .
	}
   if {![file exists $dplogDir]} {
	echo Creating directory $dplogDir
	exec mkdir -p $dplogDir
	}

   if {![info exists startCol]} {
	set startCol 1
	}
   if {![info exists endCol]} {
	set endCol 6
	}

   set biasCheck(run) $run
   set biasCheck(dataDir) $dataDir
   set biasCheck(outDir) $outDir
   set biasCheck(logDir) $logDir
   set biasCheck(dplogDir) $dplogDir
   set biasCheck(refDir) $refDir
   set biasCheck(startCol) $startCol
   set biasCheck(endCol) $endCol

#Get list of files for camcol $startCol and construct start, end fields.
   if {![info exists startField] || ![info exists endField]} {
	set files [lsort [glob -nocomplain $dataDir/1/idR-[format %06d $run]-r1-*.fit]]


	set file1 [lindex $files 0]
	set file2 [lindex $files end]
	set file1 [file tail $file1]
	set file2 [file tail $file2]
	if {![regexp "idR-[format %06d $run]-r1-(\[0-9\]+).fit" \
	   $file1 all field1]} {
	   error "Can't decode $file1 to get field1"
	   }
	if {![regexp "idR-[format %06d $run]-r1-(\[0-9\]+).fit" \
	   $file2 all field2]} {
	   error "Can't decode $file2 to get field1"
	   }
	set field1 [string trimleft $field1 0]
	if {$field1 == ""} {set field1 0}
	set field2 [string trimleft $field2 0]
	if {$field2 == ""} {set field2 0}
   } else {
	set field1 $startField
	set field2 $endField
	}

#First field: take 3rd frame
   set field1 [expr $field1+2]

#Last field: take next to last frame
   set field2 [expr $field2-1]

   if {$field2-$field1 < 2} {
	error "Not enough files found in directory $dataDir"
	}

   echo Directory $dataDir, Field1 $field1 Field2 $field2

#Cache results
   set biasCheck(field1) $field1
   set biasCheck(field2) $field2

   return
   }

###################################################################
#Copmute statistics for left and right halves
#Return a list of lists, each inner list being mean, sigma for 1 side of CCD.

proc biasStats {reg} {
   set nrow [exprGet $reg.nrow]
   set lreg [subRegNew $reg $nrow 1064 0 0]
   set rreg [subRegNew $reg $nrow 1064 0 1064]
   foreach side "l r" {
	set list [regStatsFind [set ${side}reg] -niterations 2]
	set ${side}mean [format %.2f [keylget list mean]]
	set ${side}sigma [format %.2f [keylget list sigma]]
	}
   regDel $lreg
   regDel $rreg
   return [list [list $lmean $lsigma] [list $rmean $rsigma]]
   }

###################################################################
#Compute mean level from bias columns and subtract from all.  If there is
#odd/even present, we subtract the average of the low & high.
#Add back an offset to prevent negative numbers

proc biasNorm {reg offset} {
   set nrow [exprGet $reg.nrow]
   set lreg [subRegNew $reg $nrow 20 0 0]
   set rreg [subRegNew $reg $nrow 20 0 2108]
   foreach side "l r" {
	set list [regStatsFind [set ${side}reg] -niterations 1]
	set ${side}mean [keylget list mean]
	}
   regDel $lreg
   regDel $rreg
   set lreg [subRegNew $reg $nrow 1064 0 0]
   set rreg [subRegNew $reg $nrow 1064 0 1064]
   foreach side "l r" {
	-con [set ${side}reg] [expr [set ${side}mean] - $offset]
	}
   regDel $lreg
   regDel $rreg
   return
   }

###################################################################
#Create bias vector.

proc biasGen {args} {
   set opts { \
	{biasCheck "Compare biases in a new run.  Menu will be presented \
	   to pick reference bias"} \
	{{[run]} INTEGER -1 run "new bias run"} \
	}
   if {[shTclParseArg $args $opts biasCheck] == 0} {
	return 
	}
   global biasCheck
   global env

   biasCheckSetup $run
   set dataDir $biasCheck(dataDir)
   set outDir $biasCheck(outDir)
   set logDir $biasCheck(logDir)
   set dplogDir $biasCheck(dplogDir)
   set field1 $biasCheck(field1)
   set field2 $biasCheck(field2)
   set nfield [expr $field2 - $field1 + 1]
echo nfield = $nfield
   if {$nfield <1} {
	error "Not enough fields"
	}

#Begin the work
#Region for raw data
   set reg [regNew]

#Work region for array of bias vectors.  Hardware ncol
   set workReg [regNew $nfield 2128 -type FL32]

#Final bias vector.  Hardwire ncol
   set biasReg [regNew 1 2128]

#Now read bias vectors from photo gangs
   set row(u) 3
   set row(g) 5
   set row(r) 1
   set row(i) 2
   set row(z) 4
   set fid [open $dplogDir/makebias.qa.par w]
   set camcols "1 2 3 4 5 6"
   set filters "u g r i z"
#   set camcols 5
#   set filters u
   foreach camcol $camcols {
	foreach fil $filters {
#	   regSetWithDbl $workReg 0
	   set r 0
	   loop field $field1 [expr $field2+1] {
		set file \
$dataDir/$camcol/idR-[format %06d $run]-$fil$camcol-[format %04d $field].fit
echo Reading $file
		regReadAsFits $reg $file
		set subReg [subRegNew $workReg 1 2128 $r 0]
		imgColMed $reg $subReg
		regDel $subReg
		incr r
		}
echo Master median
#Photo expects scaled ints (for whatever reason)
	   *con $workReg 32

#Now compute median of medians.
	   imgColMed $workReg $biasReg

#Create header
	   hdrInsWithInt $biasReg.hdr TSHIFT 32 "amount bias vector is \
		scaled by"                
	   hdrInsWithAscii $biasReg.hdr BUNITS "ADUs x TSCALE"
	   hdrInsWithAscii $biasReg.hdr VERSION [format %8s \
		[file tail $env(KTOOLS_DIR)]]
	   hdrInsWithInt $biasReg.hdr TSCALE 32
	   hdrInsWithInt $biasReg.hdr CAMCOL $camcol
	   hdrInsWithInt $biasReg.hdr CAMROW $row($fil)
	   hdrInsWithInt $biasReg.hdr RUN $run
	   hdrInsWithAscii $biasReg.hdr BIAS_ID [fmtclock [getclock] \
		"%Y-%m-%dT%H:%M:%S"]
	   hdrInsWithInt $biasReg.hdr FIELD1 $field1 "first field"
	   hdrInsWithInt $biasReg.hdr FIELD2 $field2 "last field"

#Write out new bias vector
	   set outFile $outDir/idB-[format %06d $run]-$fil$camcol.fit
echo Writing $outFile
	   regWriteAsFits $biasReg $outFile

#Generate some QA numbers
#Let me subtract the master bias from the row medians and compute statistics.
	   echo Median frame stats
	   -col $workReg $biasReg
	   /con $workReg 32
	   set list [biasStats $workReg]
	   set left [lindex $list 0]
	   set right [lindex $list 1]

#The output line is:
#	Left side mean difference between master bias and median array
#	Left side rms between master bias and median array
#	Right side mean difference between master bias and median array
#	Right side rms between master bias and median array

	   puts $fid "$fil$camcol-median $left $right"
	   puts stdout "$fil$camcol-median $left $right"

#Next, subtract bias from the last working image and compute the 1-frame
#read noise.  Must rescale biasReg and add offset to reg.
	   echo Single frame stats
	   /con $biasReg 32

#Let me also subtract off mean bias level from biasReg.
	   biasNorm $biasReg 100
	   -col $reg $biasReg
	   +con $reg 100
	   set list [biasStats $reg]
	   set left [lindex $list 0]
	   set right [lindex $list 1]

#The output line is:
#	Left side mean difference between master bias and single image,
#		preserving the mean bias level of the single image
#	Left side rms between master bias and single image
#	Right side mean difference between master bias and single image,
#		preserving the mean bias level of the single image
#	Right side rms between master bias and single image

	   puts $fid "$fil$camcol-single $left $right"
	   puts stdout "$fil$camcol-single $left $right"
	   }
	}
   close $fid
   regDel $biasReg
   regDel $workReg
   regDel $reg

   return 0
   }


###################################################################
#Compute statistics for left and right halves vs. opECalib

proc biasOpCompare {list row col opChain} {
   set subchain [chainSearch $opChain "{camRow = $row} {camCol = $col}"]
   if {[chainSize $subchain] == 1} {
	   set hndl [chainElementGetByPos $subchain 0]
	   set lbias [exprGet $hndl.biasLevel0]
	   set rbias [exprGet $hndl.biasLevel1]
	   set lrn [format %6.2f [exprGet $hndl.readNoiseDN0]]
	   set rrn [format %6.2f [exprGet $hndl.readNoiseDN1]]
	   handleDel $hndl
   } else {
	return "1 1"
	}
   chainDel $subchain
   set lsig [lindex $list 0]
   set rsig [lindex $list 1]
   set ldiff [format %.2f [expr abs($lsig/$lrn)]]
   set rdiff [format %.2f [expr abs($rsig/$rrn)]]
   return [list $ldiff $rdiff]
   }

###################################################################
#QC step.

proc biasCheck {args} {
   global qa
   global DEBUG
   set opts { \
	{biasCheck "Compare biases in a new run."} \
	{{<run>} INTEGER -1 run "new bias run"} \
	{{<refrun>} STRING "" refrun "reference bias run"} \
	}
   if {[shTclParseArg $args $opts biasCheck] == 0} {
	return 
	}

echo In biasCheck, run is $run, refrun is $refrun
   global biasCheck
   global sigmax1 sigmax2 diffmax3 sigmax3

#sigmax1 is the max sigma relative to opECalib comparing medians of diffent fields.
#sigmax2 is the max sigma relative to opECalib comparing pixels in a single field.
#diffmax3 is the max offset in mean level between bias and reference
#sigmax3 is the max sigma between bias and reference.

#Allow caching of setup stuff for a specific run
   if {![info exists biasCheck($run)]} {
	biasCheckSetup $run
	}
   set dataDir $biasCheck(dataDir)
   set outDir $biasCheck(outDir)
   set logDir $biasCheck(logDir)
   set dplogDir $biasCheck(dplogDir)
   set refDir $biasCheck(refDir)
   if {[info exists qa]} {unset qa}

#Try reading opECalib
   set file [glob -nocomplain $logDir/opECalib*.par]
   if {[llength $file] == 1} {
echo Reading $file
	set opChain [param2Chain $file hdr]
   } else {
	echo "No opECalib file found in directory $logDir!"
	}

#Begin the work
   set row(u) 3
   set row(g) 5
   set row(r) 1
   set row(i) 2
   set row(z) 4

#Initialize qa flags to 0
   foreach camcol "1 2 3 4 5 6" {
	foreach fil "u g r i z" {
	   set qa($fil$camcol) 0
	   }
	}

   set biasReg [regNew]
   set refReg [regNew]

#Read the qa file, compare read noise with opECalib
   foreach camcol "1 2 3 4 5 6" {
	foreach fil "u g r i z" {

#First, check noise of medians
	   if {[info exists opChain]} {
		inipar $dplogDir/makebias.qa.par
		set line [getpars $fil$camcol-median]
		set lsig [lindex $line 1]
		set rsig [lindex $line 3]
		set list "$lsig $rsig"
		set diffs [biasOpCompare $list $row($fil) $camcol \
		   $opChain]
		set ldiff [lindex $diffs 0]
		set rdiff [lindex $diffs 1]
if {[info exists DEBUG]} {
echo $fil$camcol, left, right ratios are $diffs
}
		if {$ldiff > $sigmax1 || $rdiff > $sigmax1} {
		   set qa($fil$camcol) 1
echo Median test failed for $fil$camcol, left, right ratios are $diffs, \
   limit is $sigmamax1
		   }
#Next, check noise of single field
		set line [getpars $fil$camcol-single]
		set lsig [lindex $line 1]
		set rsig [lindex $line 3]
		set list "$lsig $rsig"
		set diffs [biasOpCompare $list $row($fil) $camcol \
		   $opChain]
		set ldiff [lindex $diffs 0]
		set rdiff [lindex $diffs 1]
if {[info exists DEBUG]} {
$fil$camcol: left, right ratios are $diffs
}
		if {$ldiff > $sigmax2 || $rdiff > $sigmax2} {
		   set qa($fil$camcol) 1
echo Single test failed for $fil$camcol: left, right ratios are $diffs, \
   limit is $sigmax2
		   }
		}
#Finally, compare bias vector with reference run vector
	   if {$refrun != ""} {
		set newfile $outDir/idB-[format %06d $run]-$fil$camcol.fit
		regReadAsFits $biasReg $newfile
		set reffile $refDir/idB-[format %06d \
		   $refrun]-$fil$camcol.fit
		regReadAsFits $refReg $reffile

#Compute difference; normalize difference to 100.
#First, revert to unscaled ints.
		set fbiasReg [regNewFromReg $biasReg -type FL32]
		set frefReg [regNewFromReg $refReg -type FL32]
		/con $fbiasReg 32
		/con $frefReg 32
		+con $fbiasReg 100
		-file $fbiasReg $frefReg
		biasNorm $fbiasReg 100
		set list [biasStats $fbiasReg]
		regDel $fbiasReg
		regDel $frefReg
		set llist [lindex $list 0]
		set rlist [lindex $list 1]
		set lmean [format %.2f [expr abs([lindex $llist 0] - 100)]]
		set lsig [lindex $llist 1]
		set rmean [format %.2f [expr abs([lindex $rlist 0] - 100)]]
		set rsig [lindex $rlist 1]
if {[info exists DEBUG]} {
echo $fil$camcol: means vs. reference are $lmean $rmean.
}
	if {$lmean > $diffmax3 || $rmean > $diffmax3} {
		   set qa($fil$camcol) 1
echo Diffmax3 test failed for $fil$camcol, \
   means are $lmean $rmean, limit is $diffmax3
		   }
if {[info exists DEBUG]} {
echo $fil$camcol: sigmas vs. reference are $lsig $rsig.
}
		if {$lsig > $sigmax3 || $rsig > $sigmax3} {
		   set qa($fil$camcol) 1
echo sigmax4 test failed for $fil$camcol, sigmas are $lsig $rsig, \
   limit is $sigmax3
		   }
		}
	   }
	}
   regDel $biasReg
   regDel $refReg
   if {[info exists opChain]} {
	genericChainDel $opChain
	}

#Check qa flags to see if any are 1.
   foreach camcol "1 2 3 4 5 6" {
	foreach fil "u g r i z" {
	   if {$qa($fil$camcol) == 1} {
		return 1
		}
	   }
	}
   return 0
   }

######################################################################
#Create bias vectors

proc biasGenJob {plan} {
   global dataDir outDir logDir dplogDir
   inipar $plan
   set dataDir [getpars dataDir]
   set outDir [getpars outDir]
   set logDir [getpars logDir]
   set dplogDir [getpars dplogDir]
   set run [getpari run]

   if {[catch {set result [biasGen $run]} msg]} {
        echo $msg
	echo biasGen failed!
	return 1
   } else {
	return $result
	}
   }

######################################################################
#QC for bias vectors
proc biasQCJob {plan} {
   global dataDir outDir logDir dplogDir
   inipar $plan
   set dataDir [getpars dataDir]
   set outDir [getpars outDir]
   set logDir [getpars logDir]
   set dplogDir [getpars dplogDir]
   set run [getpari run]

#QC params
   global sigmax1 sigmax2 diffmax3 sigmax3 refDir
   set sigmax1 [getparf mb_sigmax1]
   set sigmax2 [getparf mb_sigmax2]
   set diffmax3 [getparf mb_diffmax3]
   set sigmax3 [getparf mb_sigmax3]

#Reference run is optional
   if {![catch {set refrun [getpari mb_refrun]}]} {
	set refDir [getpars mb_refDir]
   } else {
	set refrun ""
	}

   if {[catch {set result [biasCheck $run $refrun]} msg]} {
        echo $msg
	echo biasCheck failed!
	return 1
   } else {
	return $result
	}
   }

##########################################################################
#Alternate top level procs to make bias and run QC.  These are designed to be a
#drop-in replacement for the existing makebias pipeline.  I think data
#processing does the following setup:
#
#Creates a file "/data/dp3.b/data/xxxx/dplogs/makebias.xxxx.sh,
#which is a shell script that runs the job.  xxxx is the bias run no.
#
#Creates a param file "/data/dp3.b/data/standards/calib/xxxx/biPlan.par"
#which has more job control info.
#
#For test purposes, the root directory changes to a test directory
#defined in the ktools product.

# doBiasFromRun is called by dp scripts to run this automated
# checkBiasFromRun is the qa stage

proc doBiasFromRun {run } {
    global env
   global dataDir outDir logDir dplogDir

#If we have setup the DP products, then use DP_DIR/etc/dpParam.par to
#get top level directories.  If not, fetch top level directories from the
#current directory/dpParam.par.  The latter is done when testing.
    if {[info exists env(DP_DIR)]} {
	set dpdir $env(DP_DIR)
    } else {
	set dpdir .
	}

    param2Chain $dpdir/etc/dpParam.par paramHead
    foreach par "rootDir standardDir" {set $par [keylget paramHead $par]}

#Reference biPlan.par directory to rootDir, so in test mode, I don't
#perturb standardDir.  I will keep standardDir the actual standards directory
#even in test mode.
    set bfn $rootDir/standard/calib/$run/biPlan.par
    if {![file exists $bfn]} {
	echo $bfn does not exist
	return 1
    }
    param2Chain $bfn hdr
    foreach p "startField endField inputDir outputDir startCol endCol" {
	global $p
	set $p [keylget hdr $p]
    }

#Setup - convert above variables to internal variables.
   set dataDir $inputDir
   set outDir $outputDir
   set logDir [file dirname $inputDir]/logs
   set dplogDir $outputDir

#    cd $inputDir

   if {[catch {set result [biasGen $run]} msg]} {
        echo $msg
	echo biasGen failed!
	return 1
   } else {
	echo SUCCESS
	return $result
	}

   }

#######################################################################
proc checkBiasFromRun {run } {
   global env
   global dataDir outDir logDir dplogDir

#If we have setup the DP products, then use DP_DIR/etc/dpParam.par to
#get top level directories.  If not, fetch top level directories from the
#current directory/dpParam.par.  The latter is done when testing.
    if {[info exists env(DP_DIR)]} {
	set dpdir $env(DP_DIR)
    } else {
	set dpdir .
	}

    param2Chain $dpdir/etc/dpParam.par paramHead
    foreach par "rootDir standardDir" {set $par [keylget paramHead $par]}

    set bfn $rootDir/standard/calib/$run/biPlan.par
    if {![file exists $bfn]} {
	echo $bfn does not exist
	return 1
    }
    param2Chain $bfn hdr
    foreach p "startField endField inputDir outputDir startCol endCol" {
	global $p
	set $p [keylget hdr $p]
    }

#Setup - convert above variables to internal variables.
   set dataDir $inputDir
   set outDir $outputDir
   set logDir [file dirname $inputDir]/logs
   set dplogDir $outputDir

#QC params
   global sigmax1 sigmax2 diffmax3 sigmax3 refDir

   set qcfile $dpdir/etc/qc.par
   param2Chain $dpdir/etc/qc.par qcHead
   foreach par "mb_sigmax1 mb_sigmax2 mb_diffmax3 mb_sigmax3" {
       set name [string range $par 3 end]
	set $name [keylget qcHead $par]
	}

#Reference run is optional
   if {![catch {set refrun [keylget qcHead mb_refrun]}]} {
	set refDir $standardDir/calib/$refrun
   } else {
	set refrun ""
	}

#Run the biasCheck proc
   if {[catch {set result [biasCheck $run $refrun]} msg]} {
        echo $msg
	echo biasCheck failed!
	return 1
   } else {
	return $result
	}
6~}

#Helpful routine for quick comparisons of bias vectors by hand

proc biasCompare {run1 run2} {
   global biasCheck
   global qa
   global sigmax1 sigmax2 diffmax3 sigmax3
   global dataDir outDir logDir dplogDir refDir logDir refDir
   global startField endField startCol endCol
   set dataDir .

   global dataRoot
   set outDir $dataRoot/standards/calib/$run1
   set refDir $dataRoot/standards/calib/$run2

#Check for existence of directories - biasCheckSetup is run later and might
#try to create outDir, which is not desired if just running biasCompare
   if {![file exists $outDir]} {
	error "In biasCompare, outDir $outDir does not exist!"
	}
   if {![file exists $refDir]} {
	error "In biasCompare, refDir $refDir does not exist!"
	}
   set logDir .
   set dplogDir .
   set sigmax1 1.
   set sigmax2 5.
   set diffmax3 1.
   set sigmax3 1.
   set startCol 1
   set endCol 6
   set startField 1
   set endField 10
   biasCheck $run1 $run2
   set code 0
   foreach item [array names qa] {
	set code [max $code $qa($item)]
	}
   return $code
   }





