# Scripts to support comparing mt patches with 2.5 m scans.
#
#Setup routines:
#   mdInventory <mtreportdir>
#   mtstdPatchRead (must have set up mtstds product first!)
#   kaInventory <kadir>
#   patchDir plan.par
#
#Main routines to compute residuals:
#
#These do it all - creates html and gifs in current directory
#	htmlCreate		{run patch}     Residuals vs ccd col
#	htmlWidthCreate		{run patch}     Residual vs psf width
#	htmlPSFCreate		{run patch}     PSF width vs ccd col
#	htmlFFCreate		{run}           Flatfield
#
#
#	photoPatch		{run patch}
#	proc patchRCPlot	{run col {filter 2}}
#	proc patchRCWidthPlot	{run col {filter 2}}
#	qaPlot			{run filter}
#	qaWidthPlot		{run filter}
#	qaPSFPlot		{run filter}
#
# Read various MT files and store results in TCL arrays.
# First, code to read an mdReport file and store in TCL variables.
#

	
#########################################################################
proc mtGlobals {{planfile ""}} {
   set globals "report comment sequence patchmjd patchPos photodir \
	kadir patchdir mtreportdir tsdir asdir fcdir astrom starlist \
	objChain radii area mtstdPatch masterdir mtstd camcols fcalib"
   eval uplevel 1 global $globals

#Define MTCOMMENT in its most general form

typedef struct {
   int  seqId;
   char comment[240];
} MTCOMMENT;

#Read a plan file that defines various useful globals (primarily
#directories)
   if {$planfile == ""} return
   eval global $globals
   if {![file exists $planfile]} return
   set fid [open $planfile]
   while {1} {
	set line [string trim [gets $fid]]
	if {[eof $fid]} break
	if {[string index $line 0] == "#"} continue
	if {[string compare $line ""] == 0} continue
#Trim trailing comments
	regsub "\[ \t\]#.*" $line "" line
	set var [lindex $line 0]
	set arg [lrange $line 1 end]
#	global $var
	set $var $arg
	echo $var = $arg
	uplevel 1 global $var
	}
   close $fid

#Define outer radii for profmean.  Units are pixels.
   set radii(-1) 0.
   set radii(0) .564
   set radii(1) 1.69
   set radii(2) 2.58
   set radii(3) 4.41
   set radii(4) 7.51
   set radii(5) 11.58
   set radii(6) 18.58
   set radii(7) 28.55
   set radii(8) 45.50
   set radii(9) 70.51
   set radii(10) 110.5
   set radii(11) 172.5
   set radii(12) 269.5
   set radii(13) 420.5
   set radii(14) 657.5

#Area
   set pi 3.141593
   set area(0) [expr $pi*pow($radii(0),2)]
   loop i 1 15 {
        set area($i) [expr $pi*pow($radii($i),2) - \
                $pi*pow($radii([expr $i-1]),2)]
	}
   return
   }

#########################################################################
#Read list of mt patches from the mtstds product
#Must have setup mtstds first.

proc mtstdPatchRead {} {
   mtGlobals
   global  env

#Location of file
   if {![info exists env(MTSTDS_DIR)]} {error "Must setup MTSTDS product"}
   set file $env(MTSTDS_DIR)/secondary/mt_patch.par
   if {[info exists mtstdsPatch]} {catch {genericChainDel $mtstdsPatch}}
   set mtstdPatch [param2Chain $file hdr]
   chainEach hndl $mtstdPatch {
	set name [exprGet $hndl.name]
	set ra [exprGet $hndl.ra]
	set dec [exprGet $hndl.dec]
	set mtstd($name,ra) $ra
	set mtstd($name,dec) $dec
	}
   return $mtstdPatch
   }
   
#########################################################################
#Make an inventory of all mdReport files.  Save list of sequences.
#Input is directory containing all the reports.  At present this is
#now in the sdssdp home area doc area.
#Creates a global variable "sequence" with all log info.

proc mdInventory {{dir ""}} {
   mtGlobals
   if {$dir == ""} {set dir $mtreportdir}
   set files [glob -nocomplain $dir/mdReport*.par]
   foreach file $files {
	mdReportRead $file
	}
   return
   }

#########################################################################
#Read the mdReport file.  Just specify the file name.
proc mdReportRead {file} {
   mtGlobals

#Avoid extraneous files
   if {![regexp {mdReport-([0-9]+)\.par} $file all mjdate]} return
echo mjdate is $mjdate
   set chains [param2Chain $file hdr]
   set comment($mjdate) ""
   set report($mjdate) ""
   if {![info exists sequence()]} {set sequence() ""}
   if {[lsearch $sequence() $mjdate] < 0} {lappend sequence() $mjdate}
   set sequence($mjdate) ""

#Save the header
   set reportHdr($mjdate) $hdr

   set coChain ""; set expChain ""
   foreach chain $chains {
	if {[chainTypeDefine $chain] == "MTCOMMENT"} {
	   set coChain $chain
	   }
	if {[chainTypeDefine $chain] == "EXP"} {
	   set expChain $chain
	   }
	}

#Read comments, if present
   if {$coChain != ""} {
	loop i 0 [chainSize $coChain] {
	   set hndl [chainElementGetByPos $coChain $i]
	   set seqId [exprGet $hndl.seqId]
#We can have multiple comments for the same seqId!
#Just paste together with formatting to make it both netscape and human readable
	   set line [exprGet $hndl.comment]
	   if {[llength $line] == 1} {
		set line [lindex $line 0]
		}
	   if {[lsearch $comment($mjdate) $seqId] < 0} {
		lappend comment($mjdate) $seqId
		set comment($mjdate,$seqId) $line
	   } else {
		append comment($mjdate,$seqId) "\n<br>$line"
		}
	   handleDel $hndl
	   }
	genericChainDel $coChain
	}

#Read exposure log - better be present!
   if {$expChain  != ""} {
	loop i 0 [chainSize $expChain] {
	   set hndl [chainElementGetByPos $expChain $i]
	   set exposure [exprGet $hndl.exposure]
	   lappend report($mjdate) $exposure
	   set report($mjdate,$exposure) ""
	   foreach var "seqId flavor quality program expTime mjd ra dec \
		filter targetName taihms" {
		lappend report($mjdate,$exposure) $var
		set report($mjdate,$exposure,$var) [exprGet $hndl.$var]
		}

#Compile a list of sequences
	   set seqId [exprGet $hndl.seqId]
#Sometimes in hand-edited logs, an exposure appears in advance of its seqId.
	   if {[lsearch $sequence($mjdate) $seqId] < 0} {
		lappend sequence($mjdate) $seqId
		set sequence($mjdate,$seqId) ""
#Copy targetName over to sequence array and delete report array.
#This should save some space
		set sequence($mjdate,$seqId,targetName)\
		   $report($mjdate,$exposure,targetName)
		set sequence($mjdate,$seqId,ra)\
		   $report($mjdate,$exposure,ra)
		set sequence($mjdate,$seqId,dec)\
		   $report($mjdate,$exposure,dec)
		set sequence($mjdate,$seqId,flavor)\
		   $report($mjdate,$exposure,flavor)
		}
	   lappend sequence($mjdate,$seqId) $exposure
	   handleDel $hndl
	   }
	genericChainDel $expChain
	}
   unset report
   return
   }


##########################################################
#Make inventory of kaCalObj files.  Must have run mdInventory first.
#Creates a global variable patchmjd
#dir is a directory with subdirectories of all secondary patches.

proc kaInventory {{dir ""}} {
   mtGlobals
   if {$dir == ""} {set dir $kadir}

#Tucker now has files in subdirectories ordered by stripe number or old
#patch name.
   set subdirs [glob -nocomplain $dir/*]
   foreach subdir $subdirs {
	if {[file isdirectory $subdir]} {
	   set files [glob -nocomplain $subdir/kaCalObj*.fit]
	   foreach file $files {
	 	echo $file
		kaRead $file
		}
	   continue
	   }

#Maybe it's a kaCalObj file after all.
	if {[regexp kaCalObj $subdir]} {
	   echo $subdir
	   kaRead $subdir
	   }
	}
   return
   }

#############################################################
#Read a kaCalObj file and get the ra, dec, and target name.
proc kaRead {file} {
   global patchmjd
   set hdr [hdrNew]
   hdrReadAsFits $hdr $file
   set mjdate [hdrGetAsInt $hdr MJD]
   set seqId [hdrGetAsInt $hdr SEQID]
   set patchmjd($seqId) $mjdate
   hdrDel $hdr
   return
   }

##############################################################
#Reconstruct patch directory

proc patchDir {{plan plan.par}} {
   mtGlobals $plan
   echo Making inventory of reports

#Don't repeat inventory if global variables exist
   if {![info exists sequence]} {mdInventory $mtreportdir}

   echo Making inventory of kaCalObj files

   if {![info exists patchmjd]} {kaInventory $kadir}

   if {[info exists patchPos]} {unset patchPos}
   set patchPos(patch) ""
#Go through patches, get name, create directory tree
   foreach seq [lsort -integer [array names patchmjd]] {
	set mjdate $patchmjd($seq)
	if {![info exists sequence($mjdate,$seq,targetName)]} continue
	set targetName $sequence($mjdate,$seq,targetName)
	set ra $sequence($mjdate,$seq,ra)
	set dec $sequence($mjdate,$seq,dec)

#We have 3 designations for equatorial patches!

	if {![regexp {([0-9]+)[a-d]([0-9])} $targetName all stripe subfield]} {
	   if {![regexp {Field([A-Z]+)} $targetName all field] \
	   && ![regexp {^([A-Z])[0-9]} $targetName all field]} continue
	} else {
	   set field $stripe-$subfield
	   }
	echo Found field $field

#Directory for this patch
	set dir $patchdir/$field/mjd$mjdate
	if {![file exists $dir]} {
	   echo Creating $dir
	   exec mkdir -p $dir
	   }
	set link $dir/kaCalObj-[format %08d $seq].fit
	set realfile [glob \
		$kadir/*/kaCalObj-[format %08d $seq].fit \
		$kadir/kaCalObj-[format %08d $seq].fit]
	if {![file exists $link]} {
	   echo Linking $link
	   exec ln -s $realfile $link
	   }
	if {[lsearch $patchPos(patch) $field] < 0} {
	   lappend patchPos(patch) $field
	   }
	set patchPos($field,ra) $ra
	set patchPos($field,dec) $dec
	}
   return
   }

###############################################################
proc flagsDefine {} {
   global flagCode flagVal
   set flagCode(1) OBJECT1_CANONICAL_CENTER
   set flagCode(2) OBJECT1_BRIGHT
   set flagCode(4) OBJECT1_EDGE
   set flagCode(8) OBJECT1_BLENDED
   set flagCode(16) OBJECT1_CHILD
   set flagCode(32) OBJECT1_PEAKCENTER
   set flagCode(64) OBJECT1_NODEBLEND
   set flagCode(128) OBJECT1_NOPROFILE
   set flagCode(256) OBJECT1_NOPETRO
   set flagCode(512) OBJECT1_MANYPETRO
   set flagCode(1024) OBJECT1_NOPETRO_BIG
   set flagCode(2048) OBJECT1_DEBLEND_TOO_MANY_PEAKS
   set flagCode(4096) OBJECT1_CR
   set flagCode(8192) OBJECT1_MANYR50
   set flagCode(16384) OBJECT1_MANYR90
   set flagCode(32768) OBJECT1_BAD_RADIAL
   set flagCode(65536) OBJECT1_INCOMPLETE_PROFILE
   set flagCode(131072) OBJECT1_INTERP
   set flagCode(262144) OBJECT1_SATUR
   set flagCode(524288) OBJECT1_NOTCHECKED
   set flagCode(1048576) OBJECT1_SUBTRACTED
   set flagCode(2097152) OBJECT1_NOSTOKES
   set flagCode(4194304) OBJECT1_BADSKY
   set flagCode(8388608) OBJECT1_PETROFAINT
   set flagCode(16777216) OBJECT1_TOO_LARGE
   set flagCode(33554432) OBJECT1_DEBLENDED_AS_PSF
   set flagCode(67108864) OBJECT1_DEBLEND_PRUNED
   set flagCode(134217728) OBJECT1_ELLIPFAINT
   set flagCode(268435456) OBJECT1_BINNED1
   set flagCode(536870912) OBJECT1_BINNED2
   set flagCode(1073741824) OBJECT1_BINNED4
   set flagCode(-2147483648) OBJECT1_MOVED
   foreach attr [array names flagCode] {
	regsub "OBJECT1_" $flagCode($attr) "" flagCode($attr)
	}
   set flagVal(DEBLEND_PRUNED) 67108864
   set flagVal(MANYPETRO) 512
   set flagVal(PETROFAINT) 8388608
   set flagVal(BADSKY) 4194304
   set flagVal(CANONICAL_CENTER) 1
   set flagVal(BRIGHT) 2
   set flagVal(ELLIPFAINT) 134217728
   set flagVal(SATUR) 262144
   set flagVal(MOVED) -2147483648
   set flagVal(EDGE) 4
   set flagVal(BINNED4) 1073741824
   set flagVal(DEBLENDED_AS_PSF) 33554432
   set flagVal(TOO_LARGE) 16777216
   set flagVal(NOTCHECKED) 524288
   set flagVal(NOPETRO) 256
   set flagVal(INTERP) 131072
   set flagVal(MANYR90) 16384
   set flagVal(SUBTRACTED) 1048576
   set flagVal(BLENDED) 8
   set flagVal(BINNED2) 536870912
   set flagVal(INCOMPLETE_PROFILE) 65536
   set flagVal(NOSTOKES) 2097152
   set flagVal(BAD_RADIAL) 32768
   set flagVal(NODEBLEND) 64
   set flagVal(BINNED1) 268435456
   set flagVal(CR) 4096
   set flagVal(NOPROFILE) 128
   set flagVal(MANYR50) 8192
   set flagVal(PEAKCENTER) 32
   set flagVal(DEBLEND_TOO_MANY_PEAKS) 2048
   set flagVal(CHILD) 16
   set flagVal(NOPETRO_BIG) 1024
   return
   }

###############################################################
#Print out flags set for an object
proc flagsDecode {flag} {
   global flagCode
   if {![info exists flagCode]} flagsDefine
   foreach attr [array names flagCode] {
	if {($attr & $flag) != 0} {
	   echo $flagCode($attr)
	   }
	}
   return
   }

##############################################################
#Fetch reprocessed observations of patches
#Updated to handle patches for post Jan 1999 data.
#Directory layout is different from 1998 data.
#Or, maybe no more.
#mjd parameter is just used to determine which directories to search

#Define an enhanced KA type

typedef struct {
   int id;
   double ra;
   double dec;
   float mag<5>;
   float magerr<5>;
   int saturated<5>;
   char patch<5>;
   } KA;

proc singlePatch {patch {mjd *}} {
   global starlist
   mtGlobals

#Clean up old starlists
   if {[info exists starlist($patch)]} {
	foreach mjd $starlist($patch) {
	   if {![info exists starlist($patch,$mjd)]} continue
	   genericChainDel $starlist($patch,$mjd)
	   unset starlist($patch,$mjd)
	   }
	}

#Clean up old chains
    set hndls [handleListFromType CHAIN]
    foreach hndl $hndls {
	set type [chainTypeDefine $hndl]
	if {$type == "KA"} {genericChainDel $hndl}
	if {$type == "MTOBJ"} {genericChainDel $hndl}
	if {$type == "KAMTOBJMERGE"} {genericChainDel $hndl}
    }
    if {[info exists starlist]} {unset starlist}
   lappend starlist(patch) $patch
#Now all kaCalObj files are on one directory for each patch.
#This is a little bit of a fake - runs is now the directory for the
#specified patch.
#Well, now we are back to the old style
   set files [glob -nocomplain $patchdir/$patch/mjd${mjd}/ka*.fit]
   set seqList ""
   set starlist($patch) ""
   foreach file $files {
	set hdr [hdrNew]
	hdrReadAsFits $hdr $file
	set mjd [hdrGetAsInt $hdr MJD]
	set seq [hdrGetAsInt $hdr SEQID]

#Alas, some of the reprocessed kaCalObj files don't have RA or DEC in
#the header.
	catch {set ra [hdrGetAsDbl $hdr RA]}
	catch {set dec [hdrGetAsDbl $hdr DEC]}
	hdrDel $hdr

	lappend seqList $seq
echo night is $mjd, seq is $seq

#Keep record of night per patch - this works with old and new style!
	if {[lsearch $starlist($patch) $mjd] < 0} {
	   lappend starlist($patch) $mjd
	   }

echo Reading $file
	set kaChain [fits2Schema $file KA]

#Create merged output chain
	if {![info exists starlist($patch,$mjd)]} {
	   set starlist($patch,$mjd) [chainNew KA]
	   }

#Short-circuit the kaInventory and patchDir procedures.
#kaRead sets pointer between sequence and mjd
	kaRead $file
	set patchName $sequence($patchmjd($seq),$seq,targetName)
	chainEach hndl $kaChain {
	   handleSet $hndl.patch $patchName
	   }

	set subpatch [string index $patchName 2]
	if {![info exists starlist($patch,subpatch)]} {
	   set starlist($patch,subpatch) $subpatch
	} else {
	   if {[lsearch $starlist($patch,subpatch) $subpatch] < 0} {
		lappend starlist($patch,subpatch) $subpatch
		}
	   }
	chainJoin $starlist($patch,$mjd) $kaChain

#I want to update patchPos array
	if {![info exists patchPos(patch)]} {set patchPos(patch) ""}
	if {[lsearch $patchPos(patch) $patch] < 0} {
	   lappend patchPos(patch) $patch
	   }
	catch {set patchPos($patch,ra) $ra}
	catch {set patchPos($patch,dec) $dec}
	}
   return
   }

####################################################################
#Read astrometric solutions
#For now, asdir points to a specific directory, so we can only process
#one run at a time.

proc astromPatch {args} {
#We read 6 distinct tables.  Easiest to return results in an array of chains
   mtGlobals
   global astrom
   set runs [eval fieldExpand $args]
echo runs are $runs
#Read every 6th fits table
   foreach run $runs {
	set topdir $asdir($run)
	loop col 1 7 {
	   if {[info exists astrom($run,$col)]} {
		genericChainDel $astrom($run,$col)
		unset astrom($run,$col)
		}
	   set hdu [expr 1 + ($col-1)*5]
	   set astrom($run,$col) [fits2Schema \
		[format $topdir/asTrans-%06d.fit $run] ASTRANS \
		-hdu $hdu]
	   }
	}
   return
   }

####################################################################3
# Read photo data for May data.
#Just keep data in L patch: lambda = 59 to 59.5
proc photoPatch {run patch} {

   mtGlobals

#Reconstruct patch directory
#   if {![info exists patchPos]} {patchDir}

   singlePatch $patch
#singlePatch now sets patchPos

#Read astrom data.  Results returned in astrom array
   astromPatch $run

   set objChain($run,patch) $patch
   foreach col "1 2 3 4 5 6" {
	   photoPatchRC $run $col $patch
	}
   return
   }

#######################################################3
proc tobjDef {} {
   if {[lsearch [typesList] TOBJ] >= 0} return
   typedef struct {
	int id;
	int objc_type;
 	float diff;
	float ug;
	float gr;
	float ri;
	float iz;
	float rowc[5];
	float colc[5];
	float psfcounts[5];
	float fibercounts[5];
        float apmag[5];
	float petrorad[5];
	float profmean[5][15];
        float profmag;
	int flags[5];
	double ra;
	double dec;
	double lambda;
	double eta;
	} TOBJ;
   return
   }

#######################################################
#Compute summed mag
proc profSum {chain rad {filter 2}} {
   mtGlobals
echo outer radius is $radii($rad) pixels
   chainEach hndl $chain {
	set sum 0
	loop i 0 [expr $rad+1] {
	   set surf($i) [exprGet $hndl.profmean<$filter><$i>]
	   set sum [expr $sum + pow(10., -$surf($i)/2.5)*$area($i)]
	   }
#	handleSet $hndl.profmag [expr -2.5*log10($sum*.16)]
#	handleSet $hndl.objc_rowc [expr -2.5*log10($sum*.16)]
	handleSet $hndl.proferr<$filter><$rad> [expr -2.5*log10($sum*.16)]
	}
   return
   }
#######################################################
#Compute rms width
proc profRmsSum {chain radmax {filter 2}} {
   mtGlobals
   chainEach hndl $chain {
	set sum 0
	set sum2 0
	loop i 0 $radmax {
	   set surf($i) [expr pow(10., \
		-[exprGet $hndl.profmean<$filter><$i>]/2.5)]
	   }
	loop i 0 $radmax {
	   set surf($i) [expr $surf($i) - $surf([expr $radmax-1])]
	   set cnts($i) [expr $surf($i)*$area($i)]
	   set sum [expr $sum + $cnts($i)]
	   set rad [expr sqrt((pow($radii($i),2) + \
		pow($radii([expr $i-1]),2))/2.)]
	   set sum2 [expr $sum2 + $cnts($i)*$rad*$rad]
	   }
	if {$sum <= 0. || $sum2 <= 0.} continue
	set fwhm [expr 2.35*sqrt($sum2/$sum)]
	handleSet $hndl.profmag $fwhm
	}
   return
   }

#######################################################
#Read a tsobj file using the new Yanny field/file format.  Return the
#handle; don't store handle in global arrays.

proc tsobjRead {run camcol field} {
   mtGlobals
   tobjDef
   if {![info exists tsdir($run)]} {dataDirs $run}
   set files [glob -nocomplain $tsdir($run)/$camcol/tsObj-[format \
	%06d $run]-$camcol-*-[format %04d $field].fit]
   if {[llength $files] == 0} return

#If I find multiple files, take the second.
   set file [lindex $files end]
   set chain [fits2Schema $file TOBJ]
   return $chain
   }

#######################################################
#Read an objc file using the new Yanny field/file format.  Return the
#handle; don't store handle in global arrays.

proc objcRead {run camcol field} {
   mtGlobals
   tobjDef
   if {![info exists photodir($run)]} {dataDirs $run}
   set files [glob -nocomplain $photodir($run)/$camcol/fpObjc-[format \
	%06d $run]-$camcol-[format %04d $field].fit]
   if {[llength $files] == 0} return

#If I find multiple files, take the second.
   set file [lindex $files end]
   set chain [fits2Schema $file TOBJ]
   return $chain
   }

#######################################################
#Read photo data for one run, column, matched to a specific patch
proc photoPatchRC {run col patch} {
   mtGlobals
   global starlist
   global objChain
   if {![info exists astrom($run,$col)]} {astromPatch $run}

#Get N or S
   set hndl [chainGet $astrom($run,4) 0]
   set d [exprGet $hndl.d]
   if {$d > 0} {set strip N} else {set strip S}
   handleDel $hndl

#Clean up old chains
   if {[info exists objChain($run,$col)]} {
	   genericChainDel $objChain($run,$col)
	   unset objChain($run,$col)
	   }

   set objChain($run,strip) $strip

#Get patch limits.  These should be generous enough
   if {![info exists patchPos($patch,ra)]} return

#Grrr, some patches are offset by more than a degree

   set ramin [expr $patchPos($patch,ra) - 3.]
   set ramax [expr $patchPos($patch,ra) + 3.]
   tobjDef
   set objChain($run,$col) [chainNew TOBJ]

#Look at astrom solution
   chainEach hndl $astrom($run,$col) {
#a is coordinate of beginning of frame
	set a [exprGet $hndl.a]
	set b [exprGet $hndl.b]
	set field [exprGet $hndl.field]
	set amax [expr $a + $b*1361.]
	if {$amax < $ramin || $a > $ramax} continue
echo $run $col $field
	set chain($run,$col,$field) [tsobjRead $run $col $field]
	if {$chain($run,$col,$field) == ""} continue
	chainJoin $objChain($run,$col) $chain($run,$col,$field)
	}
   return
   }

#####################################################################
#Match all mjd's for an MT patch with a single photo run, column, and filter

proc patchRCMatch {run col {filter 2}} {
   global objChain
   global starlist
   global patchMatch
   if {[info exists patchMatch]} {
echo Purging patchMatch
	genericChainDel $patchMatch
	unset patchMatch
	}
   set patch [lindex $starlist(patch) 0]

#Match to all mjd's
   if {[llength $starlist($patch)] == 0} return


#Do we have the right subpatch?
   set strip $objChain($run,strip)
   if {$strip == "S"} {
	if {$col == 1 || $col == 2} {
	   set subpatch a
	} elseif {$col == 3} {
	   set subpatch b
	} elseif {$col == 4 || $col == 5} {
	   set subpatch c
	} else {
	   set subpatch d
	   }
   } else {
	if {$col == 1} {
	   set subpatch a
	} elseif {$col == 2 || $col == 3} {
	   set subpatch b
	} elseif {$col == 4} {
	   set subpatch c
	} else {
	   set subpatch d
	   }
	}

   if {[lsearch $starlist($patch,subpatch) $subpatch] < 0} {
	echo No proper subpatch for run $run, camcol $col
	}

   foreach mjd $starlist($patch) {
echo Running chainMatchType
	set matchChain($mjd) [chainMatchType $starlist($patch,$mjd) \
	   $objChain($run,$col) {{ra ra .001} {dec dec .001}}]

echo Running chainTransZXY
	chainTransZXY $matchChain($mjd) res1 object1->mag<$filter> \
	   object2->psfcounts<$filter> 0 1 -1

#Get chain type
	set type [chainTypeDefine $matchChain($mjd)]
	}
   set patchMatch [chainNew $type]

echo Combining chains
   foreach mjd $starlist($patch) {
	chainJoin $patchMatch $matchChain($mjd)
	}
   return
   }

##################################################################
###Make a plot of residuals on row-col map
proc patchRCPlot {run col {filter 2}} {
   global patchMatch

   patchRCMatch $run $col $filter

   set chain [chainSearch $patchMatch "{16 < object2->psfcounts<2> < 17} \
	{object1->magerr<$filter> > 0}"]
   global YMIN YMAX
   set YMIN -.2; set YMAX .2
   cplot $chain object2->colc<2> res1
   chainFSort $chain res1
   set len [chainSize $chain]
   if {$len > 0} {
	set mid [chainGet $chain [expr $len/2]]
	set median [exprGet $mid.res1]
	handleDel $mid
	echo Median for run $run col $col filter $filter \
		is [format %.2f $median], n = $len
	}
   chainDel $chain

#Temporary
   pgLine "0 2048" "0 0"

   return
   }

####################################################################
#Make 6 gifs to look at mt vs 2.5m residuals for 6 columns of a strip.
#Must have run singlePatch and photoPatch first.

proc qaPlot {run {filter 2}} {
   global patchMatch
   mtGlobals
   set i 0
   foreach name "u g r i z" {
	set filname($i) $name
	incr i
	}
   set patch [lindex $starlist(patch) 0]
   singlePatch $patch
#   photoPatch
   pgEnd
   global XMIN XMAX YMIN YMAX
   set YMIN -.2; set YMAX .2
   set XMIN 0; set XMAX 2050
   foreach col "1 2 3 4 5 6" {
echo Run $run Column $col Filter $filter Patch $patch
flush stdout
	pgBegin res-$run-$col-$filter-$patch.gif/GIF
	pgGeomSet 110 110 0 0

#White background and black foreground
	pgScr 0 1 1 1
	pgScr 1 0 0 0
	patchRCPlot $run $col $filter
	pgLabel "" "" "Run $run Col $col Filter $filname($filter) Patch $patch"
	pgEnd
	}
   return
   }

####################################################################
#Make an html page with residual from all across the array.
#This is a top-level routine: Call photoPatch

proc htmlCreate {run patch} {
   mtGlobals

   set i 0
   foreach filter "u g r i z" {
	set filname($i) $filter
	incr i
	}

   if {![info exists objChain($run,patch)] || \
		$objChain($run,patch) != $patch} {
	photoPatch $run $patch
	}
   set file run-$run-$patch.html
   set hid [open $file w]
   puts $hid <html>
   puts $hid "<title>PT vs. 2.5M residuals for Run $run Patch $patch</title>"
   puts $hid "<body bgcolor=white text=black>"
   puts $hid "<center><h1>PT vs. 2.5M residuals for \
	Run $run Patch $patch</h1></center>"
   puts $hid "The following set of GIFs are a comparison of photometry from"
   puts $hid "the Photometric Telescope vs. the 2.5M scans."
   puts $hid "The MT data consist of 6 fields in patch \"$patch\"."
   puts $hid "The 2.5M data are run $run"
   puts $hid <p>
   puts $hid "The following set of figures show PT - 2.5m magnitudes for stars"
   puts $hid "observed in common as a function of the 2.5m column number."
   puts $hid "Each plot is for a single CCD.  The plot are arranged"
   puts $hid "to match the actual CCD mosaic camera layout."
   puts $hid <p>
   puts $hid "<center><table width=660 border=0>"
   puts $hid "<tr align=center>"
   loop i 1 7 {
	puts $hid <th>$i
	}
   puts $hid </table></center>
   foreach filter "2 3 0 4 1" {
	qaPlot $run $filter
	puts $hid "<center>$filname($filter):  " nonewline
	loop col 1 7 {
	   puts $hid "<img src=res-$run-$col-$filter-$patch.gif>" nonewline
	   }
	puts $hid </center>
	}
   puts $hid </body>
   puts $hid </html>
   close $hid
   }

##################################################################
###Make a plot of residuals on row-col map
proc patchRCWidthPlot {run col {filter 2}} {
   global patchMatch

   patchRCMatch $run $col $filter

   set chain [chainSearch $patchMatch {{object1->mag<2> < 16}}]
   global XMIN XMAX
   global YMIN YMAX
   set XMIN 1; set XMAX 2
   set YMIN -.3; set YMAX .3
   cplot $chain object2->petrorad<$filter> res1
   chainDel $chain

#Temporary
   pgLine "0 2048" "0 0"
   return
   }

####################################################################
#Make 6 gifs to look at residual vs. petrosian radius for 6 columns of a strip.
#Must have run singlePatch and photoPatch first.

proc qaWidthPlot {run {filter 2}} {
   global patchMatch
   mtGlobals
   set i 0
   foreach name "u g r i z" {
	set filname($i) $name
	incr i
	}
   set patch [lindex $starlist(patch) 0]
   singlePatch $patch
#   photoPatch
   pgEnd
   global XMIN XMAX YMIN YMAX
   set YMIN -.3; set YMAX .3
   set XMIN 1; set XMAX 2
   foreach col "1 2 3 4 5 6" {
echo Run $run Column $col Filter $filter Patch $patch
flush stdout
	pgBegin width-$run-$col-$filter-$patch.gif/GIF
	pgGeomSet 110 110 0 0

#White background and black foreground
	pgScr 0 1 1 1
	pgScr 1 0 0 0
	patchRCWidthPlot $run $col $filter
	pgLabel "" "" "Run $run Col $col Filter $filname($filter) Patch $patch"
	pgEnd
	}
   return
   }

####################################################################
#Make an html page with residual from all across the array.
#This is a top-level routine: Call photoPatch
#The MT patch is used just to select a subset of the 2.5 m frames.

proc htmlWidthCreate {run patch} {
   mtGlobals

   set i 0
   foreach filter "u g r i z" {
	set filname($i) $filter
	incr i
	}

   if {![info exists objChain($run,patch)] || \
		$objChain($run,patch) != $patch} {
	photoPatch $run $patch
	}
   set file width-$run-$patch.html
   set hid [open $file w]
   puts $hid <html>
   puts $hid "<title>2.5M residuals vs PSF for Run $run Patch $patch</title>"
   puts $hid "<body bgcolor=white text=black>"
   puts $hid "<center><h1>2.5M residuals vs PSF Width for \
	Run $run Patch $patch</h1></center>"
   puts $hid <p>
   puts $hid "The following set of figures show 2.5m magnitudes vs"
   puts $hid "Petrosian radius, which is used as a measure of PSF width"
   puts $hid "Each plot is for a single CCD.  The plot are arranged"
   puts $hid "to match the actual CCD mosaic camera layout."
   puts $hid <p>
   puts $hid "<center><table width=660 border=0>"
   puts $hid "<tr align=center>"
   loop i 1 7 {
	puts $hid <th>$i
	}
   puts $hid </table></center>
   foreach filter "2 3 0 4 1" {
	qaWidthPlot $run $filter
	puts $hid "<center>$filname($filter):  " nonewline
	loop col 1 7 {
	   puts $hid "<img src=width-$run-$col-$filter-$patch.gif>" nonewline
	   }
	puts $hid </center>
	}
   puts $hid </body>
   puts $hid </html>
   close $hid
   }

##################################################################
###Make a plot of PSF width vs. col number
proc patchRCPSFPlot {run col {filter 2}} {
   global patchMatch objChain

   patchRCMatch $run $col $filter

   set chain [chainSearch $patchMatch {{object1->mag<2> < 16}}]
   global XMIN XMAX
   global YMIN YMAX
   set XMIN 0; set XMAX 2048
   set YMIN 4; set YMAX 8

#Fixed max radius of 6 bins
   set subchain [chainSearch $objChain($run,$col) {{psfcounts<2> < 16}}]
   profRmsSum $subchain 6 $filter
   chainDel $subchain
#   cplot $chain object2->colc<2> object2->petrorad<$filter>
   cplot $chain object2->colc<2> object2->profmag
   chainDel $chain

#Temporary
   pgLine "0 2048" "0 0"
   return
   }

####################################################################
#Make 6 gifs to look at petrosian radius vs. ccd col for 6 columns of a strip.
#Must have run singlePatch and photoPatch first.

proc qaPSFPlot {run {filter 2}} {
   global patchMatch
   mtGlobals
   set i 0
   foreach name "u g r i z" {
	set filname($i) $name
	incr i
	}
   set patch [lindex $starlist(patch) 0]
   singlePatch $patch
#   photoPatch   pgEnd
   global XMIN XMAX YMIN YMAX
   set YMIN -.3; set YMAX .3
   set XMIN 1; set XMAX 2
   foreach col "1 2 3 4 5 6" {
echo Run $run Column $col Filter $filter Patch $patch
flush stdout
	pgBegin psf-$run-$col-$filter-$patch.gif/GIF
	pgGeomSet 110 110 0 0

#White background and black foreground
	pgScr 0 1 1 1
	pgScr 1 0 0 0
	patchRCPSFPlot $run $col $filter
	pgLabel "" "" "Run $run Col $col Filter $filname($filter) Patch $patch"
	pgEnd
	}
   return
   }

####################################################################
#Make an html page with residual from all across the array.
#This is a top-level routine: Call photoPatch
#The MT patch is used just to select a subset of the 2.5 m frames.

proc htmlPSFCreate {run patch} {
   mtGlobals

   set i 0
   foreach filter "u g r i z" {
	set filname($i) $filter
	incr i
	}

   if {![info exists objChain($run,patch)] || \
		$objChain($run,patch) != $patch} {
	photoPatch $run $patch
	}
   set file psf-$run-$patch.html
   set hid [open $file w]
   puts $hid <html>
   puts $hid "<title>2.5M PSF vs. Column Number for Run $run Patch $patch</title>"
   puts $hid "<body bgcolor=white text=black>"
   puts $hid "<center><h1>2.5M PSF width vs. CCD column for \
	Run $run Patch $patch</h1></center>"
   puts $hid <p>
   puts $hid "The following set of figures show 2.5m PSF"
   puts $hid "(Petrosian radius) vs. ccd column number."
   puts $hid "Each plot is for a single CCD.  The plot are arranged"
   puts $hid "to match the actual CCD mosaic camera layout."
   puts $hid <p>
   puts $hid "<center><table width=660 border=0>"
   puts $hid "<tr align=center>"
   loop i 1 7 {
	puts $hid <th>$i
	}
   puts $hid </table></center>
   foreach filter "2 3 0 4 1" {
	qaPSFPlot $run $filter
	puts $hid "<center>$filname($filter):  " nonewline
	loop col 1 7 {
	   puts $hid "<img src=psf-$run-$col-$filter-$patch.gif>" nonewline
	   }
	puts $hid </center>
	}
   puts $hid </body>
   puts $hid </html>
   close $hid
   }

##################################################################
###Make a plot of residuals on row-col map
proc patchRCFlatPlot {run col {filter 2}} {
   mtGlobals
   set i 0
   foreach fil "u g r i z" {
       set filnam($i) $fil
       incr i
   }

   set file $tsdir($run)/photo-output/fpC-[format %06d $run]-c$col/psFF-[format %06d $run]-$filnam($filter)$col.fit
   set reg [regNew]
   regReadAsFits $reg $file
   pgEnv 0 2048 1600 2400 0 0
   regRowPlot $reg 0
   regDel $reg

   return
   }

####################################################################
#Make 6 gifs to look at flatfields

proc qaFlatPlot {run {filter 2}} {
   mtGlobals
   set i 0
   foreach name "u g r i z" {
	set filname($i) $name
	incr i
	}

   foreach col "1 2 3 4 5 6" {
echo Run $run Column $col Filter $filter
flush stdout
	pgBegin flat-$run-$col-$filter.gif/GIF
	pgGeomSet 110 110 0 0

#White background and black foreground
	pgScr 0 1 1 1
	pgScr 1 0 0 0
	patchRCFlatPlot $run $col $filter
	pgLabel "" "" "Run $run Col $col Filter $filname($filter)"
	pgEnd
	}
   return
   }

####################################################################
#Make an html page with residual from all across the array.

proc htmlFlatCreate {run} {
   mtGlobals

   set i 0
   foreach filter "u g r i z" {
	set filname($i) $filter
	incr i
	}

   set file flat-$run.html
   set hid [open $file w]
   puts $hid <html>
   puts $hid "<title>2.5M Flatfields for Run $run</title>"
   puts $hid "<body bgcolor=white text=black>"
   puts $hid "<center><h1>2.5M Flatfields</h1></center>"
   puts $hid <p>
   puts $hid "The following set of figures show 2.5m flatfields."
   puts $hid "Each plot is for a single CCD.  The plot are arranged"
   puts $hid "to match the actual CCD mosaic camera layout."
   puts $hid <p>
   puts $hid "<center><table width=660 border=0>"
   puts $hid "<tr align=center>"
   loop i 1 7 {
	puts $hid <th>$i
	}
   puts $hid </table></center>
   foreach filter "2 3 0 4 1" {
	qaFlatPlot $run $filter
	puts $hid "<center>$filname($filter):  " nonewline
	loop col 1 7 {
	   puts $hid "<img src=flat-$run-$col-$filter.gif>" nonewline
	   }
	puts $hid </center>
	}
   puts $hid </body>
   puts $hid </html>
   close $hid
   }

####################################################################
#A hack to compute aperture photometry.  Chain is output of "find" command

proc hack {reg chain rad} {
   imgGlobals
   chainInit $chain iter
   sky4 $reg
   if {![info exists img(zero)]} {
	error "Need to set photometric zero point img(zero)"
	}
   set i 0
   while {1} {
	set hndl [chainNext iter]
	if {$hndl == ""} break
	set row [int [exprGet $hndl.row]]
	set col [int [exprGet $hndl.col]]
#Get better axes
	if {[catch {axes $reg $row $col 5 5}]} {
	   continue
	   }
	handleSet $hndl.row [exprGet $img(axes).row]
	handleSet $hndl.col [exprGet $img(axes).col]
	circle $reg $rad
#Stuff counts
	handleSet $hndl.cnts $img(photonet)
#Hack mag
	if {$img(photonet) > 0.} {
	   handleSet $hndl.fwhm [expr -2.5*log10($img(photonet))+$img(zero)]
	   }
	incr i
	}
   return
   }

####################################################################
####################################################################

#Code to check mt data for internal consistency.  These routines
#are taken from the old "dark.tcl"

#Check bias levels in mt frames?

proc dark {rundir} {
   set dirs [glob $rundir/*]
   foreach dir $dirs {
	set file $dir/biasmt_stare.fit
	if {[file exists $file]} {
	   set reg [openimage $file]
	   sky4 $reg
	   }
	reset
	}
   return
   }

############################################################
#Compare multiple observations of patches
proc multiPatch {{pat *}} {
   global starlist

   mtGlobals

#This is probably not even needed.
   if {![info exists patchmjd]} {
	echo Making inventory of ka files
	kaInventory $kadir
	}

#Clean up old chains
    set hndls [handleListFromType CHAIN]
    foreach hndl $hndls {
	if {[chainTypeDefine $hndl] == "KA"} {genericChainDel $hndl}
    }
    if {[info exists starlist]} {unset starlist}
   set topdir $patchdir
   set patches [glob $topdir/$pat]
   foreach patch $patches {
       set runs [glob $patch/mjd*]
#If only one run, continue
       if {[llength $runs] < 2} continue
       echo Multiple nights for patch $patch
       set name [file tail $patch]
       lappend starlist(patch) $name
       foreach run $runs {
	   echo "   Processing $run"
	   set files [glob $run/ka*.fit]
	   foreach file $files {
	       set chain($file) [fits2Schema $file KA]
	   }
	   set night [file tail $run]
	   regexp {mjd(.+)} $night all night
	   set starlist($name,$night) [chainNew KA]
	   foreach file [array names chain] {
	       chainJoin $starlist($name,$night) $chain($file)
	   }
	   unset chain
       }
   }
   return
}

#A telling set of data is to compare i band mags for patch K,
#nights 50904 vs. 50949

#####################################################################
#Plot the locations of all nights for a given patch

proc patchPlot {patch} {
   mtGlobals
   patchDir

#Approximate survey coord limits
   set list [eqToSurvey $patchPos($patch,ra) 0]
   set lambda [keylget list lambda]
   set eta [keylget list eta]
   set lmin [expr $lambda - .5]
   set lmax [expr $lambda + 0.5]
   set emin [expr $eta - 1.5]
   set emax [expr $eta + 1.5]

#Now just use ra and dec
   set ra $patchPos($patch,ra)
   set ramin [expr $ra-3.]
   set ramax [expr $ra+3.]
   set decmin -1.5
   set decmax 1.5
   if {[pgQinf STATE] == "CLOSED"} {
	pgBegin; pgGeomSet 400 400 10 25]
	pgAsk 0
	}
   global pgSym

   set pgSym 1
#   pgEnv $lmin $lmax $emin $emax 1 0
   pgSci 1
   pgEnv $ramax $ramin $decmin $decmax 1 0

#First, plot location of 2.5 m scans
   eval astromPatch [array names asdir]
   set pgSym 0
   pgSls 4
   foreach run [array names asdir] {
	incr pgSym
	pgSci $pgSym
	foreach col "1 2 3 4 5 6" {
	   chainInit $astrom($run,$col) iter
echo run $run col $col
	   while {1} {
		set hndl [chainNext iter]
		if {$hndl == ""} break
		set a [exprGet $hndl.a]
		set b [exprGet $hndl.b]
		set d [exprGet $hndl.d]
		set f [exprGet $hndl.f]
		set ra1 $a
		set ra2 [expr $a + 1361.*$b]
		set dec1 $d
		set dec2 [expr $d + 2048.*$f]
		pgLine "$ra1 $ra2 $ra2 $ra1 $ra1" \
		   "$dec1 $dec1 $dec2 $dec2 $dec1"
#		pgLine "$ra1 $ra2" "$dec1 $dec1"
#		pgLine "$ra1 $ra2" "$dec2 $dec2"
		}
	   }
	}
   pgSls 1


   pgSlw 3
#Next, MT stuff
   set nights [glob -nocomplain $patchdir/$patch/*]
   set n 0
   foreach night $nights {
	set n [expr $n+0.05]
	regexp {mjd(.+)} $night all mjd
	set kaFiles [glob -nocomplain $night/*.fit]
	incr pgSym
	pgSci $pgSym
	set colors [pgQcr $pgSym]
	set r [int [lindex $colors 0]]
	set g [int [lindex $colors 1]]
	set b [int [lindex $colors 2]]
echo night $night, rgb = $r $g $b
	foreach kaFile $kaFiles {
	   regexp {kaCalObj-[0]*([1-9][0-9]*).fit} $kaFile all seq
#	   set chain [fits2Schema $kaFile KA]
#	   cplot $chain lambda eta y
#	   genericChainDel $chain
	   set scale [expr 1.2/3600.]
	   set ra [expr $sequence($mjd,$seq,ra)+$n]
	   set ra1 [expr $ra-1024.*$scale]
	   set ra2 [expr $ra+1024.*$scale]
	   set dec $sequence($mjd,$seq,dec)
	   set dec1 [expr $dec-1024.*$scale]
	   set dec2 [expr $dec+1024.*$scale]
	   pgLine "$ra1 $ra1 $ra2 $ra2 $ra1" "$dec1 $dec2 $dec2 $dec1 $dec1"
	   }
	}
   pgSlw 1
   pgSls 1
   pgSci 1
   pgLabel "" "" "Patch $patch, Runs [array names asdir]"
   return
   }

#####################################################################
#Make a Gif from the patchPlot

proc patchPlotGif {patch} {
   mtGlobals
   set file patch-$patch.gif
   pgBegin $file/GIF
   pgGeomSet 600 600 0 0

#White background and black foreground
	pgScr 0 1 1 1
	pgScr 1 0 0 0

   patchPlot $patch

   pgEnd
   return
}

#######################################################################3

proc patchPatchPlot {} {
   mtGlobals
   global patchMatch

    if {![info exists patchMatch]} {
	error "Need to run patchPatchMatch" first
    }

    if {[chainTypeDefine $patchMatch] != "KAKAMERGE"} {
	error "Wrong type of chain in patchMatch"
    }

    set subchain [chainSearch $patchMatch {{object1->mag<2> < 15}}]
   cplot $subchain object1->eta res1

   chainDel $subchain
   return
}

#####################################################################
#Make a Gif from the patchPlot
proc patchPatchPlotGif {mjd1 mjd2 {filter 2}} {
   mtGlobals
   global patchMatch

   set i 0
   foreach fil "u g r i z" {
       set filnam($i) $fil
       incr i
   }

    if {![info exists starlist]} {
	error "Need to run singlePatch first"
    }

   patchPatchMatch $mjd1 $mjd2 $filter
   set patch [lindex $starlist(patch) 0]

   set file patchPatch-$patch.gif
   pgBegin $file/GIF
   pgGeomSet 600 600 0 0

#White background and black foreground
	pgScr 0 1 1 1
	pgScr 1 0 0 0


   patchPatchPlot
   pgLabel "" "" "Patch $patch, $mjd1 vs $mjd2, filter $filnam($filter)"
   genericChainDel $patchMatch
   unset patchMatch

   pgEnd
   }

#####################################################################
#Plot the locations of all run and the overlapping MT patchs

proc runPlot {} {
   mtGlobals plan.par
   set runs [lsort -integer [array names tsdir]]
   eval astromPatch $runs
   patchDir

   if {[pgQinf STATE] == "CLOSED"} {
	pgBegin; pgAsk 0
	}
   pgSci 1
   pgEnv 270 100 -10 10 1 0
   pgLabel "Survey Long." "Survey Lat." "Runs"
#Just look at column 1
   set n 1
   foreach run $runs {
	incr n
	pgSci $n
	set col 1
	set topdir $tsdir($run)
#Brian now writes out multiple files per scanline
	set template $topdir/tsObj-[format %06d $run]-$col-*-*.fit
	set files [glob -nocomplain $template]
	if {[string compare $files ""] == 0} return

	set ramin 360.
	set ramax 0
	foreach file $files {
	   set hdr [hdrNew]
	   hdrReadAsFits $hdr $file
	   set field0 [hdrGetAsInt $hdr FIELD0]
	   set nfields [hdrGetAsInt $hdr NFIELDS]
	   hdrDel $hdr
	   loop i 0 [expr $nfields-1] {
		set field [expr $i+$field0]
#Look at astrom solution
		set hndl [chainGetOne $astrom($run,$col) field $field]
		if {$hndl == ""} {
		   echo Missing astrom for field $field
		   continue
		   }
#a is coordinate of beginning of frame
		set a [exprGet $hndl.a]
		set b [exprGet $hndl.b]
		handleDel $hndl
		set amax [expr $a + $b*1361.]
		set ramin [min $ramin $a]
		set ramax [max $ramax $a]
		}
	   }
	pgLine "$ramin $ramax $ramax $ramin $ramin" \
	       "-1.25 -1.25 1.25 1.25 -1.25"
	}

#Now plot patches
   if {![info exists patchPos]} patchDir
   pgSci 1
   foreach patch $patchPos(patch) {
	set ra $patchPos($patch,ra)
	pgLine "$ra $ra" "-1.25 1.25"
	pgText $ra 2 $patch
	}
   return
   }

#####################################################################
#Compute residuals between two different mjd's of a single patch
#Must have run multiPatch first with just one patch.

proc patchPatchMatch {mjd1 mjd2 {filter 2}} {
   global starlist
   global patchMatch
   if {[info exists patchMatch]} {
	genericChainDel $patchMatch
	unset patchMatch
	}
   set patch [lindex $starlist(patch) 0]
   set patchMatch [chainMatchType $starlist($patch,$mjd1) \
	$starlist($patch,$mjd2) {{lambda lambda .001} {eta eta .001}}]
#Compute residuals
   chainTransZXY $patchMatch res1 object1->mag<$filter> \
	object2->mag<$filter> 0 1 -1
   return $patchMatch
   }

################################################################
#Convert a list of kaCalObj objects from eta, lambdat to ra, dec

proc kaConvert {chain} {
   chainInit $chain iter
   while {1} {
	set hndl [chainNext iter]
	if {$hndl == ""} break
	set eta [exprGet $hndl.eta]
	set lambda [exprGet $hndl.lambda]
	set list [surveyToEq $lambda $eta]
	handleSet $hndl.lambda [keylget list ra]
	handleSet $hndl.eta [keylget list dec]
	}
   return
   }

################################################################
#Convert sexadecimal to decimal

proc sex2dec {string} {
   set sign 1
   set hrdeg hr
   set string [string trim $string]
   if {[string index $string 0] == "+"} {
	set hrdeg deg
	set string [string range $string 1 end]
	}
   if {[string index $string 0] == "-"} {
	set sign -1
	set hrdeg deg
	set string [string range $string 1 end]
	}
   set list [split $string :]
   set hr [lindex $list 0]
   set min [lindex $list 1]
   set sec [lindex $list 2]
#Deal with leading 0's
   regsub {0([0-9])} $hr {\1} hr
   regsub {0([0-9])} $min {\1} min
   regsub {0([0-9])} $sec {\1} sec
   set result [expr $sign * ($hr + ($min/60.) + ($sec/3600.))]
   if {$hrdeg == "hr"} {
	set result [expr $result*15.]
	}
   return $result
   }

##############################################################
#Convert GK patch def's to standard survey format

proc gkConvert {chain} {

#Define a new patch format
   typedef struct {
	double ramin;
	double ramax;
	double decmin;
	double decmax;
	char zone[2];
	} PATCH;

   set newChain [chainNew PATCH]
   chainInit $chain iter
   while {1} {
	set hndl [chainNext iter]
	if {$hndl == ""} break
	set new [genericNew PATCH]
	handleSet $new.zone [exprGet $hndl.zone]
	foreach var "ramin ramax decmin decmax" {
	   set $var [exprGet $hndl.$var]
	   set $var [sex2dec [set $var]]
	   handleSet $new.$var [set $var]
	   }
	chainAdd $newChain $new
	}
   return $newChain
   }

##############################################################
#Input is a list of lists (many variations possible).
#Expans any range specifications.  Return a single list

proc fieldExpand {args} {
   set outlist ""
   foreach arg $args {
        foreach field $arg {
	   if {[llength $field] > 1} {
		set list [eval fieldExpand $field]
		foreach subfield $list {
		   lappend outlist $subfield
		   }
		continue
		}
           if {[regexp {([0-9]+)-([0-9]+)} $field all f1 f2]} {
                loop i $f1 [expr $f2+1] {
                   lappend outlist $i
                   }
           } else {
                lappend outlist $field
                }
           }
        }
   return $outlist
   }

##########################################################################
proc dataDirs {run} {
   mtGlobals

### Log file directory
#Grr, Yanny is now creating lots of directories for log files.
#The files that I want are often in logs-0, etc.
#Make a list of all possible directories.
    set dirs [glob -nocomplain $masterdir/$run/log $masterdir/$run/logs \
		  $masterdir/$run/logs-\[0-9\] $masterdir/$run/\[0-9\]/logs]
    if {[info exists logdir($run)]} {unset logdir($run)}
    foreach dir $dirs {
        set files [glob -nocomplain $dir/idFrameLog*]
        if {[llength $files] == 0} continue
        set logdir($run) $dir
        break
    }
   echo logdir($run) = $logdir($run)

### tsObj file directory
   set tsdir($run) $masterdir/$run/calibChunks
    if {![file exists $tsdir($run)]} {
	set file [lindex [glob -nocomplain \
	   $masterdir/$run/calibChunks-\[0-9\]] 0]
	if {[file exists $file]} {set tsdir($run) $file}
    }
   echo tsdir(run) = $tsdir($run)

### astrom directory
   set asdir($run) $masterdir/$run/astrom
    if {![file exists $asdir($run)]} {
	set file [lindex [glob -nocomplain $masterdir/$run/astrom-\[0-9\]] \
		      0]
	if {[file exists $file]} {set asdir($run) $file}
    }
   if {![file exists $asdir($run)]} {
	set file [lindex [glob -nocomplain $masterdir/$run/\[0-9\]/astrom] end]
	if {[file exists $file]} {set asdir($run) $file}
	}
   echo asdir($run) = $asdir($run)

### objcs directory
   set photodir($run) $masterdir/$run/objcs
    if {![file exists $photodir($run)]} {
	set file [lindex [glob -nocomplain $masterdir/$run/objcs-\[0-9\]] \
			  0]
	if {[file exists $file]} {set photodir($run) $file}
	}
   echo photodir($run) = $photodir($run)

### corrected frames directory
	if {[file exists $masterdir/$run/corr]} {
        set corrdir($run) $masterdir/$run/corr
	} else {
        set corrdir($run) $photodir($run)
        }
   echo corrdir($run) = $corrdir($run)

### final calibrations directory
   set fcdir($run) $masterdir/$run/fcalib
   if {![file exists $fcdir($run)]} {
	set file [lindex [glob -nocomplain $masterdir/$run/fcalib-\[0-9\]] \
			  0]
	if {[file exists $file]} {set fcdir($run) $file}
	}
   if {![file exists $fcdir($run)]} {
	set file $masterdir/$run/nfcalib
	if {[file exists $file]} {set fcdir($run) $file}
	}

###Sometimes the ncalib directory lives one level down.
   if {![file exists $fcdir($run)]} {
	if {![catch {set calibdir [file readlink $tsdir($run)]}]} {
	   set rerun [file dirname $calibdir]
	   set file $masterdir/$run/$rerun/nfcalib
	   if {[file exists $file]} {set fcdir($run) $file}
	   }
	}

#Backup fcalib directory
   if {![file exists $fcdir($run)] && [info exists fcalibdir]} {
	set file $fcalibdir
	if {[file exists $file]} {set fcdir($run) $file}
	}
   echo fcdir($run) = $fcdir($run)

   return
    }
