# Top level commands
# First create a plan.par file with entries
#   dataDir
#   outDir
#   biasDir
#   thresh
# oddEvenCopyJob plan.par
# oddEvenCheckJob plan.par
# oddEvenJob plan.par
# For boundary frames...
# oddEvenSplitJob plan.par frame row  (+ to fix bottom of frame, - to fix top).
#
# Biases:
# oddEvenBiasCheckJob plan.par
# oddEvenBiasJob plan.par
#
#Fix a region for the odd/even effect

#We compute a vector that will be subtracted from the region in order
#to remove some desired effect.  We may wish to match a bias frame in
#which the oddEven effect is actually present.
#   shift: Mean level in original frame minus mean level in corrected frame.
#	This is normally the mean level in the matching bias minus the
#	mean level in the original frame.
#   diffoff: After computing odd/even difference in original frame, subtract
#	diffoff.  This is normally the odd/even difference in the matching
#	bias frame.
proc oddEven {reg {shift 0} {diffoff 0}} {
   global DEBUG
#Compute median
#Get subregions first.  Allow for the fact that "reg" might be a subregion
   set nrow [exprGet $reg.nrow]
   set subreg1 [subRegNew $reg $nrow 16 0 0]
   set subreg2 [subRegNew $reg $nrow 16 0 2112]
   set median1 [colMed $subreg1]
   set median2 [colMed $subreg2]
   regDel $subreg1
   regDel $subreg2
   set low 0; set high 0
   loop i 0 8 {
	set low [expr $low + [regPixGet $median1 0 [expr 2*$i]] + \
	   [regPixGet $median2 0 [expr 2*$i]]]
	set high [expr $high + [regPixGet $median1 0 [expr 2*$i+1]] + \
	   [regPixGet $median2 0 [expr 2*$i +1]]]
	}
   set low [expr $low/16.]
   set high [expr $high/16.]
   regDel $median1
   regDel $median2

#Check to see if TSCALE parameter is in header; if so, divide by it.
   if {![catch {set scale [hdrGetAsInt $reg.hdr TSCALE]}]} {
	if {[info exists DEBUG]} {echo Scaling by $scale}
	set low [expr $low/$scale]
	set high [expr $high/$scale]
	}

   set mean [expr ($low + $high)/2.]
   set diff [expr ($high-$low)]
   if {[info exists DEBUG]} {
	echo Mean [format %.1f $mean], diff [format %.1f $diff]
	}

#New low, high values for median vector
   set low [expr -1.*(($diff-$diffoff)/2.) - $shift]
   set high [expr $low + $diff - $diffoff]

   set median [regNew 1 2128 -type FL32]
   loop i 0 1064 {
	regPixSetWithDbl $median 0 [expr 2*$i] $low
	regPixSetWithDbl $median 0 [expr 2*$i+1] $high
	}
   return $median
   }

######################################################################
#Process the specified file.
proc oddEvenFrame {file {shift 0} {diffoff 0}} {
   set reg [regNew]
   regReadAsFits $reg $file
   set vec [oddEven $reg $shift $diffoff]

#Dither in the difference
   imgColSub $reg $vec -dither
   regDel $vec
   regWriteAsFits $reg $file
   regDel $reg
   return
   }

#####################################################################
#Process all frames according to a plan

proc oddEvenJob {{plan oePlan.par} {qcpar ""}} {
   global DEBUG
#Run copy job
   set planDir [file dirname $plan]
   oddEvenCopyJob $plan

   inipar $plan
   set dataDir [getpars dataDir]
   set outDir [getpars outDir]

   if {$qcpar == ""} {
	set thresh [getpars thresh]
   } else {
	inipar $qcpar
	set thresh [getpars oe_thresh]
	inipar $plan
	}
   set fid [open $planDir/parameters.par]
   while {1} {
	set line [string trim [gets $fid]]
	if {[eof $fid]} break
	set file $dataDir/$line
	set outfile $outDir/[file tail $file]

#Get filter and shift
	if {![regexp {idR-[0-9]+-(g|r).+} $line all filter]} {
	   echo Can't get filter for $line
	   continue
	   }
	set shift [getpars ${filter}shift]
	exec cp $file $outfile
	set bdiff [oddEvenBias $plan $filter]

#The sign of shift is that we expect bias to have odd/even pattern,
#frames not to.  If the opposite, holds, we need to reverse shift
	if {abs($bdiff) < $thresh} {
	   set shift [expr -1.*$shift]
	   } 
        echo Processing [file tail $outfile]
        oddEvenFrame $outfile $shift $bdiff
        }
   close $fid
   return 0
   }

######################################################################
#Process the specified file as a split frame.  We check each odd/even
#difference against that of the bias.

proc oddEvenSplitFrame {file row thresh shift bdiff} {
   global DEBUG
   set reg [regNew]
   regReadAsFits $reg $file

   set nrow [exprGet $reg.nrow]
   set ncol [exprGet $reg.ncol]

#Check upper part of frame

   set sub [subRegNew $reg $row $ncol 0 0]
   set diff [oddEvenCheck $sub]
   set diffdiff [expr $diff - $bdiff]
   if {abs($diffdiff) > $thresh} {
	echo Repairing top section of [file tail $file]
	set vec [oddEven $sub $shift $bdiff]
	imgColSub $sub $vec -dither
	regDel $vec
	}
   regDel $sub

#Check lower part of frame
   set sub [subRegNew $reg [expr $nrow-$row] $ncol $row 0]
   set diff [oddEvenCheck $sub]
   set diffdiff [expr $diff - $bdiff]
   if {abs($diffdiff) > $thresh} {
	echo Repairing bottom section of [file tail $file]
	set vec [oddEven $sub $shift $bdiff]
	imgColSub $sub $vec -dither
	regDel $vec
	}
   regDel $sub

   regWriteAsFits $reg $file
   regDel $reg
   return
   }

#####################################################################
#Process a single frame as a split frame.  Read dirs from plan

proc oddEvenSplitJob {plan file row} {
   inipar $plan
   set dataDir [getpars dataDir]
   set outDir [getpars outDir]
   set thresh [getpars thresh]
   set file $dataDir/[file tail $file]
   set outfile $outDir/[file tail $file]
   echo Processing $file to $outfile
   exec cp $file $outfile
#Get filter and shift
   if {![regexp {idR-[0-9]+-(g|r).+} [file tail $file] all filter]} {
	echo Can't get filter for $line
	continue
	}
   set shift [getpars ${filter}shift]

#Get diff for bias
   set bdiff [oddEvenBias $plan $filter]

#The sign of shift is that we expect bias to have odd/even pattern,
#frames not to.  If the opposite, holds, we need to reverse shift
   if {abs($bdiff) < $thresh} {
	set shift [expr -1.*$shift]
	}
   oddEvenSplitFrame $outfile $row $thresh $shift $bdiff
   return
   }

#####################################################################
#Copy all files from dataDir to outDir.  This job is
#run before oddEvenJob

proc oddEvenCopyJob {plan} {
   global DEBUG
   inipar $plan
   set dataDir [getpars dataDir]
   set outDir [getpars outDir]
   set files [split [exec find $dataDir -follow -type f -name *.fit] \n]
   foreach file $files {
	set outfile $outDir/[file tail $file]
	if {[file exists $outfile]} {
	   if {[file size $file] == [file size $outfile]} {
		if {[info exists DEBUG]} {echo $outfile in place}
		continue
		}
	   }
        echo Copying $file to $outfile
	exec cp $file $outfile
	exec chmod u+w $outfile
        }
   return
   }

#######################################################################
#Check a region for the odd/even effect
proc oddEvenCheck {reg {shift 0} {diffoff 0}} {
   global DEBUG
#Compute median
#Get subregions first
   set nrow [exprGet $reg.nrow]
   set subreg1 [subRegNew $reg $nrow 16 0 0]
   set subreg2 [subRegNew $reg $nrow 16 0 2112]
   set median1 [colMed $subreg1]
   set median2 [colMed $subreg2]
   regDel $subreg1
   regDel $subreg2
   set low 0; set high 0
   loop i 0 8 {
	set low [expr $low + [regPixGet $median1 0 [expr 2*$i]] + \
	   [regPixGet $median2 0 [expr 2*$i]]]
	set high [expr $high + [regPixGet $median1 0 [expr 2*$i+1]] + \
	   [regPixGet $median2 0 [expr 2*$i +1]]]
	}
   set low [expr $low/16.]
   set high [expr $high/16.]

#Check to see if TSCALE parameter is in header; if so, divide by it.
   if {![catch {set scale [hdrGetAsInt $reg.hdr TSCALE]}]} {
	if {[info exists DEBUG]} {echo Scaling by $scale}
	set low [expr $low/$scale]
	set high [expr $high/$scale]
	}

   regDel $median1
   regDel $median2
   set mean [expr ($low + $high)/2.]
   set diff [expr ($high-$low)]
   if {[info exists DEBUG]} {
	echo Mean [format %.1f $mean], diff [format %.1f $diff]
	}

#New low, high values for median vector
   set low [expr -1.*(($diff-$diffoff)/2.) - $shift]
   set high [expr $low + $diff - $diffoff]

   return $diff
   }

#######################################################################
#Check a region for the odd/even effect
proc oddEvenCheckFrame {file} {

   set reg [regNew]
   regReadAsFits $reg $file
   set diff [oddEvenCheck $reg]
   regDel $reg
   return $diff
   }

#####################################################################
#Verify after running oddEven
proc oddEvenVerifyJob {{plan oePlan.par} {qcpar ""} {qcfile ""}} {
   global VERIFY
   set VERIFY 1
   set code [oddEvenCheckJob $plan $qcpar $qcfile]
   unset VERIFY
   return $code
   }

#####################################################################
#Check all frames according to a plan
#When calling oddEvenCheckJob directory, qcfile is not used and can
#be omitted.

proc oddEvenCheckJob {{plan oePlan.par} {qcpar ""} {qcfile ""}} {
   global DEBUG
   global VERIFY
   inipar $plan
   set dataDir [getpars dataDir]
   set outDir [getpars outDir]
   set planDir [file dirname $plan]

#Read threshold from qcpar file, if provided, else the oePlan.par file
   if {$qcpar == ""} {
	set thresh [getpars thresh]
   } else {
	inipar $qcpar
	set thresh [getpars oe_thresh]
	}
   set ngood 0
   set nbad 0

#Track transition fields.
   set transfield ""
   set badfield ""
   if {[info exists VERIFY]} {
	set dir $outDir
   } else {
	set dir $dataDir
	set fid [open $planDir/parameters.par w]
	}
   foreach filter "g r" {

#First, get diff for bias
	set bdiff [oddEvenBias $plan $filter]
	set files [lsort [glob $dir/idR-*-$filter*.fit]]

#Search for transitions
	set laststate good
	set oldfile [lindex $files 0]
	set startField 9999
	set endField 0
	foreach file [lsort $files] {
	   if {[info exists DEBUG]} {echo [file tail $file]}
	   if {![regexp {idR-[0-9]+-[ugriz][0-9]+-([0-9]+).fit} $file all field]} {\
		echo Can't decode file name $file
		set field -1
		}
	   set field [string trimleft $field 0]
	   if {$field == ""} {set field 0}
	   set startField [min $startField $field]
	   set endField [max $endField $field]
	   set diff [oddEvenCheckFrame $file]
	   set diffdiff [expr $diff - $bdiff]
	   if {abs($diffdiff) > $thresh} {
		if {![info exists VERIFY]} {
		   puts $fid [file tail $file]
		   }
		echo [file tail $file] exceeds threshold
		incr nbad
		if {$file != $oldfile && $laststate == "good"} {
		   echo Check [file tail $oldfile] [file tail $file] \
			for transitions.
#Get actual field number
#Both the current field and the one before are suspect.
		   lappend transfield [list [expr $field-1] $filter]
		   lappend transfield [list $field $filter]
		   }

#Make TCL list of all bad fields
		lappend badfield [list $field $filter]
		set laststate bad
	   } else {
		incr ngood
		if {$file != $oldfile && $laststate == "bad"} {
		   echo Check [file tail $oldfile] \
			[file tail $file] for transitions.
#Get actual field number
#Both the current field and the one before are suspect.
		   lappend transfield [list [expr $field-1] $filter]
		   lappend transfield [list $field $filter]
		   }
		set laststate good
		}
	   set oldfile $file
	   }
	}
   if {![info exists VERIFY]} {
	close $fid

#Write out transition fields
	set fid [open $planDir/oeTransition.par w]
	foreach pair $transfield {
	   puts $fid $pair
	   }
	close $fid
   } elseif {$qcfile != ""} {

#Write out qc file	
#Output the start and end fields, the tolerance, and any bad info.
	set fid [open $qcfile w]
	puts $fid "startField $startField"
	puts $fid "endField $endField"
	puts $fid "oe_thresh $thresh"
	if {[llength $badfield] > 0} {
	   puts $fid camcol nonewline
	   loop i 0 [llength $badfield] {
		puts $fid " 1" nonewline
		}
	   puts $fid ""
	   puts $fid field nonewline
	   foreach pair $badfield {
		puts $fid " [lindex $pair 0]" nonewline
		}
	   puts $fid ""
	   puts $fid filter nonewline
	   foreach pair $badfield {
		puts $fid " [lindex $pair 1]" nonewline
		}
	   puts $fid ""
	   }
	close $fid
	}
   echo nbad = $nbad, ngood = $ngood
   if {$nbad == 0} {
	return 0
   } else {
	return 1
	}
   }

#####################################################################

#Compute diff for a bias frame in a specified filter

proc oddEvenBias {plan filter} {
   global DEBUG
   inipar $plan
   set biasDir [getpars biasDir]
   set thresh [getpars thresh]
   set files [glob $biasDir/idB*.fit]
   set file [lindex [glob -nocomplain $biasDir/idB*${filter}1*.fit] 0]
   if {$file == ""} {
	echo No bias file for filter $filter
	return
	}
   set diff [oddEvenCheckFrame $file]

   if {[info exists DEBUG]} {
	echo Diff for bias in $filter = [format %.1f $diff]
	}
   return $diff
   }

###################################################################
#Make a plan.  Presume default directory structure layout

proc oddEvenPlan {run biasrun {rerun 0}} {

#top directory
   set topdir /data/dp3.b/data/$run/$rerun
   if {![file exists $topdir]} {
	echo Can't find $topdir
	return
	}
   echo topdir is $topdir
#   set biasDir $topdir/photo/calib
#Biases are usually not available yet.  Do next best thing
   set biasDir /data/dp3.b/data/standards/calib/$biasrun

   if {![file exists $biasDir]} {
	echo Caution: Can't find $biasDir
	set biasDir ""
	}
   echo bias directory is $biasDir
   set bias [lindex [glob $biasDir/idB-*-g1.fit] 0]
   if {![file exists $bias]} {
	echo Caution: can't find bias file $bias
   } else {
	echo bias frame is $bias
	set hdr [hdrNew]
	hdrReadAsFits $hdr $bias
	set brun [hdrGetAsInt $hdr RUN]
	hdrDel $hdr
	echo bias run is $brun
	}
   echo Creating oePlan.par
   set fid [open oePlan.par w]

#Data directory
   set dataDir $topdir/fields/1
   if {![file exists $dataDir]} {
	echo Caution: No data directory $dataDir
   } else {
	echo dataDir is $dataDir
	}
   puts $fid "dataDir $dataDir"

#Output directory
   set outrerun [expr $rerun+1]
   set outDir /data/dp3.b/data/$run/$outrerun/fields/1
   if {![file exists $outDir]} {
	echo Caution: No output directory $outDir
   } else {
	echo outDir is $outDir
	}
   puts $fid "outDir $outDir"

#Bias directory - use the one in the run directory, not in standars/calib
   puts $fid "biasDir $biasDir"

#Hardwired constants
   puts $fid "thresh 10"
   puts $fid "gshift -8.2"
   puts $fid "rshift -5.4"
   close $fid
   return
   }

