#Fetch photometric zero points and sky values from opdb
#Procedures (Must be running photoSch on sdssdp11)
#pget <run>		Fetch info for a run
#photoComp		Summary sky residuals
#photoPlot <filter>	photoComp for 1 filter with plots
#photoWrite		Write out residuals to pres.tcl
#

proc photoGet {run rerun {altfield0 ""} {altnfield ""}} {
   global median photom secz extinct za
   if {![inSession]} {
	opdbConnect
	}

#Convert filter to ccdrow
   set ccdrow(u) 3
   set ccdrow(g) 5
   set ccdrow(r) 1
   set ccdrow(i) 2
   set ccdrow(z) 4

   foreach camcol "1 2 3 4 5 6" {
echo Camcol $camcol
	set frames [pipeGet frames $run $camcol -rerun $rerun]
	set calib [$frames -> currentPhotometricCalibration]
	if {![$calib isValid]} {
	   echo No valid calibration for run $run rerun $rerun camcol $camcol
	   objDel $calib
	   objDel $frames
	   return
	   }
	set field0 [$frames -> field0]
	set nFields [$frames -> nFields]
	if {$altfield0 != ""} {set field0 $altfield0}
	if {$altnfield != ""} {set nFields $altnfield}

#Fetch calibrations from field 0
	set ptrans [$calib -> ptrans $field0]
	foreach filter "u g r i z" {
	   set trans [$ptrans $filter]
	   set a [$trans a]
	   set k [$trans a]
	   set photom($run,$camcol,$filter) [format %.3f [$a value]]
	   set extinct($run,$camcol,$filter) [format %.3f [$k value]]
	   handleDel $a
	   handleDel $trans
	   }
	handleDel $ptrans

#Fetch skies
	foreach filter "u g r i z" {
	   set skies($run,$camcol,$filter) ""
	   }
	loop field $field0 [expr $field0 + $nFields] {
	   set objList [$frames -> objectList $field]
	   foreach filter "u g r i z" {
		set string AR_[string toupper $filter]
		set int [enumStringToInt arFilter $string]
		set frameStat [$objList -> frameStat $int]
		set sky [$frameStat sky_frames]
		handleDel $frameStat
		lappend skies($run,$camcol,$filter) [format %.2f $sky]
		}
	   objDel $objList
	   }
#Fetch airmass
	foreach filter "u g r i z" {
	   set secz($run,$camcol,$filter) ""
	   }
	set asCalib [$frames -> currentAstrometricCalibration]
	loop field $field0 [expr $field0 + $nFields] {
	   foreach filter "u g r i z" {
		set trans [$asCalib -> trans $ccdrow($filter)$camcol $field]
		set airmass [$trans airmass]

#Use elevation angle - weirdly, sky brightness changes more linearly with
#this than secz!
		set airmass [expr 180.*acos(1./$airmass)/3.14159]
		handleDel $trans
		lappend secz($run,$camcol,$filter) [format %.2f $airmass]
		}
	   }
	objDel $asCalib
	objDel $frames

#Median of skies
	foreach filter "u g r i z" {
	   set skies($run,$camcol,$filter) \
		 [lsort -real $skies($run,$camcol,$filter)]
	   set len [llength $skies($run,$camcol,$filter)]
	   set median($run,$camcol,$filter) \
		[lindex $skies($run,$camcol,$filter) [expr $len/2]]
	   }
#Average airmass
	foreach filter "u g r i z" {
	   set sum 0; set n 0
	   foreach airmass $secz($run,$camcol,$filter) {
		set sum [expr $sum+$airmass]
		incr n
		}
	   set za($run,$camcol,$filter) [format %.3f [expr $sum/$n]]
	   }
	}

#If all worked successfully, register run number
   if {[info exists photom()]} {
	lappend photom() $run
   } else {
	set photom() $run
	}
   return
   }

#####################################################################
#Make plot of zero point vs. sky.  I normalize both the averages across
#6 camcols
proc photoPlot {filter} {
   global median photom slistres gres balance za

   echo Filter $filter
#Slope: d(sky)/d(za) in mags per degree
   set slope(u) .0068
   set slope(g) .0114
   set slope(r) .0131
   set slope(i) .0147
   set slope(z) .0150
   if {![info exists photom]} return
   if {[pgQinf STATE] == "CLOSED"} {
	pgBegin; pgGeomSet 400 400 10 25; pgAsk 0
	}
   foreach run $photom() {
	set aavg ""
	set savg ""
	foreach camcol "1 2 3 4 5 6" {
	   set alist($camcol) $photom($run,$camcol,$filter)
	   set slist($camcol) [expr -2.5*log10($median($run,$camcol,$filter))]

#Correct for zenith angle.  Make things fainter.
	   set slist($camcol) [expr $slist($camcol) + \
		$slope($filter) * $za($run,$camcol,$filter)]
	   set aavg [expr $aavg + $alist($camcol)]
	   set savg [expr $savg + $slist($camcol)]
	   }
	set aavg [expr $aavg/6.]
	set savg [expr $savg/6.]
	foreach camcol "1 2 3 4 5 6" {
	   set alist($run,$camcol) [expr $alist($camcol) - $aavg]
	   set slist($run,$camcol) [expr $slist($camcol) - $savg]

#Residual
	   set slistres($run,$camcol) [expr $slist($run,$camcol) - \
		$alist($run,$camcol)]
	   }
	}

#Compute mean sky and photometric zero pt. balance factors by themselves,
#then the difference.
   foreach camcol "1 2 3 4 5 6" {
	set alist($camcol) 0
	set slist($camcol) 0
	}
   foreach run $photom() {
	foreach camcol "1 2 3 4 5 6" {
	   set alist($camcol) [expr $alist($camcol) + $alist($run,$camcol)]
	   set slist($camcol) [expr $slist($camcol) + $slist($run,$camcol)]
	   }
	}
   foreach camcol "1 2 3 4 5 6" {
	set alist($camcol) [expr $alist($camcol)/[llength $photom()]]
	set slist($camcol) [expr $slist($camcol)/[llength $photom()]]
	set difference [expr $alist($camcol) - $slist($camcol)]
#echo Camcol $camcol, mean z.p. = [format %8.3f $alist($camcol)], \
		      mean sky = [format %8.3f $slist($camcol)], \
            zp - sky = [format %8.3f $difference]
	}

#Now compute rms scatter in sky by itself, photo zp by themselves.
   foreach camcol "1 2 3 4 5 6" {
	set arms($camcol) 0
	set srms($camcol) 0
	}
   foreach run $photom() {
	foreach camcol "1 2 3 4 5 6" {
	   set arms($camcol) [expr $arms($camcol) + pow($alist($run,$camcol) \
		- $alist($camcol),2)]
	   set srms($camcol) [expr $srms($camcol) + pow($slist($run,$camcol) \
		- $slist($camcol),2)]
	   }
	}
   foreach camcol "1 2 3 4 5 6" {
	set arms($camcol) [expr sqrt($arms($camcol)/[llength $photom()])]
	set srms($camcol) [expr sqrt($srms($camcol)/[llength $photom()])]
	set difference [expr $alist($camcol) - $slist($camcol)]

#Print this stuff out with the tables below.
#echo Camcol $camcol, rms z.p. = [format %8.3f $arms($camcol)], \
#		      rms sky = [format %8.3f $srms($camcol)]
	}

#Print out sky residuals by run
#List residuals
   echo
   echo Sky Residuals
   set format "%-4s%6s%8s%8s%8s%8s%8s"
   puts stdout [format $format Run 1 2 3 4 5 6]
   set format "%4s%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f"
   foreach camcol "1 2 3 4 5 6" {
	set rms($camcol) 0
	}
   set n 0
   foreach run [lsort -integer $photom()] {
	incr n
	foreach camcol "1 2 3 4 5 6" {
	   set diff($camcol) [expr $slist($run,$camcol) - $slist($camcol)]
	   }
	puts stdout [format $format $run \
	   $diff(1) $diff(2) $diff(3) $diff(4) $diff(5) $diff(6)]
	}
   puts stdout ""
   set format "%4s%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f"
   puts stdout [format $format "" $srms(1) $srms(2) $srms(3) $srms(4) \
	$srms(5) $srms(6)]
   puts stdout ""

#Print out a residuals by run
#List residuals
   echo
   echo Zero point Residuals
   set format "%-4s%6s%8s%8s%8s%8s%8s"
   puts stdout [format $format Run 1 2 3 4 5 6]
   set format "%4s%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f"
   foreach camcol "1 2 3 4 5 6" {
	set rms($camcol) 0
	}
   set n 0
   foreach run [lsort -integer $photom()] {
	incr n
	foreach camcol "1 2 3 4 5 6" {
	   set diff($camcol) [expr $alist($run,$camcol) - $alist($camcol)]
	   }
	puts stdout [format $format $run \
	   $diff(1) $diff(2) $diff(3) $diff(4) $diff(5) $diff(6)]
	}
   puts stdout ""
   set format "%4s%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f"
   puts stdout [format $format "" $arms(1) $arms(2) $arms(3) $arms(4) \
	$arms(5) $arms(6)]
   puts stdout ""

   pgAsk 1

#Make a plot of the sky fluxes themselves - look to see how sky ratios vary
#among runs
   pgSci 1
   pgEnv 0 7 -.5 .5 0 0
   pgLabel "Camcol" "Sky flux" "Normalize sky flux"
   set sci 0

   foreach run $photom() {
	incr sci
	if {$sci > 7} {set sci 1}
	pgSci $sci
	foreach camcol "1 2 3 4 5 6" {
	   pgPoint $camcol $slist($run,$camcol) 3
	   }
       }

#Make a plot of the zero points themselves - look to see how they
#among runs
   pgSci 1
   pgEnv 0 7 -.5 .5 0 0
   pgLabel "Camcol" "Zero point" "Normalize zero point"
   set sci 0
   foreach run $photom() {
	incr sci
	if {$sci > 7} {set sci 1}
	pgSci $sci
	foreach camcol "1 2 3 4 5 6" {
	   pgPoint $camcol $alist($run,$camcol) 3
	   }
       }


#Make plot of photometric calibs relative to sky
   pgSci 1
   pgEnv 0 7 -.2 .2 0 0
   pgLabel "Normalized zeropoint" "Sky Residual (mag)" "Sky vs. zero pt."
   pgLine "0 7" "0 0"
   set sci 0
   foreach run $photom() {
	incr sci
	if {$sci > 7} {set sci 1}
	pgSci $sci
	foreach camcol "1 2 3 4 5 6" {
	   pgPoint $camcol $slistres($run,$camcol) 3
	   }
       }

   pgSci 1

#Compute a global residual unless we already read in balance factors
   if {![info exists balance(1,$filter)]} {
	foreach camcol "1 2 3 4 5 6" {
	   set gres($camcol) ""
	   foreach run $photom() {
		lappend gres($camcol) $slistres($run,$camcol)
		}
	   set gres($camcol) [lindex [lsort -real $gres($camcol)] \
		[expr [llength $photom()]/2]]
	   echo Camcol $camcol Balance factor [format %.3f $gres($camcol)]
	   echo Setting balance factor
	   set balance($camcol,$filter) $gres($camcol)
	   }
   } else {
	foreach camcol "1 2 3  4 5 6" {
	   set gres($camcol) $balance($camcol,$filter)
	   }
	}

#List residuals
   set format "%-4s%6s%8s%8s%8s%8s%8s"
   puts stdout [format $format Run 1 2 3 4 5 6]
   set format "%4s%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f"
   foreach camcol "1 2 3 4 5 6" {
	set rms($camcol) 0
	}
   set n 0
   foreach run [lsort -integer $photom()] {
	incr n
	foreach camcol "1 2 3 4 5 6" {
	   set diff($camcol) [expr $slistres($run,$camcol) - $gres($camcol)]
	   set rms($camcol) [expr $rms($camcol) + pow($diff($camcol),2)]
	   }
	puts stdout [format $format $run \
	   $diff(1) $diff(2) $diff(3) $diff(4) $diff(5) $diff(6)]
	}
   set rrms 0
   foreach camcol "1 2 3 4 5 6" {
	set rms($camcol) [expr sqrt($rms($camcol)/$n)]
	set rrms [expr $rrms + pow($rms($camcol),2)]
	}
   set rrms [expr sqrt($rrms/6.)]
   puts stdout ""
   set format "%4s%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f"
   puts stdout [format $format "" $rms(1) $rms(2) $rms(3) $rms(4) \
	$rms(5) $rms(6)]
   puts stdout ""
   puts stdout "Grand rms: [format %8.3f $rrms]"
   return
   }

#######################################################################
#Easier routine to get stuff.  Allow optional start field, nfield

proc pget {run {field0 ""} {nfield ""}} {
   global runTable
   if {![inSession]} {opdbConnect}
   if {![info exists runTable]} {runInventory}
   if {[lsearch $runTable() $run] < 0} {
	echo No run $run
	return
	}
   photoGet $run $runTable($run) $field0 $nfield
   return
   }

#####################################################################
#Make plot of zero point vs. sky.  I normalize both the averages across
#6 camcols.  PhotoComp just computes the sky offsets relative to nominal
#balance factors in 5 filters.  No plots.

proc photoComp {} {
   global median photom slistres gres balance za pres

#Slope: d(sky)/d(za) in mags per degree
   set slope(u) .0068
   set slope(g) .0114
   set slope(r) .0131
   set slope(i) .0147
   set slope(z) .0150
   if {![info exists photom]} return

#Compute mean in each CCD - look for trends (i.e. potential flaws) in
#balance factors.
   foreach camcol "1 2 3 4 5 6" {
	foreach filter "u g r i z" {
	   set sum($camcol,$filter) 0
	   }
	}

   if {[info exists pres]} {unset pres}
   set pres() [lsort -integer $photom()]

foreach run [lsort -integer $photom()] {
#List residuals
   set format1 "%-4s  %6s%8s%8s%8s%8s%8s"
   puts stdout [format $format1 $run 1 2 3 4 5 6]
   set format "%4s  %8.2f%8.2f%8.2f%8.2f%8.2f%8.2f"
   foreach filter "u g r i z" {
	set aavg ""
	set savg ""
	foreach camcol "1 2 3 4 5 6" {
	   set alist($camcol) $photom($run,$camcol,$filter)
	   set slist($camcol) [expr -2.5*log10($median($run,$camcol,$filter))]

#Correct for zenith angle.  Make things fainter.
	   set slist($camcol) [expr $slist($camcol) + \
		$slope($filter) * $za($run,$camcol,$filter)]
	   set aavg [expr $aavg + $alist($camcol)]
	   set savg [expr $savg + $slist($camcol)]
	   }
	set aavg [expr $aavg/6.]
	set savg [expr $savg/6.]
	foreach camcol "1 2 3 4 5 6" {
	   set alist($run,$camcol) [expr $alist($camcol) - $aavg]
	   set slist($run,$camcol) [expr $slist($camcol) - $savg]

#Residual
	   set slistres($run,$camcol) [expr $slist($run,$camcol) - \
		$alist($run,$camcol)]
	   }

#Compute a global residual unless we already read in balance factors
	if {![info exists balance(1,$filter)]} {
	   echo No balance factors found.  Skipping filter $filter
	   continue
	   }
	foreach camcol "1 2 3  4 5 6" {
	   set gres($camcol) $balance($camcol,$filter)
	   }

	foreach camcol "1 2 3 4 5 6" {
	   set diff($camcol) [expr $slistres($run,$camcol) - $gres($camcol)]
	   set pres($run,$camcol,$filter) $diff($camcol)
	   set sum($camcol,$filter) \
		[expr $sum($camcol,$filter) + $diff($camcol)]
	   }
	puts stdout [format $format $filter \
	   $diff(1) $diff(2) $diff(3) $diff(4) $diff(5) $diff(6)]
	}
#Don't compute rms.  I want succinct output.
}
   set nrun [llength $photom()]
   puts stdout ""
   puts stdout [format $format1 "" 1 2 3 4 5 6]
   foreach filter "u g r i z" {
	foreach camcol "1 2 3 4 5 6" {
	   set diff($camcol) [expr 1.*$sum($camcol,$filter)/$nrun]
	   }
	puts stdout [format $format $filter \
	   $diff(1) $diff(2) $diff(3) $diff(4) $diff(5) $diff(6)]
	}

   return
   }

#####################################################################
#Print a file with sky residuals.

proc photoWrite {} {
   global median photom slistres gres balance za pres

   if {![info exists pres]} return

   set fid [open pres.tcl w]
   puts $fid "global pres"
   puts $fid {if {[info exists pres]} {unset pres}}
   puts $fid "set pres() \"$pres()\""

   foreach run $pres() {
#List residuals
	foreach filter "u g r i z" {
	   foreach camcol "1 2 3 4 5 6" {
		puts $fid "set pres($run,$camcol,$filter) [format %.3f \
		   $diff($camcol)]"
		}
	   }
	}
   close $fid
   return
   }

########################################
#Helper procs to calculate sky vs. za from theory
proc brighten {filter za} {
   set k(u) .55
   set k(g) .22
   set k(r) .13
   set k(i) .08
   set k(z) .08

   set secz [expr 1./cos($za/57.3)]
   set brighten [expr -$k($filter)*($secz-1) + 2.5*log10($secz)]
   return $brighten
   }

proc btable {} {
   loop za 0 70 10 {
	puts stdout $za nonewline
	foreach filter "u g r i z" {
	   puts stdout [format %7.2f [brighten $filter $za]] nonewline
	   }
	puts stdout ""
	}
   return
   }

##################################################################
#Predict median offsets in QA plots that Zeljko products
#Input is 1st run, it's strip, and 2nd run

proc medOffset {run1 strip1 run2} {
   global gres slistres
   foreach run "$run1 $run2" {
	foreach camcol "1 2 3 4 5 6" {
	   set diff($run,$camcol) \
		[expr $slistres($run,$camcol) - $gres($camcol)]
	   }
	}
   foreach camcol "1 2 3 4 5 6" {
	set res [format %.2f [expr $diff($run1,$camcol) - \
	   $diff($run2,$camcol)]]
	puts stdout "$camcol-$camcol: $res"
	if {$camcol == 6} continue
	if {$strip1 == "N"} {
	   set res [format %.2f [expr $diff($run1,$camcol) - \
		$diff($run2,[expr $camcol+1])]]
	} else {
	   set res [format %.2f [expr $diff($run1,[expr $camcol+1]) - \
		$diff($run2,$camcol)]]
	   }
	puts stdout "$camcol-[expr $camcol+1]: $res"
	}
   return
   }

#####################################################################
#Look at trends in zero point alone.

proc zpComp {} {
   global median photom slistres gres zpbal za zres

   if {![info exists photom]} return

#Compute mean in each CCD - look for trends (i.e. potential flaws) in
#balance factors.
   foreach camcol "1 2 3 4 5 6" {
	foreach filter "u g r i z" {
	   set sum($camcol,$filter) 0
	   set reslist($camcol,$filter) ""
	   }
	}

   foreach run [lsort -integer $photom()] {
	foreach filter "u g r i z" {
	   set aavg ""
	   foreach camcol "1 2 3 4 5 6" {
		set alist($camcol) $photom($run,$camcol,$filter)
		set aavg [expr $aavg + $alist($camcol)]
		}
	   set aavg [expr $aavg/6.]
	   foreach camcol "1 2 3 4 5 6" {
		set alist($run,$camcol,$filter) [expr $alist($camcol) - $aavg]
		lappend greslist($camcol,$filter) $alist($run,$camcol,$filter)
		}

	   }
	}

#Balance factors
   foreach camcol "1 2 3 4 5 6" {
	foreach filter "u g r i z" {
	   set zpbal($camcol,$filter) [lindex [lsort -real \
		$greslist($camcol,$filter)] \
		[expr [llength $greslist($camcol,$filter)]/2]]
	   }
	}

#List residuals
   set zres() ""
   foreach run [lsort -integer $photom()] {
	lappend zres() $run
	set format1 "%-4s  %6s%8s%8s%8s%8s%8s"
	puts stdout [format $format1 $run 1 2 3 4 5 6]
	set format "%4s  %8.2f%8.2f%8.2f%8.2f%8.2f%8.2f"
	foreach filter "u g r i z" {
	   foreach camcol "1 2 3 4 5 6" {
		set zres($run,$camcol,$filter) [expr \
		   $alist($run,$camcol,$filter) - $zpbal($camcol,$filter)]

#To put on equal footing with sky residuals and cross-scan residuals,
#multiply by -1
		set zres($run,$camcol,$filter) [expr \
		   -1*$zres($run,$camcol,$filter)]
		}

	   puts stdout [format $format $filter \
		$zres($run,1,$filter) $zres($run,2,$filter) \
		$zres($run,3,$filter) $zres($run,4,$filter) \
		$zres($run,5,$filter) $zres($run,6,$filter)]
	   }
	}
   return
   }

##################################################################
#Write out zero point residuals

proc zpWrite {} {
   global zres

   set fid [open zres.tcl w]
   puts $fid "global zres"
   puts $fid {if {[info exists zres]} {unset zres}}
   puts $fid "set zres() \"$zres()\""
   foreach run $zres() {
       foreach filter "u g r i z" {
	   foreach camcol "1 2 3 4 5 6" {
		puts $fid "set zres($run,$camcol,$filter) [format %.3f \
		$zres($run,$camcol,$filter)]"
		}
	   }
	}
   close $fid
   return
   }
