#Extract seeing statistics for a run
###Top level command:
### seeingRun run - combines seeingStat, seeingJudge, seeingJudgeDetail,
###                 seeingJudgePlot.
### Other useful commands:
### ppdRead run - reads seeing from astroline files.
### seeingStat run - computes median seeing for each CCD in 10 field
###	stretches
### seeingJudge run - computes median of medians from seeingStat combining
###	5 groups, 6 camcols, 5 filters.  Applied Kolmogorov correction.
###	Also computes worst median of all groups, camcols, filters.
###	Assigns quality to each group of fields based on these 2 stats.
### seeingJudgeDetail run - Expands results of seeingJudge and assigns good/bad
###	quality to each field in a run
### seeingJudgePlot run - makes a plot of median per field, min per field,
###	and max per field.  Note that these values are not quite the same
###	as were used by seeingJudge (they are computed field by field, not
###	grouping 10 fields together).
### seeingJudgePlotGif run - make a gif
### seeingPlot run camcol filter - plot seeing for specified CCD

#For APO, read astroline files and decode seeing information
proc ppdRead {run} {
   global seeing ppdir
   set ppdir /data/dp3.b/data/ppd

   if {![info exists seeing()]} {set seeing() ""}
   if {[lsearch $seeing() $run] < 0} {lappend seeing() $run}
   set seeing($run) ""

#Look for ppd directory
   if {![info exists ppdir]} {
	if {[file exists /data/dp3.b/data/ppd]} {
	   set ppdir /data/dp3.b/data/ppd
	} else {
	   set ppdir .
	   }
	}

#On sdsshost, we can also search the /astrolog directory
   if {[file exists /astrolog]} {
	set astrodir /astrolog
   } else {
	set astrodir .
	}

   set files [glob -nocomplain $astrodir/*/idAstroline-c1-[format %06d \
	$run].par $ppdir/*/idAstroline-c1-[format %06d $run].par]
   if {$files != ""} {
	set dir [file dirname [lindex $files 0]]
   } else {
	set dir .
	}
   foreach camcol "1 2 3 4 5 6" {
	lappend seeing($run) $camcol
	set seeing($run,$camcol) ""
	set file $dir/idAstroline-c$camcol-[format %06d $run].par
	set fid [open $file]
	set line ""

	while {1} {
	   set nextline [gets $fid]
	   if {[eof $fid]} break
	   set end [expr [clength $nextline]-1]
	   if {[cindex $nextline $end] == "\\"} {
		regsub {\\$} $nextline "" nextline
		append line " $nextline"
		continue
		}
	   if {![regexp _ASTROLINE $line]} {
		set line ""
		continue
		}
	   set nline 0
	   set frame [lindex $line 3]
	   lappend seeing($run,$camcol) $frame
	   foreach ifil "1 2 3 4 5" {
		set filter [lindex "x r i u z g" $ifil]
		set index [expr $ifil+71]
		set major [lindex $line $index]
		set index [expr $ifil+78]
		set minor [lindex $line $index]

#FWHM is avg of major & minor axes; convert sigma in pixels to FWHM in arcsec
#Do NOT convert to PSP psf_width scale.  We do NOT apply kolmogorov
#corrections here - that is done in seeingJudge and seeingJudgePlot.
		set width [expr ($major + $minor)*.94/2.]

#Don't bother with field offsets
		set field $frame
		set seeing($run,$camcol,$field,$filter) [format %.2f $width]
		}
	   set line ""
	   }
	close $fid
	}
   return
   }

#####################################################################
set nsmooth 10

#For APO seeing, we use 1.6 arcsec (actually should be 1.56)
set medlim 1.6
set maxlim 2.0
set gradlim .2

set iscale(u) .89
set iscale(g) .95
set iscale(r) 1
set iscale(i) 1.04
set iscale(z) 1.04

proc seeingStat {run} {
   global median seeing nsmooth
   if {![info exists median()]} {set median() ""}
   if {[lsearch $median() $run] < 0} {lappend median() $run}
   set median($run) ""
   foreach camcol "1 2 3 4 5 6" {
	lappend median($run) $camcol
	set median($run,$camcol) ""
	set fields $seeing($run,$camcol)
	set firstfield [lindex $fields 0]
	set lastfield [lindex $fields end]
	if {$firstfield == "" || $lastfield == ""} {
	   echo No fields!
	   return
	   }
	foreach filter "u g r i z" {
	   loop i $firstfield [expr $lastfield+1] $nsmooth { 
		set list ""
		set jlist ""
		set gradlist ""
		loop j $i [min [expr $i+$nsmooth] [expr $lastfield+1]] {
		   if {![info exists seeing($run,$camcol,$j,$filter)]} continue
		   lappend list $seeing($run,$camcol,$j,$filter)
		   lappend jlist $j
		   }
		set list [lsort -real $list]
		set jlist [lsort -integer $jlist]

#We tabulate median at median frame # of a string of frames.
		set jmed [lindex $jlist [expr \
		   [llength $jlist]/2]]
		if {$filter == "u"} {
		   lappend median($run,$camcol) $jmed
		   }
		set median($run,$camcol,$jmed,$filter) [lindex $list [expr \
		   [llength $list]/2]]
		}
	   }

	}
   }

#####################################################################
#Determine which segments of data are bad.  Must run seeingStat first.
#Go through medians, group by 5 fields of 10.  Take median of medians.
#Demand that median of medians not exceed 1.7 arcsec, and that no
#single point exceed 2.0 arcsec.  Otherwise, tag this group as bad.

proc seeingJudge {run} {
   global median quality iscale
   global medlim maxlim gradlim

   set camcol1 [lindex $median($run) 0]
   if {![info exists quality()]} {set quality() ""}
   if {[lsearch $quality() $run] < 0} {lappend quality() $run}
   set quality($run) ""
   loop i 0 [llength $median($run,$camcol1)] {
	set medlist ""
	set max 0
	set maxgrad 0
	set g1index [max 0 [expr $i-2]]
	set g2index [min [expr [llength $median($run,$camcol1)]-1] \
		[expr $i+2]]
	set group [lindex $median($run,$camcol1) $i]
	lappend quality($run) $group
	foreach camcol $median($run) {
	   loop j $g1index $g2index {
		set g [lindex $median($run,$camcol1) $j]
		foreach filter "u g r i z" {
		   if {![info exists median($run,$camcol,$g,$filter)]} {
			echo Missing median for run $run, camcol $camcol, \
			   group $g, filter $filter
			continue
			}
		   set medscale [format %.4f [expr $iscale($filter) * \
			$median($run,$camcol,$g,$filter)]]
		   lappend medlist $medscale
		   set max [max $max $medscale]
		   }
		}
	   }
	set medlist [lsort -real $medlist]
	set med [lindex $medlist [expr [llength $medlist]/2]]
	set quality($run,$group,med) $med
	set quality($run,$group,max) $max
	if {$med <= $medlim && $max <= $maxlim} {
	   set quality($run,$group) good
	} elseif {$max <= $maxlim} {
#	   echo Marginal run $run group $group, median = $med, max = $max
	   set quality($run,$group) acceptable
	} else {
#	   echo Reject run $run group $group, median = $med, max = $max
	   set quality($run,$group) bad
	   }
	}
   return
   }

#####################################################################
#After running seeingRead, seeingStat, seeingJudge, run this proc to
#assign a quality to each field.

proc seeingJudgeDetail {run} {
   global seeing quality iscale

   if {![info exists seeing($run)]} {
	echo No data for run $run
	return
	}
   if {[llength $seeing($run)] < 6} {
	echo Less than 6 camcols
	return
	}

#Set up quality
   set fields ""
   foreach group $quality($run) {
	set f1 [expr $group-5]
	set f2 [expr $group+5]
	for {set field $f1} {$field <= $f2} {incr field} {
	   set quality($run,$field) $quality($run,$group)
	   lappend fields $field
	   }
	}

#Get better list of fields
   global seeing
   set quality($run) $seeing($run,3)
   return
   }

################################################################3
#Must have run seeingJudgeDetail first.

proc seeingJudgePlot {run} {
   global seeing quality iscale medlim
#Plot
   if {[pgQinf STATE] == "CLOSED"} {
	pgBegin; pgGeomSet 1200 400 10 25; update; pgAsk 0
	}

   pgSci 1
   pgEnv 0 700 0 2 0 0
   pgLabel Field "Median Seeing" "Run $run"
   pgSci 2
   pgLine "0 1000" "$medlim $medlim"

   foreach field $seeing($run,3) {
	set seeingList ""
	if {![info exists quality($run,$field)]} continue
	if {![info exists seeing($run,1,$field,r)]} continue
	foreach camcol "1 2 3 4 5 6" {
	   if {[lsearch $seeing($run,$camcol) $field] < 0} {
		echo No data for camcol $camcol, field $field
		continue
		}
	   foreach filter "u g r i z" {
		set seeScale [expr $seeing($run,$camcol,$field,$filter) \
		   * $iscale($filter)]
		lappend seeingList $seeScale
		}
	   }
	set seeingList [lsort -real $seeingList]
	set median [lindex $seeingList [expr [llength $seeingList]/2]]
	set low [lindex $seeingList 0]
	set high [lindex $seeingList end]
	if {$quality($run,$field) == "good"} {
	   pgSci 5
	} elseif {$quality($run,$field) == "acceptable"} {
	   pgSci 6
	} else {
	   pgSci 7
	   }
	pgPoint $field $median 3
	pgLine "$field $field" "$low $high"
	}
   return
   }

################################################################3
#Must have run seeingJudgeDetail first.

proc seeingJudgePlotGif {run} {
   global seeing quality scale
#Plot
   if {[pgQinf STATE] == "OPEN"} {pgEnd}
   pgBegin seeing-$run.gif/GIF
   pgGeomSet 1200 400 10 25
   pgAsk 1
   seeingJudgePlot $run
   return
   }

################################################################3
#Proc to run seeingStat, seeingJudge, seeingJudgeDetail
proc seeingRun {args} {
   set opts { \
        {seeingRun "Compute seeing for a run vs. survey spec."} \
        {{[run]} INTEGER -1 run "Imaging run"} \
    }
if {[shTclParseArg $args $opts biasCheck] == 0} {
        return 
}
   echo Reading data ...
   ppdRead $run
   echo Analyzing ...
   seeingStat $run
   seeingJudge $run
   seeingJudgeDetail $run
   seeingJudgePlotGif $run
   return
   }

#####################################################################
#Plot a single camcol & filter.
proc seeingPlot {run camcols filters} {
   global seeing median
   pgSlw 3
   pgSci 1
   pgEnv 0 [lindex $seeing($run,[lindex $camcols 0]) end] 1 2 0 0
   pgLabel Field FWHM "Run $run, camcol $camcols, filter $filters"
   foreach camcol $camcols {
	foreach filter $filters {
	   foreach field $seeing($run,$camcol) {
		pgSci 2
		pgSch .7
		pgPoint $field $seeing($run,$camcol,$field,$filter) 2
		pgSch 2
		if {![info exists median($run,$camcol,$field,$filter)]} \
			continue
		if {$median($run,$camcol,$field,$filter) == ""} continue
		pgSci 1
		pgPoint $field $median($run,$camcol,$field,$filter) 3
		pgSch 1
		}
	   pgSch 1
	   }
	}
   return
   }


