#include <math.h>
#include "dervish.h"
#include "imgAxes.h"
#include "imgProto.h"
#include "img_msg_c.h"

/* #define PI 3.14159265358979 */
#define sqr(x) (sqrt(fabs(x)))


/*============================================================================  
**============================================================================
**
** ROUTINE: box
**
** DESCRIPTION:
** Convert centroid and width to min and max with bounds checking.
** My limits are inclusive for both lower and upper limits.
**
** RETURN VALUES:
**	None
**
**============================================================================
*/

static void box(REGION *reg, int row0, int col0, int npr, int npc,
	int *row1, int *row2, int *col1, int *col2) {
   *row1 = row0 - npr;
   if (*row1 < 1) *row1 = 1;
   *row2 = row0 + npr;
   if (*row2 > reg->nrow-2) *row2 = reg->nrow - 2;
   *col1 = col0 - npc;
   if (*col1 < 1) *col1 = 1;
   *col2 = col0 + npc;
   if (*col2 > reg->ncol-2) *col2 = reg->ncol - 2;
   }


/*============================================================================  
**============================================================================
**
** ROUTINE: max
**
**============================================================================
*/
static double max (double x, double y) {if (x >= y) return x; else return y;}


/*============================================================================  
**============================================================================
**
** ROUTINE: imgAxes
**
** DESCRIPTION:
**	Compute centroid in a region given initial guess.
**
** RETURN VALUES:
**	IMG_SUCCESS	Success
**
**============================================================================
*/
IMG_CODE imgAxes (REGION *reg, int row0, int col0, int npr, int npc,
	AXES *axes) {

   double threshold;
   double sum, wsumr, wsumc; 	 		/*hold weighted sum */
   int row1, row2, col1, col2;
/* Local indexes */
   int i, j;
/* Temporary storage */
   double pix, pixl, pixr, pixd, pixu, pixld, pixrd, pixlu, pixru;

   /* local temps  for 2nd moments */
   double qmomrr=0.0; double qmomrc=0.0; double qmomcc=0.0; 	 
   double eigen1,eigen2;	/*real-valued eigenvalues for axes */ 	 
   double temp,rmaj,cmaj,rminor,cminor; 
   double cs, sn;
   double *buf;
   double tmp;
   int nbuf, flag;

   float centroidr,centroidc;		/*real-valued coordinate of
					centriod*/
   float eccentricity, ellipticity;	/*real-valued eccentricity for
					ellipse*/
   float angminaxis, angmajaxis;
   float sigminaxis, sigmajaxis;   
   double rx, ry, ax, bx, cx, ay, by, cy, d, sumx, sumxx, sumy, sumyy;
   double axy, sumxy;

/* Pointer to function to return integer pixel value */
/*   int (*intpix)(REGION *, int, int) = regPixIntFunc (reg); */

/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg);


/* Get array limits with bounds checking; */

   box(reg, row0, col0, npr, npc, &row1, &row2, &col1, &col2);

/* Threshold is peak along boundary */
/* No more! Take median.  I get too many bogus high points */

   buf = (double *)shMalloc((2*(row2-row1+1) + 2*(col2-col1+1)) * 
	sizeof(double));
   nbuf=0;
   for (i = row1; i <= row2; i++) {
	buf[nbuf] = dblpix(reg,i,col1);
	nbuf++;
	buf[nbuf] = shRegPixGetAsDbl(reg,i,col2);
	nbuf++;
/*
	threshold = max (threshold, dblpix (reg, i, col1));
	threshold = max (threshold, dblpix (reg, i, col2));
*/
	}
   for (j = col1; j <= col2; j++) {
	buf[nbuf] = dblpix(reg,row1,j);
	nbuf++;
	buf[nbuf] = dblpix(reg,row2,j);
	nbuf++;
/*
	threshold = max (threshold, dblpix (reg, row1, j));
	threshold = max (threshold, dblpix (reg, row2, j));
*/
	}

/* The dreaded bubble sort! */
   flag = 1;
   while (flag) {
	flag = 0;
	for (i=1; i<nbuf; i++) {
	   if (buf[i-1] > buf[i]) {
		tmp = buf[i-1];
		buf[i-1] = buf[i];
		buf[i] = tmp;
		flag = 1;
		}
	   }
	}
   threshold = buf[nbuf/2];
   shFree(buf);

/* find the centroid by subtracting threshold from all points
within the * box greater than threshold, and summing in the x and y
directions */ 

   sum=0; wsumr=0; wsumc=0;
   for (i = row1; i <= row2; i++) {
      for (j = col1; j <= col2; j++) {
        if ((pix = dblpix (reg, i, j)) > threshold) { 	 
	   pix = pix-threshold; 	 
           sum += pix; 	 

/* The following complex math fits parabolas to intensity in x and y
 * directions independently and then computes 2nd moments accordingly.
 * This seems to give much more accurate 2nd moments than my old box-car
 * integration routines even for undersampled images.  The cross-terms
 * are slightly less accurate since I don't yet incorporate cross-derivs.
*/
	   rx = -i-.5;
	   ry = -j-.5;

	   pixl = dblpix (reg, i-1,j)-threshold;
	   pixr = dblpix (reg, i+1,j)-threshold;
	   pixd = dblpix (reg, i,j-1)-threshold;
	   pixu = dblpix (reg, i,j+1)-threshold;
	   bx = (pixr - pixl)/2.;
	   cx = (pixr + pixl - 2.*pix)/2.;
	   ax = pix-(1./12.)*cx;
	   ax = ax/pix;
	   bx = bx/pix;
	   cx = cx/pix;
	   by = (pixu - pixd)/2.;
	   cy = (pixu + pixd - 2.*pix)/2.;
	   ay = pix-(1./12.)*cy;
	   ay = ay/pix;
	   by = by/pix;
	   cy = cy/pix;
	   sumx = (-ax*rx + (bx-cx*rx)/12.);
	   sumy = (-ay*ry + (by-cy*ry)/12.);
/*
           wsumr += pix*(i+0.5);
           wsumc += pix*(j+0.5); 	 
*/
	   wsumr += pix*sumx;
	   wsumc += pix*sumy;
 	   }
         }     
      }
   if (sum == 0) {return IMG_GENERIC_ERROR;} /* trouble */ 
   centroidr = wsumr/sum;
   centroidc = wsumc/sum;

/* quadupole moment about centroidr and centroidc are used to get eigen values,
* and then the major and minor axes for the ellipse.
*/

/* Try more accurate centroid code */

   for (i = row1; i <= row2; i++) {
      for (j = col1; j <= col2; j++) {
        if ((pix = dblpix (reg, i, j)) > threshold) { 	 
	   pix = pix-threshold; 	 
	   pixl = dblpix (reg, i-1,j)-threshold;
	   pixr = dblpix (reg, i+1,j)-threshold;
	   pixd = dblpix (reg, i,j-1)-threshold;
	   pixu = dblpix (reg, i,j+1)-threshold;
	   pixld = dblpix (reg, i-1,j-1)-threshold;
	   pixrd = dblpix (reg, i+1,j-1)-threshold;
	   pixlu = dblpix (reg, i-1,j+1)-threshold;
	   pixru = dblpix (reg, i+1,j+1)-threshold;
	   bx = (pixr - pixl)/2.;
	   cx = (pixr + pixl - 2.*pix)/2.;
	   ax = pix-(1./12.)*cx;
	   ax = ax/pix;
	   bx = bx/pix;
	   cx = cx/pix;
	   by = (pixu - pixd)/2.;
	   cy = (pixu + pixd - 2.*pix)/2.;
	   ay = pix-(1./12.)*cy;
	   ay = ay/pix;
	   by = by/pix;
	   cy = cy/pix;
/*
	   qmomrr += (i+0.5-centroidr)*(i+0.5-centroidr)*pix; 		 
	   qmomrc += (i+0.5-centroidr)*(j+0.5-centroidc)*pix; 		 
	   qmomcc += (j+0.5-centroidc)*(j+0.5-centroidc)*pix; 	
*/
	   rx = centroidr - i - .5;
	   ry = centroidc - j - .5;

	   sumxx = (ax*rx*rx + (cx*rx*rx-2.*bx*rx+ax)/12.+cx/80.);
	   sumyy = (ay*ry*ry + (cy*ry*ry-2.*by*ry+ay)/12.+cy/80.);

/* xy term needs a little more work.
*/
	   axy = (ax + ay - 1.);
	   d = (pixru + pixld - pixlu - pixrd)/4.;
	   d = d/pix;
	   sumxy = axy*rx*ry + (rx*ry*(cx+cy) - bx*ry - by*rx)/12. + d/144.;
	   qmomrr += sumxx*pix;
	   qmomrc += sumxy*pix;
	   qmomcc += sumyy*pix;
 	   }
         }
      }     
   qmomrr = qmomrr /sum;
   qmomrc = qmomrc/sum;
   qmomcc = qmomcc/sum;

   temp = sqr((qmomrr-qmomcc)*(qmomrr-qmomcc) + 4*(qmomrc*qmomrc));
   eigen1 = ((qmomrr+qmomcc)+temp)/2.0;
   eigen2 = ((qmomrr+qmomcc)-temp)/2.0; 
   if (eigen1 == 0.) eigen1 = 1.;
   eccentricity = sqr(1.0-(eigen2/eigen1));
   if(qmomrc == 0.0) {
      if (qmomrr >= qmomcc)  {
	rmaj = 1.0;
	cmaj = 0.0;
	rminor=0.0; cminor = 1.0;
	} else {
	rmaj = 0.0; 
	cmaj = 1.0;
	rminor=1.0;
	cminor = 0.0;
	}
      } else {
 /* qmomrc != 0.0 */ 
      rmaj = 1.0;
      cmaj = (eigen1-qmomrr)/(qmomrc);
      cminor = (eigen2-((qmomrr)/(qmomrc))); 
      rminor = 1.0; 
      } 
   angmajaxis = atan2(cmaj,rmaj) * 180.0/PI;
   angminaxis = atan2(cminor,rminor) * 180.0/PI;
   ellipticity = 1.0 - sqr (1.0- (eccentricity)*(eccentricity));
	   /* compute sigma across major and minor axes */ 	
   cs = cos(angmajaxis * PI/180.0);
   sn = sin(angmajaxis * PI/180.0);
   sigmajaxis  = sqr ((qmomrr)*cs*cs + 2.0*(qmomrc)*cs*sn + (qmomcc)*sn*sn); 
   sigminaxis  = sqr ((qmomrr)*sn*sn - 2.0*(qmomrc)*cs*sn + (qmomcc)*cs*cs);
/* Fill output array */
   axes->row = centroidr;
   axes->col = centroidc;
   axes->pos_maj = angmajaxis;
   axes->eps = ellipticity;
   axes->cnts = sum;
   axes->sigmaj = sigmajaxis;
   axes->sigmin = sigminaxis;
   axes->rmax = max((double)npr,(double)npc);
   return IMG_SUCCESS;
}
