#include <math.h>
#include "dervish.h"
#include "imgAxes.h"
#include "imgProto.h"
#include "img_msg_c.h"

/* #define PI 3.14159265358979 */
#define sqr(x) (sqrt(fabs(x)))


/*============================================================================  
**============================================================================
**
** ROUTINE: box
**
** DESCRIPTION:
** Convert centroid and width to min and max with bounds checking.
**
** RETURN VALUES:
**	None
**
**============================================================================
*/

static void box(REGION *reg, int row0, int col0, int npr, int npc,
	int *row1, int *row2, int *col1, int *col2) {
   *row1 = row0 - npr;
   if (*row1 < 0) *row1 = 0;
   *row2 = row0 + npr;
   if (*row2 >= reg->nrow) *row2 = reg->nrow - 1;
   *col1 = col0 - npc;
   if (*col1 < 0) *col1 = 0;
   *col2 = col0 + npc;
   if (*col2 >= reg->ncol) *col2 = reg->ncol - 1;
   }


/*============================================================================  
**============================================================================
**
** ROUTINE: max
**
**============================================================================
*/
static double max (double x, double y) {if (x >= y) return x; else return y;}


/*============================================================================  
**============================================================================
**
** ROUTINE: imgCaxes
**
** DESCRIPTION:
**	Compute centroid in a region given initial guess.  Weighting of
**	points is 1 or 0, not by pixel intensity
**
** RETURN VALUES:
**	IMG_SUCCESS	Success
**
**============================================================================
*/
IMG_CODE imgCaxes (REGION *reg, int row0, int col0, int npr, int npc,
	AXES *axes) {

   double threshold;
   double sum, wsumr, wsumc; 	 		/*hold weighted sum */
   int row1, row2, col1, col2;
/* Local indexes */
   int i, j;
/* Temporary storage */
   double pix;

   /* local temps  for 2nd moments */
   double qmomrr=0.0; double qmomrc=0.0; double qmomcc=0.0; 	 
   double eigen1,eigen2;	/*real-valued eigenvalues for axes */ 	 
   double temp,rmaj,cmaj,rminor,cminor; 
   double cs, sn;

   float centroidr,centroidc;		/*real-valued coordinate of
					centriod*/
   float eccentricity, ellipticity;	/*real-valued eccentricity for
					ellipse*/
   float angminaxis, angmajaxis;
   float sigminaxis, sigmajaxis;   

/* Get array limits with bounds checking; */

   box(reg, row0, col0, npr, npc, &row1, &row2, &col1, &col2);

/* Threshold is peak along boundary */

   threshold = -1.e20;
   for (i = row1; i <= row2; i++) {
	threshold = max (threshold, shRegPixGetAsDbl (reg, i, col1));
	threshold = max (threshold, shRegPixGetAsDbl (reg, i, col2));
	}
   for (j = col1; j <= col2; j++) {
	threshold = max (threshold, shRegPixGetAsDbl (reg, row1, j));
	threshold = max (threshold, shRegPixGetAsDbl (reg, row2, j));
	}

/* find the centroid by subtracting threshold from all points
within the * box greater than threshold, and summing in the x and y
directions */ 

   sum=0; wsumr=0; wsumc=0;
   for (i = row1; i <= row2; i++) {
      for (j = col1; j <= col2; j++) {
        if ((pix = shRegPixGetAsDbl (reg, i, j)) > threshold) { 	 
           sum += 1.; 	 
           wsumr += (i+0.5);
           wsumc += (j+0.5); 	 
 	   }		 	 
         }     
      }
   if (sum == 0) {return IMG_GENERIC_ERROR;} /* trouble */ 
   centroidr = wsumr/sum;
   centroidc = wsumc/sum;

/* quadupole moment about centroidr and centroidc are used to get
eigen values, * and then the major and minor axes for the ellipse.  */

   for (i = row1; i <= row2; i++) {
      for (j = col1; j <= col2; j++) {
        if ((pix = shRegPixGetAsDbl (reg, i, j)) > threshold) { 	 
	   qmomrr += (i+0.5-centroidr)*(i+0.5-centroidr);
	   qmomrc += (i+0.5-centroidr)*(j+0.5-centroidc);
	   qmomcc += (j+0.5-centroidc)*(j+0.5-centroidc);
 	   }		 	 
         }
      }     
   qmomrr = qmomrr /sum;
   qmomrc = qmomrc/sum;
   qmomcc = qmomcc/sum;

   temp = sqr((qmomrr-qmomcc)*(qmomrr-qmomcc) + 4*(qmomrc*qmomrc));
   eigen1 = ((qmomrr+qmomcc)+temp)/2.0;
   eigen2 = ((qmomrr+qmomcc)-temp)/2.0; 
   if (eigen1 == 0.) eigen1 = 1.;
   eccentricity = sqr(1.0-(eigen2/eigen1));
   if(qmomrc == 0.0) {
      if (qmomrr >= qmomcc)  {
	rmaj = 1.0;
	cmaj = 0.0;
	rminor=0.0; cminor = 1.0;
	} else {
	rmaj = 0.0; 
	cmaj = 1.0;
	rminor=1.0;
	cminor = 0.0;
	}
      } else {
 /* qmomrc != 0.0 */ 
      rmaj = 1.0;
      cmaj = (eigen1-qmomrr)/(qmomrc);
      cminor = (eigen2-((qmomrr)/(qmomrc))); 
      rminor = 1.0; 
      } 
   angmajaxis = atan(cmaj/rmaj) * 180.0/PI;
   angminaxis = atan(cminor/rminor) * 180.0/PI;
   ellipticity = 1.0 - sqr (1.0- (eccentricity)*(eccentricity));
	   /* compute sigma across major and minor axes */ 	
   cs = cos(angmajaxis * PI/180.0);
   sn = sin(angmajaxis * PI/180.0);
   sigmajaxis  = sqr ((qmomrr)*cs*cs + 2.0*(qmomrc)*cs*sn + (qmomcc)*sn*sn); 
   sigminaxis  = sqr ((qmomrr)*sn*sn - 2.0*(qmomrc)*cs*sn + (qmomcc)*cs*cs);
/* Fill output array */
   axes->row = centroidr;
   axes->col = centroidc;
   axes->pos_maj = angmajaxis;
   axes->eps = ellipticity;
   axes->cnts = sum;
   axes->sigmaj = sigmajaxis;
   axes->sigmin = sigminaxis;
   axes->rmax = max((double)npr,(double)npc);
   return IMG_SUCCESS;
}
