/*****************************************************************************
******************************************************************************
**
** FILE:
**	imgDecimate.c
**
** ABSTRACT:
**	This file contains code to decimate and/or compress a region
**
**
******************************************************************************
******************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "imgProto.h"
#include "imgTypes.h"


/*============================================================================  
**============================================================================
**
** ROUTINE: imgDecimate
**
** DESCRIPTION:
**	Decimate a region
**
** RETURN VALUES:
**	IMG_SUCCESS -   Successful completion.
**	IMG_GENERIC_ERROR -   Failed miserably.
**
** CALLS TO:
**
**============================================================================
*/
/*
 */

IMG_CODE imgDecimate(
	REGION *reg,
	REGION *regout,
	int factor
	  )
{

   int i,j;
   int type = pxType(reg);

   int nr = reg->nrow;
   int nc = reg->ncol;

/* Pointer to function to return integer pixel value */
   int (*intpix)(REGION *, int, int) = regPixIntFunc (reg);
/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg);
/* Pointer to a function to set an int pixel value */
   void (*intpixset)(REGION *, int, int, int) =
        regPixIntFuncSet (reg);
/* Pointer to a function to set a double pixel value */
   void (*dblpixset)(REGION *, int, int, double) =
        regPixDblFuncSet (reg);

/* For speed, process ints and floats separately.  We assume without checking
 * that the input and output regions have the same type.
*/

   if (type == IMG_INT) {
	for (j=0; j<=nc/factor; j++) {
	   for (i=0; i<nr/factor; i++) {
		intpixset(regout, i, j, intpix(reg, i*factor, j*factor));
		}
	   }

   } else {
	for (j=0; j<=nc/factor; j++) {
	   for (i=0; i<nr/factor; i++) {
		dblpixset(regout, i, j, dblpix(reg, i*factor, j*factor));
		}
	   }
	}
   return IMG_SUCCESS;

}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgCompress
**
** DESCRIPTION:
**	Compress a region
**
** RETURN VALUES:
**	IMG_SUCCESS -   Successful completion.
**	IMG_GENERIC_ERROR -   Failed miserably.
**
** CALLS TO:
**
**============================================================================
*/
/*
 */

IMG_CODE imgCompress(
	REGION *reg,
	REGION *regout,
	int factor
	  )
{

   int i, j, k, l;
   int isum, n;
   double sum;

   int type = pxType(reg);

   int nr = reg->nrow;
   int nc = reg->ncol;

/* Pointer to function to return integer pixel value */
   int (*intpix)(REGION *, int, int) = regPixIntFunc (reg);
/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg);
/* Pointer to a function to set an int pixel value */
   void (*intpixset)(REGION *, int, int, int) =
        regPixIntFuncSet (reg);
/* Pointer to a function to set a double pixel value */
   void (*dblpixset)(REGION *, int, int, double) =
        regPixDblFuncSet (reg);

/* For speed, process ints and floats separately.  We assume without checking
 * that the input and output regions have the same type.
*/

   if (type == IMG_INT) {
	for (j=0; j<nc/factor; j++) {
	   for (i=0; i<nr/factor; i++) {
		isum=0;
		n=0;
		for (l=j*factor; l<(j+1)*factor; l++) {
		   if (l >= nc) continue;
		   for (k=i*factor; k<(i+1)*factor; k++) {
			if (k >= nr) continue;
			isum += intpix(reg, k, l);
			n++;
			}
		   }
		if (n == 0) {
		   isum=0;
		   }
		else {
		   isum = ((double)isum)/n + 0.5;
		   }
		intpixset(regout, i, j, isum);
		}
	   }

   } else {
	for (j=0; j<nc/factor; j++) {
	   for (i=0; i<nr/factor; i++) {
		sum=0.;
		n=0;
		for (l=j*factor; l<(j+1)*factor; l++) {
		   if (l >= nc) continue;
		   for (k=i*factor; k<(i+1)*factor; k++) {
			if (k >= nr) continue;
			sum += dblpix(reg, k, l);
			n++;
			}
		   }
		if (n == 0) {
		   sum=0.;
		   }
		else {
		   sum = sum/n;
		   }
		dblpixset(regout, i, j, sum);
		}
	   }
	}
   return IMG_SUCCESS;
}


