/*****************************************************************************
******************************************************************************
**
** FILE:
**	imgFuncFit.c
**
** ABSTRACT:
**	This file contains a routine to fit a function to a profile.
**	The functions are hardwired at present.
**
**
** ENTRY POINT		SCOPE	DESCRIPTION
** -------------------------------------------------------------------------
** imgFuncFit		public	Do the fit
**
** ENVIRONMENT:
**	ANSI C.
**
** REQUIRED PRODUCTS:
**	FTCL	 	TCL + Fermi extension
**      DERVISH
**	IMAGE
**
** AUTHORS:
**	Creation date:  Feb 95
**	Steve Kent
**
******************************************************************************
******************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "dervish.h"
#include "imgProto.h"

/*----------------------------------------------------------------------------
**
** GLOBAL VARIABLES
*/

/*----------------------------------------------------------------------------
**
** STATIC FUNCTIONS
**	All hyperbolics
*/

static double sinh(double x) {return (exp(x) - exp(-x))/2.;}
static double cosh(double x) {return (exp(x) + exp(-x))/2.;}
static double tanh(double x) {return sinh(x)/cosh(x);}
static double sech(double x) {return 1./cosh(x);}
static double sech2(double x) {return pow(sech(x),2);}
static double sech2deriv(double x) {return (-2.*tanh(x)*pow(sech(x),2.));}

static double (*func)(double x) = sech2;
static double (*funcderiv)(double x) = sech2deriv;

/*============================================================================  
**============================================================================
**
** ROUTINE: imgFuncFit
**
** DESCRIPTION:
**
**============================================================================
*/
/*
 * Input is chain of PROF's
 * return is 2 numbers: Intensity and FWHM
 */

IMG_CODE
imgFuncFit(CHAIN *chain,	/* IN: Chain of PROF's */
	  double *amp,		/* MODIFIED: Amplitude */
	  double *fwhm,		/* MODIFIED: Full width at half max */
	  double *fract)	/* IN: Fractional update */
{
   PROF *prof, *profmax;
   double z0;
   double cmat[3];
   double rhs[2];
   double prt[2];
   int size = shChainSize(chain);
   int i,j,l;
   CURSOR_T cursor;
   double obs, th, res;
   double duma[2], dumb[2];
   int idumc[2];
   double weight;

   if (size < 3) return IMG_GENERIC_ERROR;

/* 2.49 is better than 2.35 for FWHM of a sech^2 */
   z0 = (*fwhm) * sqrt(2.) / 2.49;
   if (z0 <= 0.) return IMG_GENERIC_ERROR;
   l=0;
   for (i=0; i<2; i++) {
	rhs[i]=0.;
	for (j=0; j<=i; j++) {
	   cmat[l]=0;
	   l++;
	   }
	}
   cursor = shChainCursorNew(chain);
   profmax = shChainElementGetByPos(chain, size-1);
   while (1) {
	prof = shChainWalk(chain, cursor, NEXT);
	if (prof == NULL) break;
	obs = prof->cnts - profmax->cnts;
	prt[0] = func(prof->rad/z0) - func(profmax->rad/z0);
	prt[1] = (*amp) * (funcderiv(prof->rad/z0)*(-prof->rad/pow(z0,2.))
	   - funcderiv(profmax->rad/z0)*(-profmax->rad/pow(z0,2.)));
	th = prt[0] * (*amp);
	res = obs-th;
	weight=pow(prof->rad+1.,2);
	l=0;
	for (i=0; i<2; i++) {
	   rhs[i]+= res*prt[i] * weight;
	   for (j=0; j<=i; j++) {
		cmat[l] += prt[i]*prt[j]*weight;
		l++;
		}
	   }
	}
   shChainCursorDel (chain, cursor);
   syminv(cmat, rhs, 2, duma, dumb, idumc);
   *amp += rhs[0] * (*fract);
   z0   += rhs[1] * (*fract);
   *fwhm = z0*2.49/sqrt(2.);
   return IMG_SUCCESS;
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgFunc
**
** DESCRIPTION:
**	Just compute the function without fitting
**
**============================================================================
*/
/*
 * Input is chain of PROF's, amplitude, fwhm, and index into chain
 * return is intensity at radius specified by index.
 * I am not sure that this is the best implementation, but it will work for now.
 */

double 
imgFunc	 (CHAIN *chain,		/* IN: Chain of PROF's */
	  double *amp,		/* IN: Amplitude */
	  double *fwhm,		/* IN: Full width at half max */
	  int *index)		/* IN: Index into chain */
{
   PROF *prof, *profmax;
   double z0;
   int size = shChainSize(chain);
   double th;

   if (size < 3) return 0.;
   z0 = (*fwhm) * sqrt(2.) / 2.35;
   if (z0 <= 0.) return 0.;
   prof = shChainElementGetByPos(chain, (*index));
   if (prof == NULL) return 0.;
   profmax = shChainElementGetByPos(chain, size-1);
   th = (*amp) * (func(prof->rad/z0) - func(profmax->rad/z0)) + profmax->cnts;
   return th;
}
