/**********************************************************************
 * Fit a gaussian to an asteroid trail
 *********************************************************************/

#include <math.h>
#include "dervish.h"
#include "img_msg_c.h"
#include "imgProto.h"

#ifndef PI
#define PI 3.14159265358979
#endif

#define sqr(x) (sqrt(fabs(x)))

/*============================================================================  
**============================================================================
**
** ROUTINE: imgGauss
**
** DESCRIPTION:
**	Fit a Gaussian to a trail.  Just adjust amplitude and offset - presume
**	that center and orientation and width are found by centroid
**	or taken from seeing width.
**
** RETURN VALUES:
**	IMG_SUCCESS	Success
**
**============================================================================
*/

/* rowc, colc are center of trail */

IMG_CODE imgGauss      (REGION *reg,	/* Input region */
			double rowc,	/* Row center */
			double colc,	/* Col center */
			double ang,	/* Position angle */
			double FWHM,	/* Seeing FWHM */
			double rnoise,	/* Read noise, including sky noise */
			double gain,	/* Electrons per ADU */
			int xmax,	/* Number of pixels along trail */
			int ymax,	/* Number of pixels perp. to trail */
			double *amp,	/* OUT: Amplitude of gaussian */
			double *erramp,	/* OUT: Amplitude error */
			double *offset,	/* OUT: Offset */
			double *erroff,	/* OUT: Offset error */
			double *chi2)	/* OUT: goodness of fit */
{

/* Local variables */
   int i, j;
   double rang = PI*ang/180.;
   double sig = FWHM/2.35;
   double comp;
   double sum00 = 0;
   double sum01 = 0;
   double sum11 = 0;
   double rhs0 = 0;
   double rhs1 = 0;
   double inv00, inv01, inv11;
   double gfit = 0;
   double net;
   int n = 0;
   double obs, res, r, c, determ, weight;

/* Pointer to function to return double pixel value */
/*   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg); */

/* Loop through x, y in Gaussian frame */
   for (i = -xmax; i <= xmax; i++) {
	for (j = -ymax; j <= ymax; j++) {
	   r = i*cos(rang) - j*sin(rang) + rowc;
	   c = j*cos(rang) + i*sin(rang) + colc;
	   if (r < 0.5 || r > reg->nrow - 0.5) continue;
	   if (c < 0.5 || c > reg->ncol - 0.5) continue;
	   comp = exp(-j*j/(2.*sig));

/* For faint stuff, pixels are all weighted the same, so presume so for
 * the least squares */
	   weight = rnoise*rnoise;
	   obs = pixbin(reg, r, c);
	   res = obs-comp;
	   sum00 += 1./weight;
	   sum01 += comp/weight;
	   sum11 += comp*comp/weight;
	   rhs0 += res/weight;
	   rhs1 += res*comp/weight;
	   }
	}

/* Invert */
   determ = sum00*sum11 - sum01*sum01;
   if (determ == 0) {
	*amp = 0.;
	*erramp = 0.;
	*offset = 0.;
	*erroff = 0.;
	*chi2 = 0.;
	return IMG_SUCCESS;
	}
   inv00 = sum11/determ;
   inv01 = -sum01/determ;
   inv11 = sum00/determ;

/* Solve for parameters */
   *offset = inv00*rhs0 + inv01*rhs1;
   *amp = inv01*rhs0 + inv11*rhs1;

/* Estimated errors in parameters */
   *erroff = sqrt(inv00);
   *erramp = sqrt(inv11);

/* printf ("Offset = %.1f, amp = %.1f\n", offset, amp); */
/* Loop through x, y in Gaussian frame and compute goodness-of-fit */
   n=0;
   for (i = -xmax; i <= xmax; i++) {
	for (j = -ymax; j <= ymax; j++) {
	   r = i*cos(rang) - j*sin(rang) + rowc;
	   c = j*cos(rang) + i*sin(rang) + colc;
	   if (r < 0.5 || r > reg->nrow - 0.5) continue;
	   if (c < 0.5 || r > reg->ncol - 0.5) continue;
	   comp = (*amp)*exp(-j*j/(2.*sig)) + (*offset);
           if (comp-(*offset) > 0) {
             net = comp-(*offset);
           } else {
             net = 0;
             }

/* Statistical weight (variance) */
	   weight = rnoise*rnoise + net/gain;

/* For faint stuff, pixels are all weighted the same, so presume so for
 * the least squares */
	   obs = pixbin(reg, r, c);
/* printf ("r = %.1f c = %.1f obs = %.1f, comp = %.1f\n", r, c, obs, comp);
*/
	   res = obs-comp;
	   gfit += pow(res,2)/weight;
	   n++;
	   }
	}

   if (n == 0) {
	gfit = -1;
   } else {
     gfit = sqrt(gfit/n);
	}

   *chi2 = gfit;
   return IMG_SUCCESS;
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgBoxPhoto
**
** DESCRIPTION:
**	Do photometry of a trail.  Mimic the imgPhoto interface.
**
** RETURN VALUES:
**	IMG_SUCCESS	Success
**
**============================================================================
*/

/* rowc, colc are center of trail */

IMG_CODE imgBoxPhoto      (REGION *reg,	/* Input region */
			double rowc,	/* Row center */
			double colc,	/* Col center */
			double ang,	/* Position angle */
			int xmax,	/* Number of pixels along trail */
			int ymax,	/* Number of pixels perp. to trail */
			double *counts,	/* OUT: Total counts */
			double *area)	/* OUT: Area */
{

/* Local variables */
   int i, j;
   double r,c;
   double rang = PI*ang/180.;

/* Pointer to function to return double pixel value */
/*   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg); */

/* Loop through x, y in Gaussian frame */
   *counts = 0.;
   *area = 0.;
   for (i = -xmax; i <= xmax; i++) {
	for (j = -ymax; j <= ymax; j++) {
	   r = i*cos(rang) - j*sin(rang) + rowc;
	   c = j*cos(rang) + i*sin(rang) + colc;
	   if (r < 0.5 || r > reg->nrow - 0.5) continue;
	   if (c < 0.5 || c > reg->ncol - 0.5) continue;
	   *counts += pixbin(reg, r, c);
	   *area += 1;
	   }
	}
   return IMG_SUCCESS;
}

