/*****************************************************************************
******************************************************************************
**
** FILE:
**      imgHist.c
**
** ABSTRACT:
**      This file contains C routines to create and use histograms of
**	pixel values.
**
** ENTRY POINT          SCOPE   DESCRIPTION
** -------------------------------------------------------------------------
** imgStats		Public	Compute statics on a part of a region
** imgHist		Public	Compute a histogram of pixel values.
**
** ENVIRONMENT:
**      ANSI C.
**
** REQUIRED PRODUCTS:
**      dervish
**      profile
**
** AUTHORS:
**      Creation date:  May 93
**      Steve Kent
**
******************************************************************************
******************************************************************************
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "img_msg_c.h"
#include "imgProto.h"

/************************** imgTemplate ***********************************/

/* IMG_CODE imgTemplate (REGION *reg) { */

/* Pointer to function to return integer pixel value */
/*   int (*intpix)(REGION *, int, int) = regPixIntFunc (reg); */
/* Pointer to function to return double pixel value */
/* double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg); */

/* To access a pixel as a double: dblpix (reg, row, col) */
/* To access a pixel as an int:   intpix (reg, row, col) */

/*   return IMG_SUCCESS;
}
*/

/*Where applicable, the x dimension is the column number dimension, and
**the y dimension is the row number dimension; this accords with the new
**desired of JEG and is counter to the blue book */

/*******************Utility: double min function *************************/
static double dmin(double i, double j) {
   if (i < j) {return i;} else {return j;}
   }

/*******************Utility: double max function *************************/
static double dmax(double i, double j) {
   if (i > j) {return i;} else {return j;}
   }


/*******************Utility: int min function *************************/
static int imin (int i, int j) {
   if (i < j) {return i;} else {return j;}
   }

/*******************Utility: int max function *************************/
static int imax (int i, int j) {
   if (i > j) {return i;} else {return j;}
   }

/*******************Utility: integer abs function *************************/
static int iabs(int i) {
   if (i >= 0) {return i;} else {return -i;}
   }


/*******************Utility: double abs function *************************/
static double dabs(double i) {
   if (i >= 0) {return i;} else {return -i;}
   }


/****************************************************************************/
/********************** 	imgStats		*********************/
/****************************************************************************/

/* Given a region, compute image statistics.  For generality, a section
 * of a region is input
*/

IMG_CODE imgStats (
	REGION *reg, 	/* IN: Region */
	int row1,	/* IN: Starting row */
	int nrow,	/* IN: Number of rows */
	int col1,	/* IN: Starting column */
	int ncol,	/* IN: Number of columns */
	double *mean,	/* OUT: Mean */
	double *stdev,	/* OUT: 1 sigma */
	int *minval,	/* OUT: Minimum value */
	int *rowmin,	/* OUT: Row of pixel with minimum value */
	int *colmin,	/* OUT: Col of pixel with minimum value */
	int *maxval,	/* OUT: Maximum value */
	int *rowmax,	/* OUT: Row of pixel with maximum value */
	int *colmax)	/* OUT: Col of pixel with maximum value */
{

   int nr = reg->nrow;
   int nc = reg->ncol;

/* Make pix a double, not an int.  This avoids truncation on floating images */
   double pix;
   int npix;
   int i,j;
   double sum, sum2;

/* Pointer to function to return integer pixel value */
/*   int (*intpix)(REGION *, int, int) = regPixIntFunc (reg); */
/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg);

/* To access a pixel as a double: dblpix (reg, row, col) */
/* To access a pixel as an int:   intpix (reg, row, col) */


/* Begin code */
/* First, adjust r1, etc if needed */

   if (row1 < 0) row1 = 0;
   if (row1 + nrow > nr) nrow = nr - row1;
   if (col1 < 0) col1 = 0;
   if (col1 + ncol > nc) ncol = nc - col1;

   sum=0.;  sum2=0.; *minval = 65535; *maxval = -65536; npix = 0;
   *mean = 0.; *stdev = 0.;

   for (i=row1; i<row1+nrow; i++) {
	for (j=col1; j<=col1+ncol; j++) {
	   npix++;
	   pix = dblpix (reg, i, j);
	   sum = sum + pix;
	   sum2 = sum2 + pix*pix;
	   if (pix < *minval) {
		*minval = pix;
		*rowmin = i;
		*colmin = j;
		}
	   if (pix > *maxval) {
		*maxval = pix;
		*rowmax = i;
		*colmax = j;
		}
	   }
	}
   if (npix == 0) return IMG_SUCCESS;
   *mean = sum / npix;

   *stdev = sqrt (dmax(sum2 / npix - (*mean) * (*mean), 0.));

   return IMG_SUCCESS;
}


/****************************************************************************/
/**********************		imgHist			*********************/
/****************************************************************************/

/* Given a region, compute a histogram of pixel values.  Size and range
 * of pixel values is input.  For generality, a section of an image is
 * input.  Storage for the histogram must have been allocated externally.
*/

IMG_CODE imgHist (
	REGION *reg,	/* IN: Region to work on */
	int row1,	/* IN: Starting row */
	int col1,	/* IN: Starting col */
	int row2,	/* IN: Ending row */
	int col2,	/* IN: Ending col */
	int pixmin,	/* IN: Minimum pix value */
	int pixmax,	/* IN: Maximum pix value */
	int skip,	/* IN: Look at only every n'th pixel (for speed) */
	int *pixval,	/* MODIFIED: Array of pixel values */
	int *bin)	/* MODIFIED: Histogram */
{

/* pixval and bin arrays must have pixmax-pixmin+1 storage locations already
 * allocated
*/

   int nr = reg->nrow;
   int nc = reg->ncol;

/* Pointer to function to return integer pixel value */
   int (*intpix)(REGION *, int, int) = regPixIntFunc (reg);
/* Pointer to function to return double pixel value */
/*   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg); */

/* To access a pixel as a double: dblpix (reg, row, col) */
/* To access a pixel as an int:   intpix (reg, row, col) */

   int r1, r2, c1, c2;
   int pix;
   int i,j;

/* Begin code */
/* First, adjust r1, etc if needed */

   if (row1 < 0) {r1 = 0;} else {r1 = row1;}
   if (row2 > nr-1) {r2 = nr-1;} else {r2 = row2;}

   if (col1 < 0) {c1 = 0;} else {c1 = col1;}
   if (col2 > nc-1) {c2 = nc-1;} else {c2 = col2;}

/* Zero out the bin array */
/* Try for speed? */
   for (i = pixmin; i <= pixmax; i++) {
	pixval[i - pixmin] = i;
/*	bin[i - pixmin] = 0; */
	}


   memset (bin, 0, (pixmax-pixmin+1)*sizeof(int));

/* loop through all the pixels */
   for (i=r1; i<=r2; i=i+skip) {
	for (j=c1; j<=c2; j=j+skip) {
	   pix = intpix (reg, i, j);
	   if ((pix >= pixmin) && (pix <= pixmax)) {
		bin[pix - pixmin]++;
		}
	   }
	}

   return IMG_SUCCESS;
   }
