/* pixbin.c --
*  June 1993 Steve Kent
*   
* The following routines fetch pixel values of a specified row and column.
* They all check for frame boundaries.  A value 0 is returned if a pixel
* is out of bounds.
*
*/

#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

#include "dervish.h"
#include "imgProto.h"

#define min(x,y) (x) < (y) ? (x) : (y)
#define max(x,y) (x) > (y) ? (x) : (y)

/******************************* pixbin ************************************/

/* Remember that .5, .5 is the center of a pixel */

double pixbin (REGION *reg, double row, double col) {

   double pix1, pix2, pix3, pix4 ;
   int nr, nc ;
   int r1,r2,c1,c2 ;

   nr = reg->nrow;
   nc = reg->ncol;
   row = row-.5;
   col = col-.5;
   r1 = row;
   c1 = col;
   r2 = r1 + 1;
   c2 = c1 + 1;

/* Kick out if we are out of bounds */
   if (r1 < 0 || r1 >= nr-1) return 0.;
   if (c1 < 0 || c1 >= nc-1) return 0.;

   pix1 = shRegPixGetAsDbl (reg, r1, c1);
   pix2 = shRegPixGetAsDbl (reg, r1, c2);
   pix3 = shRegPixGetAsDbl (reg, r2, c1);
   pix4 = shRegPixGetAsDbl (reg, r2, c2);

   pix1 = (r2-row)*(c2-col)*pix1;
   pix2 = (r2-row)*(col-c1)*pix2;
   pix3 = (row-r1)*(c2-col)*pix3;
   pix4 = (row-r1)*(col-c1)*pix4;
   return (pix1 + pix2 + pix3 + pix4);
   }

/******************************* xpixbin ************************************/

double xpixbin (REGION *reg, double row, double col) {

   int n=2, n2=5;
   double beta=.75;
/* double PI=3.141593; */
   double a[5][5];
   int ix,iy;
   double dx,dy;
   double ax, ay, px, py;
   int i,j;
   int nx, ny;
   int nr, nc;
   double sum, bin;

   for (i=0; i<n2; i++) {
	for (j=0; j<n2; j++) {
	   a[i][j]=1.;
	   }
	}
   ix=row;
   iy=col;
   dx=row-ix;
   dy=col-iy;
   if (dx>0.5) {
	dx=dx-1.;
	ix=ix+1;
	}
   if (dy>0.5) {
	dy=dy-1.;
	iy=iy+1;
	}
   for (i=0; i<n2; i++) {
	ax=i-n-dx;
	ay=i-n-dy;
	if (ax == 0.) px=1.;
		else px = sin(PI*ax)*exp(-pow(ax*beta,2))/(PI*ax);
	if (ay == 0.) py=1.;
		else py = sin(PI*ay)*exp(-pow(ay*beta,2))/(PI*ay);
	for (j=0; j<n2; j++) {
	   a[j][i] *= px;
	   a[i][j] *= py;
	   }
	}
   sum=0.;
   for (i=0; i<n2; i++) {
	for (j=0; j<n2; j++) {
	   sum += a[i][j];
	   }
	}
   for (i=0; i<n2; i++) {
	for (j=0; j<n2; j++) {
	   a[i][j] /= sum;
	   }
	}
   bin=0.;
   nr = reg->nrow;
   nc = reg->ncol;
   for (i=0; i<n2; i++) {
	for (j=0; j<n2; j++) {
	   nx = min(nr, max(0,ix+j-n));
	   ny = min(nc, max(0,iy+i-n));
	   bin += a[i][j]*shRegPixGetAsDbl (reg, nx, ny);
	   }
	}
   return bin;
   }

/* Functions to get a pixel value as an int */
static int regPixIntFromU8 (REGION *reg, int i, int j) {
   return (reg->rows_u8)[i][j];
   }
static int regPixIntFromS8 (REGION *reg, int i, int j) {
   return (reg->rows_s8)[i][j];
   }
static int regPixIntFromU16 (REGION *reg, int i, int j) {
   return (reg->rows_u16)[i][j];
   }
static int regPixIntFromS16 (REGION *reg, int i, int j) {
   return (reg->rows_s16)[i][j];
   }
static int regPixIntFromU32 (REGION *reg, int i, int j) {
   return (reg->rows_u32)[i][j];
   }
static int regPixIntFromS32 (REGION *reg, int i, int j) {
   return (reg->rows_s32)[i][j];
   }
static int regPixIntFromFL32 (REGION *reg, int i, int j) {
   return (reg->rows_fl32)[i][j];
   }

void *regPixIntFunc (REGION *reg) {
   if (reg->type == TYPE_U8) return regPixIntFromU8;
   if (reg->type == TYPE_S8) return regPixIntFromS8;
   if (reg->type == TYPE_U16) return regPixIntFromU16;
   if (reg->type == TYPE_S16) return regPixIntFromS16;
   if (reg->type == TYPE_U32) return regPixIntFromU32;
   if (reg->type == TYPE_S32) return regPixIntFromS32;
   return regPixIntFromFL32;
   }

/* Functions to get a pixel value as a double */
static double regPixDblFromU8 (REGION *reg, int i, int j) {
   return (reg->rows_u8)[i][j];
   }
static double regPixDblFromS8 (REGION *reg, int i, int j) {
   return (reg->rows_s8)[i][j];
   }
static double regPixDblFromU16 (REGION *reg, int i, int j) {
   return (reg->rows_u16)[i][j];
   }
static double regPixDblFromS16 (REGION *reg, int i, int j) {
   return (reg->rows_s16)[i][j];
   }
static double regPixDblFromU32 (REGION *reg, int i, int j) {
   return (reg->rows_u32)[i][j];
   }
static double regPixDblFromS32 (REGION *reg, int i, int j) {
   return (reg->rows_s32)[i][j];
   }
static double regPixDblFromFL32 (REGION *reg, int i, int j) {
   return (reg->rows_fl32)[i][j];
   }

void *regPixDblFunc (REGION *reg) {
   if (reg->type == TYPE_U8) return regPixDblFromU8;
   if (reg->type == TYPE_S8) return regPixDblFromS8;
   if (reg->type == TYPE_U16) return regPixDblFromU16;
   if (reg->type == TYPE_S16) return regPixDblFromS16;
   if (reg->type == TYPE_U32) return regPixDblFromU32;
   if (reg->type == TYPE_S32) return regPixDblFromS32;
   return regPixDblFromFL32;
   }



/*============================================================================
**============================================================================
**
** LOCAL MACROS, DEFINITIONS, ETC.
**
**============================================================================
*/

/* macro to "round down", assisting in the conversion of floating point 
** types to ints.
*/ 
#define ROUND_DOWN(d) ((d)>= 0.0 ? (d) + 0.5: -(-(d) + 0.5))

static void set_u8FromInt (REGION *reg, int i, int j, int num)
{
   if (num < U8_MIN) {
	reg->rows_u8[i][j]= U8_MIN; return;}
   if (num > U8_MAX) {
	reg->rows_u8[i][j]= U8_MAX; return;}
   reg->rows_u8[i][j]= num; return;
}

static void set_s8FromInt (REGION *reg, int i, int j, int num)
{
   if (num < S8_MIN) {
	reg->rows_s8[i][j]= S8_MIN; return;}
   if (num > S8_MAX) {
	reg->rows_s8[i][j]= S8_MAX; return;}
   reg->rows_s8[i][j]= num; return;
}

static void set_u16FromInt (REGION *reg, int i, int j, int num)
{
   if (num < U16_MIN) {
	reg->rows_u16[i][j]= U16_MIN; return;}
   if (num > U16_MAX) {
	reg->rows_u16[i][j]= U16_MAX; return;}
   reg->rows_u16[i][j]= num; return;
}

static void set_s16FromInt (REGION *reg, int i, int j, int num)
{
   if (num < S16_MIN) {
	reg->rows_s16[i][j]= S16_MIN; return;}
   if (num > S16_MAX) {
	reg->rows_s16[i][j]= S16_MAX; return;}
   reg->rows_s16[i][j]= num; return;
}

static void set_u32FromInt (REGION *reg, int i, int j, int num)
{
   if (num < U32_MIN) {
	reg->rows_u32[i][j]= U32_MIN; return;}
   if (num > U32_MAX) {
	reg->rows_u32[i][j]= U32_MAX; return;}
   reg->rows_u32[i][j]= num; return;
}

static void set_s32FromInt (REGION *reg, int i, int j, int num)
{
   if (num < S32_MIN) {
	reg->rows_s32[i][j]= S32_MIN; return;}
   if (num > S32_MAX) {
	reg->rows_s32[i][j]= S32_MAX; return;}
   reg->rows_s32[i][j]= num; return;
}

static void set_fl32FromInt (REGION *reg, int i, int j, int num)
{
   reg->rows_fl32[i][j]= num; return;
}


void (*regPixIntFuncSet (REGION *reg)) \
   (REGION *, int, int, int) {
   if (reg->type == TYPE_U8) return set_u8FromInt;
   if (reg->type == TYPE_S8) return set_s8FromInt;
   if (reg->type == TYPE_U16) return set_u16FromInt;
   if (reg->type == TYPE_S16) return set_s16FromInt;
   if (reg->type == TYPE_U32) return set_u32FromInt;
   if (reg->type == TYPE_S32) return set_s32FromInt;
   return set_fl32FromInt;
   }

/************************************************************************/
/* NOTE: The follow routines are flawed, because the ROUND_DOWN function
 * can round a valid number outside the valid range!
 */

static void set_u8FromDouble (REGION *reg, int i, int j, double dbl)
{
   if (dbl < (double)U8_MIN) {
	reg->rows_u8[i][j]= U8_MIN; return;}
   if (dbl > (double)U8_MAX) {
	reg->rows_u8[i][j]= U8_MAX; return;}
   reg->rows_u8[i][j]= ROUND_DOWN(dbl); return;
}

static void set_s8FromDouble (REGION *reg, int i, int j, double dbl)
{
   if (dbl < (double)S8_MIN) {
	reg->rows_s8[i][j]= S8_MIN; return;}
   if (dbl > (double)S8_MAX) {
	reg->rows_s8[i][j]= S8_MAX; return;}
   reg->rows_s8[i][j]= ROUND_DOWN(dbl); return;
}

/* This is fixed. */
static void set_u16FromDouble (REGION *reg, int i, int j, double dbl)
{
   dbl = ROUND_DOWN(dbl);
   if (dbl < (double)U16_MIN) {
	reg->rows_u16[i][j]= U16_MIN; return;}
   if (dbl > (double)U16_MAX) {
	reg->rows_u16[i][j]= U16_MAX; return;}
   reg->rows_u16[i][j]= dbl; return;
}

static void set_s16FromDouble (REGION *reg, int i, int j, double dbl)
{
   if (dbl < (double)S16_MIN) {
	reg->rows_s16[i][j]= S16_MIN; return;}
   if (dbl > (double)S16_MAX) {
	reg->rows_s16[i][j]= S16_MAX; return;}
   reg->rows_s16[i][j]= ROUND_DOWN(dbl); return;
}

static void set_u32FromDouble (REGION *reg, int i, int j, double dbl)
{
   if (dbl < (double)U32_MIN) {
	reg->rows_u32[i][j]= U32_MIN; return;}
   if (dbl > (double)U32_MAX) {
	reg->rows_u32[i][j]= U32_MAX; return;}
   reg->rows_u32[i][j]= ROUND_DOWN(dbl); return;
}

static void set_s32FromDouble (REGION *reg, int i, int j, double dbl)
{
   if (dbl < (double)S32_MIN) {
	reg->rows_s32[i][j]= S32_MIN; return;}
   if (dbl > (double)S32_MAX) {
	reg->rows_s32[i][j]= S32_MAX; return;}
   reg->rows_s32[i][j]= ROUND_DOWN(dbl); return;
}

static void set_fl32FromDouble (REGION *reg, int i, int j, double dbl)
{
   reg->rows_fl32[i][j]= dbl; return;
}


void (*regPixDblFuncSet (REGION *reg)) \
   (REGION *, int, int, double) {
   if (reg->type == TYPE_U8) return set_u8FromDouble;
   if (reg->type == TYPE_S8) return set_s8FromDouble;
   if (reg->type == TYPE_U16) return set_u16FromDouble;
   if (reg->type == TYPE_S16) return set_s16FromDouble;
   if (reg->type == TYPE_U32) return set_u32FromDouble;
   if (reg->type == TYPE_S32) return set_s32FromDouble;
   return set_fl32FromDouble;
   }

