/*****************************************************************************
******************************************************************************
**
** FILE:
**      imgRowCol.c
**
** ABSTRACT:
**      This file contains functions to perform various row and
**	column oriented operations on a region
**
** ENTRY POINT          SCOPE   DESCRIPTION
** -------------------------------------------------------------------------
** imgRowMed		public	Compute flatfield vector by medianing
** imgColMed		public	Compute flatfield vector by medianing
** imgRowAvg		public	Compute bias vector by averaging
** imgColAvg		public	Compute bias vector by averaging
** imgRowSub		public	Subtract a row vector from a region
** imgColSub		public	Subtract a col vector from a region
** imgConAdd		public	Add a constant to a region
** imgCsmooth		public	Smooth a region in the column direction
**
** ENVIRONMENT:
**      ANSI C.
**
** REQUIRED PRODUCTS:
**      FTCL            TCL + Fermi extension
**      DERVISH
**      IMAGE
**
** AUTHORS:
**      Creation date:  Oct. 1993
**      Steve Kent
**
******************************************************************************
******************************************************************************
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "img_msg_c.h"
#include "imgProto.h"

/***************************************************************************/
/**************************	imgRowMed	****************************/
/***************************************************************************/
/* Compute the histogram of pixel distributions and extract the
** value of the median.  Do this row by row.
** regOut is allocated by the user (who selects the data type).  Must
** have ncol = 1, nrow = nrow(regIn).
*/

IMG_CODE imgRowMed (
	REGION *regIn,		/* IN: Full 2-d image */
	REGION *regOut)		/* MODIFIED: Output 1-D row vector */
{

/************************ Declarations *********************************/

   int col1, col2;
/* skip factor - fixed for now */
   int skip=1;
/* bin storage */
   int *pixval, *bin;
/* working constants */
   int i, irow, nbin, ntot=0;
   double fract;
   double sum, half;
   int ifirst=0;


/************************ Begin code	*********************************/

/* allocate storage for the working arrays */
   nbin = 65536;
   pixval = (int *)shMalloc(nbin * sizeof (int) );
   bin = (int *)shMalloc (nbin * sizeof (int) );

/* Now do the real work */

   col1=0;
   col2=regIn->ncol-1;
   for (irow=0; irow<regIn->nrow; irow++) {
	if (imgHist(regIn, irow, col1, irow, col2, 0, nbin-1,
	skip, pixval, bin) !=0) return (TCL_ERROR);

/* How many pixels total? We need do this only once. */
	if (ifirst == 0) {
	   ntot = 0;
	   for (i=0; i<nbin; i++) {
		ntot = ntot + bin[i];
		}
	   if (ntot == 0) return IMG_GENERIC_ERROR;
	   ifirst = 1;
	   }

/* Now compute the ntile values */
	sum = 0;
	half = .5*ntot;
	for (i=0; i<nbin; i++) {
	   sum = sum + bin[i];
	   if (sum >= half) {
		fract = 1. + (.5 * ntot - sum) / bin[i];
		pxSetWithDbl (regOut, irow, 0, i - 0.5 + fract);
		break;
		}
	   }
	}
   shFree(bin);
   shFree(pixval);   
   return (IMG_SUCCESS);
}

/***************************************************************************/
/**************************	imgColMed	****************************/
/***************************************************************************/
/* Compute the histogram of pixel distributions and extract the
** value of the median.  Do this column by column.
** regOut is allocated by the user (who selects the data type).  Must
** have nrow = 1, ncol = ncol(regIn).
*/

IMG_CODE imgColMed (
	REGION *regIn,		/* IN: Full 2-d image */
	REGION *regOut)		/* MODIFIED: Output 1-D row vector */
{

/************************ Declarations *********************************/

   int row1, row2;
/* skip factor - fixed for now */
   int skip=1;
/* bin storage */
   int *pixval, *bin;
/* working constants */
   int i, icol, nbin, ntot=0;
   double fract;
   double sum, half;
   int ifirst=0;

/************************ Begin code	*********************************/

/* allocate storage for the working arrays */
   nbin = 65536;
   pixval = (int *)shMalloc(nbin * sizeof (int) );
   bin = (int *)shMalloc (nbin * sizeof (int) );

/* Now do the real work */

   row1=0;
   row2=regIn->nrow-1;
   for (icol=0; icol<regIn->ncol; icol++) {
	if (imgHist(regIn, row1, icol, row2, icol, 0, nbin-1,
	skip, pixval, bin) !=0) return (TCL_ERROR);

/* How many pixels total? We need do this only once. */
	if (ifirst == 0) {
	   ntot = 0;
	   for (i=0; i<nbin; i++) {
		ntot = ntot + bin[i];
		}
	   if (ntot == 0) return IMG_GENERIC_ERROR;
	   ifirst = 1;
	   }

/* Now compute the ntile values */
	sum = 0;
	half = ntot*.5;
	for (i=0; i<nbin; i++) {
	   sum = sum + bin[i];
	   if (sum >= half) {
		fract = 1. + (.5 * ntot - sum) / bin[i];
		pxSetWithDbl (regOut, 0, icol, i - 0.5 + fract);
		break;
		}
	   }
	}
   shFree(bin);
   shFree(pixval);   
   return (IMG_SUCCESS);
}

/***************************************************************************/
/**************************	imgRowAvg	****************************/
/***************************************************************************/
/* Compute the row by row average of columns.
** regOut is allocated by the user (who selects the data type).  Must
** have ncol = 1, nrow = nrow(regIn).
*/

IMG_CODE imgRowAvg (
	REGION *regIn,		/* IN: Full 2-d image */
	REGION *regOut)		/* MODIFIED: Output 1-D row vector */
{

/************************ Declarations *********************************/

/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (regIn);
   void (*dblpixset)(REGION *, int, int, double) = regPixDblFuncSet (regOut);

   int irow;
   int icol;
   double sum;

/************************ Begin code	*********************************/

/* Now do the real work */

   if (regIn->ncol == 0) return IMG_GENERIC_ERROR;
   for (irow=0; irow<regIn->nrow; irow++) {
	sum = 0.;
	for (icol=0; icol<regIn->ncol; icol++) {
	   sum += dblpix(regIn, irow, icol);
	   }
	dblpixset (regOut, irow, 0, sum/regIn->ncol);
	}
   return IMG_SUCCESS;
}

/***************************************************************************/
/**************************	imgColAvg	****************************/
/***************************************************************************/
/* Compute the column by column average of rows.
** regOut is allocated by the user (who selects the data type).  Must
** have nrow = 1, ncol = ncol(regIn).
*/

IMG_CODE imgColAvg (
	REGION *regIn,		/* IN: Full 2-d image */
	REGION *regOut)		/* MODIFIED: Output 1-D row vector */
{

/************************ Declarations *********************************/

/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (regIn);
   void (*dblpixset)(REGION *, int, int, double) = regPixDblFuncSet (regOut);

   int irow;
   int icol;
   double sum;

/************************ Begin code	*********************************/

/* Now do the real work */

   if (regIn->nrow == 0) return IMG_GENERIC_ERROR;
   for (icol=0; icol<regIn->ncol; icol++) {
	sum = 0.;
	for (irow=0; irow<regIn->nrow; irow++) {
	   sum += dblpix(regIn, irow, icol);
	   }
	dblpixset (regOut, 0, icol, sum/regIn->nrow);
	}
   return IMG_SUCCESS;
}

/***************************************************************************/
/**************************	imgRowSub	****************************/
/***************************************************************************/
/* Subtract a vector row by row from a region.  Aplly dither if specified.
*/

IMG_CODE imgRowSub (
	REGION *reg,		/* MODIFIED: Full 2-d image */
	REGION *regRow,		/* IN: 1-D row vector */
	int dither)		/* IN: TRUE = dither */
{

/************************ Declarations *********************************/

/* Pointer to function to return double pixel value */
   double (*intpix)(REGION *, int, int) = regPixIntFunc (reg);
   double (*dblpix1)(REGION *, int, int) = regPixDblFunc (reg);
   double (*dblpix2)(REGION *, int, int) = regPixDblFunc (regRow);
   void (*dblpixset)(REGION *, int, int, double) = regPixDblFuncSet (reg);

   int irow;
   int icol;
   double pix;
   double diff;
   int iseed;
   float seed;

/************************ Begin code	*********************************/

/* Let the seed be the sum of pixels in the 1st row of an image */
   seed = 0;
   for (icol=0; icol<reg->ncol; icol++) {
	seed += dblpix1(reg, 0, icol);
	}
   iseed = seed/reg->ncol;
   reseed(iseed);

/* Now do the real work */

   for (irow=0; irow<reg->nrow; irow++) {
	pix = dblpix2(regRow, irow, 0);
	for (icol=0; icol<reg->ncol; icol++) {
	   diff = dblpix1(reg, irow, icol) - pix;
	   if (dither) diff += drand() - 0.5;
	   dblpixset (reg, irow, icol, diff);
	   }
	}
   return IMG_SUCCESS;
}


/***************************************************************************/
/**************************	imgColSub	****************************/
/***************************************************************************/
/* Subtract a vector col by col from a region.  Aplly dither if specified.
*/

IMG_CODE imgColSub (
	REGION *reg,		/* MODIFIED: Full 2-d image */
	REGION *regCol,		/* IN: 1-D row vector */
	int dither)	   	/* IN: 1 = dither */
{

/************************ Declarations *********************************/

/* Pointer to function to return double pixel value */
   double (*intpix)(REGION *, int, int) = regPixIntFunc (reg);
   double (*dblpix1)(REGION *, int, int) = regPixDblFunc (reg);
   double (*dblpix2)(REGION *, int, int) = regPixDblFunc (regCol);
   void (*dblpixset)(REGION *, int, int, double) = regPixDblFuncSet (reg);

   int irow;
   int icol;
   double pix;
   double diff;
   int iseed;
   float seed;

/************************ Begin code	*********************************/

/* Let the seed be the sum of pixels in the 1st col of an image */
   seed = 0;
   for (irow=0; irow<reg->nrow; irow++) {
     	seed += dblpix1(reg, irow, 0);
	}
   iseed = seed/reg->nrow;
   reseed(iseed);

/* Now do the real work */

   for (icol=0; icol<reg->ncol; icol++) {
	pix = dblpix2(regCol, 0, icol);
	for (irow=0; irow<reg->nrow; irow++) {
	   diff = dblpix1(reg, irow, icol) - pix;
	   if (dither) diff += drand() - 0.5;
	   dblpixset (reg, irow, icol, diff);
	   }
	}
   return IMG_SUCCESS;
}

/***************************************************************************/
/**************************	imgConAdd	****************************/
/***************************************************************************/
/* Add a constant value to a region.  Dither if specified.
*/

IMG_CODE imgConAdd (
	REGION *reg,		/* MODIFIED: Full 2-d image */
	double value,		/* IN: The constant value */
	int dither)	   	/* IN: 1 = dither */
{

/************************ Declarations *********************************/

/* Pointer to function to return double pixel value */
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg);
   void (*dblpixset)(REGION *, int, int, double) = regPixDblFuncSet (reg);

   int irow;
   int icol;
   double sum;
   int seed;

/************************ Begin code	*********************************/

/* Let the seed be the constant itself. */
   seed = value;
   reseed(seed);

/* Now do the real work */

   for (icol=0; icol<reg->ncol; icol++) {
	for (irow=0; irow<reg->nrow; irow++) {
	   sum = dblpix(reg, irow, icol) + value;
	   if (dither) sum += drand() - 0.5;
	   dblpixset (reg, irow, icol, sum);
	   }
	}
   return IMG_SUCCESS;
}

/***************************************************************************/
/**************************	imgCsmooth	****************************/
/***************************************************************************/
/* Subtract a vector col by col from a region.  Aplly dither if specified.
*/

IMG_CODE imgCsmooth (
	REGION *reg,		/* MODIFIED: Full 2-d image */
	int dither) 	  	/* IN: 1 = dither */
{

/************************ Declarations *********************************/

/* Pointer to function to return double pixel value */
   double (*intpix)(REGION *, int, int) = regPixIntFunc (reg);
   double (*dblpix)(REGION *, int, int) = regPixDblFunc (reg);
   void (*dblpixset)(REGION *, int, int, double) = regPixDblFuncSet (reg);

   int irow;
   int icol;
   double pix1, pix2, pix3;
   double sum;
   int iseed;
   float seed;

/************************ Begin code	*********************************/

   if (reg->ncol < 3) return IMG_SUCCESS;

/* Let the seed be the sum of pixels in the 1st col of an image */
   seed = 0;
   for (irow=0; irow<reg->nrow; irow++) {
	seed += dblpix(reg, irow, 0);
	}
   iseed = seed/reg->nrow;
   reseed(iseed);

/* Now do the real work */

   for (irow=0; irow<reg->nrow; irow++) {
	pix2 = dblpix(reg, irow, 0);
	pix3 = dblpix(reg, irow, 1);
	for (icol=1; icol<reg->ncol-1; icol++) {
	   pix1 = pix2;
	   pix2 = pix3;
	   pix3 = dblpix(reg, irow, icol+1);
	   sum = (pix1 + 2.*pix2 + pix3)/4.;
	   if (dither) sum += drand() - 0.5;
	   dblpixset (reg, irow, icol, sum);
	   }
	}
   return IMG_SUCCESS;
}
