/* sqrt.c --
*  Square root compression and uncompression routines.  Work on an image in
* place
*  Mar 1993 Steve Kent
*   
* Utilities:
* imin
* imax
* sqr
*	
* Tcl functions defined in this module:
*
* imgSqrt
* imgSqr
* imgSim
*
*/
#include <math.h>
#include <stdio.h>
#include "imgProto.h"

/*******************Utility: sqr function *************************/
static double sqr (double x) {
   return x*x;}

/*******************Utility: integer min function *************************/
static int imin(int i, int j) {
   if (i < j) {return i;} else {return j;}
   }

/*******************Utility: integer max function *************************/
static int imax(int i, int j) {
   if (i > j) {return i;} else {return j;}
   }

/************************** square root function ****************************/

IMG_CODE imgSqrt (UDATA **rows, int nr, int nc, double rnoise, double gain,
	double quant) {

   double rval;
   int i,j;


/* convert readout noise to rms photons */
   rnoise = rnoise * gain;

/* Compute square root of each pixel value */

   for (i=0; i<nr; i++) {
      for (j=0; j<nc; j++) {
	   rval = rows[i][j];
	   rval = 2. * sqrt (rnoise*rnoise + rval*gain) * quant;
           rows[i][j] = rval + 0.5;
        }
   }

/*   printf ("Done with square rooting\n"); */
   return (IMG_SUCCESS);
}


/**********************squaring function ****************************** */

IMG_CODE imgSqr (UDATA **rows, int nr, int nc, double rnoise, double gain,
	double quant) {

   double rval, avg, range;
   int i,j,itemp;

/* Tables of min, max input codes that lead to a given output code */
   int inmin[65536], inmax[65536];
/* Floating point output codes that correspond to the range covered by
** the input codes
*/
   double outmin[65536], outmax[65536];

/* convert readout noise to rms photons */
   rnoise = rnoise * gain;

/* Compute the range of input codes that produce a given output code */
/* First, initialize the tables */
   for (i=0; i<65535; i++) {
	inmin[i] = 100000;
	inmax[i] = 0;
   }
/* Next, loop through all possible input codes. */
   for (i=0; i<65535; i++) {
	rval = i;
	rval = 2. * sqrt (rnoise*rnoise + rval*gain) * quant;
        itemp = rval + 0.5;
	if (itemp < 65535) {
		inmin[itemp] = imin(inmin[itemp], i);
		inmax[itemp] = imax(inmax[itemp], i);
	   }
   }
/* Finally, loop through all output codes and get floating point endpoints
** of input codes.
*/
   for (i=0; i<65535; i++) {
	if (inmin[i] == 100000) {
		outmin[i]=0.;
		outmax[i]=0.;
		}
	   else {
		rval = inmin[i] - 0.5;
		rval = 2. * sqrt (rnoise*rnoise + rval*gain) * quant;
		outmin[i] = rval;
		rval = inmax[i] + 0.5;
		rval = 2. * sqrt (rnoise*rnoise + rval*gain) * quant;
		outmax[i] = rval;
		}
   }
/* debug */
/*
   for (i=95; i<105; i++) {
	printf ("Output code=%d, range of input codes=%d to %d.\n",
	   i, inmin[i], inmax[i]);
   }
*/
/* Compute square of each pixel value. Dither beforehand. */
   for (i=0; i<nr; i++) {
      for (j=0; j<nc; j++) {
	   itemp = rows[i][j];
/* First dither to a uniform distribution over the possible range of all output
** codes that are quantized to this value
*/
	   avg = (outmin[itemp] + outmax[itemp])/2.;
	   range = (outmax[itemp] - outmin[itemp]);
	   rval = avg + (drand() - 0.5) * range;
/* add one more unit of dithering to rval */
	   rval = rval + drand() - 0.5;
	   rval = (sqr (rval / (2. * quant)) - sqr(rnoise)) / gain;
	   itemp = rval+0.5;
	   if (itemp < 0) itemp = 0;
/*	   if (itemp > 65535) itemp = 65535;  */
/* Clip at 32767 for test purposes here; most external software cannot
handle unsigned 16 bit ints */
	   if (itemp > 32767) itemp = 32767;
           rows[i][j] = itemp;
        }
   }

/*   printf ("Done with squaring\n"); */
   return (IMG_SUCCESS);


}

/************************** simulation function ****************************/

IMG_CODE imgSim (UDATA **rows, int nr, int nc, double rnoise, double gain,
	double sky) {

   int seed=17;
   double variance, stdev;
   double rval;
   int i,j;


/* variance in sky signal per pixel and standard deviation */
   variance = sqr(rnoise * gain) + sky * gain;
   stdev = sqrt(variance) / gain;
/* Compute simulated sky value for each pixel value */

   for (i=0; i<nr; i++) {
      for (j=0; j<nc; j++) {
/* add three random numbers together to approximate a Gaussian */
/* standard deviation is sqrt(3/12) = 1/2.  Convert to unit stdev. */
	rval = gaussRan(3);
/* insert slope */
           rows[i][j] = sky + rval * stdev + 0.5;
/* + 40. * sqr((j - (nc/2.))/nc); */
        }
   }

   printf ("Done with simulation\n");
   return (IMG_SUCCESS);
}
