/*
 * TCL support for reading a binary file into a structure
 */
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "shChain.h"

static char *module = "contrib";	/* name of this set of code */

/*****************************************************************************/
/*************************** tclBinRead **********************************/
/*****************************************************************************/
/*
 * Read binary data from a file into a buffer pointed to by a handle
 */
static char *tclBinRead_use =
  "USAGE: binRead fileId handle [size]";
static char *tclBinRead_hlp =
"Read binary data from file; <size> bytes if specified; else size from schema";

static int
tclBinRead(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   HANDLE handle;
   void *userPtr;
   int size;
   int nread;
   char buf[80];
   const SCHEMA *schema;
   FILE *file;
 
   if(argc != 3 && argc != 4) {
	Tcl_SetResult(interp,tclBinRead_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Translate file handle */
   if (Tcl_GetOpenFile (interp, argv[1], 0, 0, &file) != TCL_OK)
	return TCL_ERROR;

/* Get Handle */
   if (shTclHandleExprEval (interp, argv[2], &handle, &userPtr) \
	!= TCL_OK) return TCL_ERROR;

/* Get size if input */
   if (argc == 4) {
	if (Tcl_GetInt (interp, argv[3], &size) != TCL_OK) return TCL_ERROR;
   } else {
	schema = shSchemaGet (shNameGetFromType(handle.type));
	if (schema == NULL) {
	   Tcl_SetResult (interp, "Can't get schema for input handle",
		TCL_STATIC);
	   return TCL_ERROR;
	   }
	size = schema->size;
	}
   nread = fread (handle.ptr, 1, size, file);
   sprintf (buf, "%d", nread);
   Tcl_SetResult (interp, buf, TCL_VOLATILE);
   return TCL_OK;
}
/*****************************************************************************/
/*************************** tclBinWrite **********************************/
/*****************************************************************************/
/*
 * Write binary data to a file from a buffer pointed to by a handle
 */
static char *tclBinWrite_use =
  "USAGE: binWrite fileId handle [size]";
static char *tclBinWrite_hlp =
"Write binary data to a file; <size> bytes if specified; else size from schema";

static int
tclBinWrite(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   HANDLE handle;
   void *userPtr;
   int size;
   int nwrite;
   char buf[80];
   const SCHEMA *schema;
   FILE *file;
 
   if(argc != 3 && argc != 4) {
	Tcl_SetResult(interp,tclBinWrite_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Translate file handle */
   if (Tcl_GetOpenFile (interp, argv[1], 0, 0, &file) != TCL_OK)
	return TCL_ERROR;

/* Get Handle */
   if (shTclHandleExprEval (interp, argv[2], &handle, &userPtr) \
	!= TCL_OK) return TCL_ERROR;

/* Get size if input */
   if (argc == 4) {
	if (Tcl_GetInt (interp, argv[3], &size) != TCL_OK) return TCL_ERROR;
   } else {
	schema = shSchemaGet (shNameGetFromType(handle.type));
	if (schema == NULL) {
	   Tcl_SetResult (interp, "Can't get schema for input handle",
		TCL_STATIC);
	   return TCL_ERROR;
	   }
	size = schema->size;
	}
   nwrite = fwrite (handle.ptr, 1, size, file);
   sprintf (buf, "%d", nwrite);
   Tcl_SetResult (interp, buf, TCL_VOLATILE);
   return TCL_OK;
}
/*****************************************************************************/
/*************************** tclBinCopy **********************************/
/*****************************************************************************/
/*
 * Copy binary data from buffer to another using handles
 * Handles must be pointers.  E.G. if a structure has a member
 *   double *pixels;
 * then the handle expression would be $hndl.pixels.
 */
static char *tclBinCopy_use =
  "USAGE: binCopy <hndl-src-ptr> <hndl-dest-ptr> size";
static char *tclBinCopy_hlp =
"Copy binary data from one handle to another; <size> bytes";

static int
tclBinCopy(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   int size;
   HANDLE hndl1;
   HANDLE hndl2;
   void *userPtr1;
   void *userPtr2;
 
   if(argc != 4) {
	Tcl_SetResult(interp,tclBinCopy_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Translate handles */

   if (shTclHandleExprEval (interp, argv[1], (void *)&hndl1, &userPtr1) \
	!= TCL_OK) return TCL_ERROR;
   if (hndl1.type != shTypeGetFromName("PTR")) {
	Tcl_SetResult (interp, "Handle must be to a pointer", TCL_STATIC);
	return TCL_ERROR;
	}

   if (shTclHandleExprEval (interp, argv[2], (void *)&hndl2, &userPtr2) \
	!= TCL_OK) return TCL_ERROR;
   if (hndl2.type != shTypeGetFromName("PTR")) {
	Tcl_SetResult (interp, "Handle must be to a pointer", TCL_STATIC);
	return TCL_ERROR;
	}

   if (Tcl_GetInt (interp, argv[3], &size) != TCL_OK) return TCL_ERROR;

   memcpy (*(void **)(hndl2.ptr), *(void **)(hndl1.ptr), size);
   return TCL_OK;
}
/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
shTclBinReadDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"binRead", 
		(Tcl_CmdProc *)tclBinRead, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclBinRead_hlp, tclBinRead_use);

   shTclDeclare(interp,"binWrite", 
		(Tcl_CmdProc *)tclBinWrite, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclBinWrite_hlp, tclBinWrite_use);

   shTclDeclare(interp,"binCopy", 
		(Tcl_CmdProc *)tclBinCopy, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclBinCopy_hlp, tclBinCopy_use);

   return;
}

