/*****************************************************************************
******************************************************************************
**
** FILE:
**	imgTclDecimate.c
**
** ABSTRACT:
**	This file contains code to decimate a region
**
**
******************************************************************************
******************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "imgProto.h"

static char *module = "img";		/* name of this facility */


/*============================================================================  
**============================================================================
**
** ROUTINE: tclDecimate
**
** DESCRIPTION:
**	Call routine to take sqrt of each pixel in an unsigned short int image.
**	This processing is done to enhance compressibility of an image.
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**
**============================================================================
*/
/*
 */

static char *tclDecimate_use =
  "USAGE: imgDecimate <region-handle> factor";
static char *tclDecimate_hlp =
  "Compress an image via decimation by <factor>";

static int
tclDecimate(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   REGION *regout = NULL;
   int nr, nc;
   int factor = 1;
   char tclRegionName[MAXTCLREGNAMELEN];

   shErrStackClear();

   if(argc != 3) {
      Tcl_SetResult(interp,tclDecimate_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);


   if (Tcl_GetInt (interp, argv[2], &factor) != TCL_OK) return TCL_ERROR;
   if (factor <= 0) {
	Tcl_SetResult (interp, "Invalid factor", TCL_STATIC);
	return TCL_ERROR;
	}

/* Get a new region */

   if (shTclRegNameGet(interp, tclRegionName) != TCL_OK) {
	Tcl_SetResult(interp ,"Cannot find region",TCL_VOLATILE);
	return(TCL_ERROR);
	}

   nr = reg->nrow;
   nc = reg->ncol;

   regout = shRegNew(tclRegionName, nr/factor, nc/factor,
                           reg->type);
   if (regout == NULL) {
        shTclRegNameDel(interp, tclRegionName);
        return(TCL_ERROR);
	}
/* Associate region with TCL region name */

   if (shTclRegNameSetWithAddr(interp, regout, tclRegionName) != TCL_OK) {
	shRegDel(regout);
	shTclRegNameDel(interp, tclRegionName);
	Tcl_SetResult (interp, "Error creating region",TCL_STATIC);
	return TCL_ERROR;
	}

   imgDecimate (reg, regout, factor);

   Tcl_SetResult (interp, tclRegionName, TCL_VOLATILE);
   return(TCL_OK);
}


/*============================================================================  
**============================================================================
**
** ROUTINE: tclCompress
**
** DESCRIPTION:
**	Call routine to take sqrt of each pixel in an unsigned short int image.
**	This processing is done to enhance compressibility of an image.
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**
**============================================================================
*/
/*
 */

static char *tclCompress_use =
  "USAGE: imgCompress <region-handle> factor";
static char *tclCompress_hlp =
  "Compress an image by <factor>";

static int
tclCompress(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   REGION *regout = NULL;
   int nr, nc;
   int factor = 1;
   char tclRegionName[MAXTCLREGNAMELEN];

   shErrStackClear();

   if(argc != 3) {
      Tcl_SetResult(interp,tclCompress_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);


   if (Tcl_GetInt (interp, argv[2], &factor) != TCL_OK) return TCL_ERROR;
   if (factor <= 0) {
	Tcl_SetResult (interp, "Invalid factor", TCL_STATIC);
	return TCL_ERROR;
	}

/* Get a new region */

   if (shTclRegNameGet(interp, tclRegionName) != TCL_OK) {
	Tcl_SetResult(interp ,"Cannot find region",TCL_VOLATILE);
	return(TCL_ERROR);
	}

   nr = reg->nrow;
   nc = reg->ncol;

   regout = shRegNew(tclRegionName, nr/factor, nc/factor,
                           reg->type);
   if (regout == NULL) {
        shTclRegNameDel(interp, tclRegionName);
        return(TCL_ERROR);
	}
/* Associate region with TCL region name */

   if (shTclRegNameSetWithAddr(interp, regout, tclRegionName) != TCL_OK) {
	shRegDel(regout);
	shTclRegNameDel(interp, tclRegionName);
	Tcl_SetResult (interp, "Error creating region",TCL_STATIC);
	return TCL_ERROR;
	}

   imgCompress (reg, regout, factor);

   Tcl_SetResult (interp, tclRegionName, TCL_VOLATILE);
   return(TCL_OK);
}


/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclDecimateDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"imgDecimate",
		(Tcl_CmdProc *)tclDecimate,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclDecimate_hlp, tclDecimate_use);

   shTclDeclare(interp,"imgCompress",
		(Tcl_CmdProc *)tclCompress,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclCompress_hlp, tclCompress_use);

   return;
}

