/*
 * TCL support for image utilities
 *
 * ENTRY POINT		SCOPE
 * imgTclFbsDeclare	public	declare all the verbs defined in this module
 * tclFbs		static	find bright stars
 * tclAxes		static	compute axes at one position
 * tclCAxes		static	compute axes at one position, don't weight pixs
 *
 * REQUIRED PRODUCTS:
 *	FTCL	 	TCL + Fermi extension
 *      DERVISH
 *
 */
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "shGeneric.h"
#include "imgStar.h"
#include "img_msg_c.h"
#include "imgProto.h"

static char *module = "img";		/* name of this facility */

/*****************************************************************************/
/*************************** tclFBS     **************************************/
/*****************************************************************************/
/*
 * Input is handle to image and a threshold
 * return handles to : a list of stars 
 */

static char *tclFbs_use =
  "USAGE: imgFbs <region-handle> thresh max";
static char *tclFbs_hlp =
  "Find bright stars above a given threshold but below a maximum";

static int
tclFbs(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   char listName[HANDLE_NAMELEN];

   REGION *reg = NULL;
   int thresh, pixmax;
   CHAIN *starlist;

   shErrStackClear();

   if(argc != 4) {
      Tcl_SetResult(interp,tclFbs_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclFbs_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the treshold and max */

   if (Tcl_GetInt (interp, argv[2], &thresh)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[3], &pixmax)  != TCL_OK) return(TCL_ERROR);


/*
 * ok, get a handle for our new list of STARs
 */

   if (shTclChainHandleNew (interp, listName, "STAR", &starlist) != TCL_OK) {
      Tcl_SetResult(interp,tclFbs_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Done with the handles, etc. Now the action begins */


   if (imgFbs (reg, thresh, pixmax, starlist) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in FBS", TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Return handle as the result */
      Tcl_AppendResult (interp, listName, (char*)NULL);
   return(TCL_OK);
}


/*****************************************************************************/
/*************************** tclAxes **************************************/
/*****************************************************************************/
/*
 * Input is handle to image, row, column, and box size
 * Return handle to AXES
 */

static char *tclAxes_use =
  "USAGE: imgAxes <region-handle> row col xsize ysize";
static char *tclAxes_hlp =
  "Compute centroid";

static int
tclAxes(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   char handleName[HANDLE_NAMELEN];

   REGION *reg = NULL;
   int row, col, xsize, ysize;
   AXES *axes;

   shErrStackClear();

   if(argc != 6) {
      Tcl_SetResult(interp,tclAxes_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclFbs_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the row, column, and box size */

   if (Tcl_GetInt (interp, argv[2], &row)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[3], &col)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[4], &xsize)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[5], &ysize)  != TCL_OK) return(TCL_ERROR);


/*
 * ok, get a handle for our new AXES
 */
   if ((axes = (AXES *)shGenericNew("AXES")) == NULL) return TCL_ERROR;

/* Now the action begins.  Must delete axes struct if error. */

   if (imgAxes (reg, row, col, xsize, ysize, axes) != IMG_SUCCESS) {
	shGenericDel(axes);
	Tcl_SetResult(interp,"Error in Axes", TCL_STATIC);
	return(TCL_ERROR);
	}

/* Handle stuff is later to make error handling a little easier. */
   if (shTclHandleNew (interp, handleName, "AXES", (void *)axes) != TCL_OK) {
	shGenericDel(axes);
	Tcl_SetResult(interp,tclAxes_use,TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Return handle as the result */
      Tcl_AppendResult (interp, handleName, (char*)NULL);
   return(TCL_OK);
}


/*****************************************************************************/
/*************************** tclCaxes **************************************/
/*****************************************************************************/
/*
 * Input is handle to image, row, column, and box size
 * Return handle to AXES
 */

static char *tclCaxes_use =
  "USAGE: imgCaxes <region-handle> row col xsize ysize";
static char *tclCaxes_hlp =
  "Compute centroid";

static int
tclCaxes(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   char handleName[HANDLE_NAMELEN];

   REGION *reg = NULL;
   int row, col, xsize, ysize;
   AXES *axes;

   shErrStackClear();

   if(argc != 6) {
      Tcl_SetResult(interp,tclCaxes_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclFbs_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the row, column, and box size */

   if (Tcl_GetInt (interp, argv[2], &row)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[3], &col)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[4], &xsize)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[5], &ysize)  != TCL_OK) return(TCL_ERROR);


/*
 * ok, get a handle for our new AXES
 */
   if ((axes = (AXES *)shGenericNew("AXES")) == NULL) return TCL_ERROR;
   if (shTclHandleNew (interp, handleName, "AXES", (void *)axes) != TCL_OK) {
	shGenericDel(axes);
	Tcl_SetResult(interp,tclAxes_use,TCL_STATIC);
	return(TCL_ERROR);
   }

/* Done with the handles, etc. Now the action begins */

   if (imgCaxes (reg, row, col, xsize, ysize, axes) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in Axes", TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Return handle as the result */
      Tcl_AppendResult (interp, handleName, (char*)NULL);
   return(TCL_OK);
}


/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclFbsDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"imgFbs", 
		(Tcl_CmdProc *)tclFbs,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclFbs_hlp, tclFbs_use);

   shTclDeclare(interp,"imgAxes",
		(Tcl_CmdProc *)tclAxes,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclAxes_hlp, tclAxes_use);

   shTclDeclare(interp,"imgCaxes",
		(Tcl_CmdProc *)tclCaxes,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclCaxes_hlp, tclCaxes_use);

   return;
}

