/*****************************************************************************
******************************************************************************
**
** FILE:
**	imgTclFuncFit.c
**
** ABSTRACT:
**	This file contains interface routines to compute circularly averaged
**	intensity profiles
**
** ENTRY POINT		SCOPE	DESCRIPTION
** -------------------------------------------------------------------------
** imgTclFuncFitDeclare	public	declare all the verbs defined in this module
** tclFuncFit		static	Call to funcFit
**
** ENVIRONMENT:
**	ANSI C.
**
** REQUIRED PRODUCTS:
**	FTCL	 	TCL + Fermi extension
**      DERVISH
**	IMAGE
**
** AUTHORS:
**	Creation date:  Feb 95
**	Steve Kent
**
******************************************************************************
******************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "imgProto.h"

/*----------------------------------------------------------------------------
**
** GLOBAL VARIABLES
*/

static char *module = "img";		/* name of this facility */


/*============================================================================  
**============================================================================
**
** ROUTINE: tclFuncFit
**
** DESCRIPTION:
**	Interface to fit sech2 function.
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	imgFuncFit
**
**
**============================================================================
*/
/*
 * Input is chain, amplitude, fwhm.
 * return is updated amplitude, fwhm.
 */

static char *tclFuncFit_use =
  "USAGE: imgFuncFit <chain-PROF> amp fwhm fract";
static char *tclFuncFit_hlp =
  "Fit a sech2 function to a profile";

static int
tclFuncFit(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   double amp, fwhm, fract;
   CHAIN *chain;
   char buf[80];

   if(argc != 5) {
      Tcl_SetResult(interp,tclFuncFit_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get CHAIN pointer */

   if (shTclChainAddrGetFromName(interp, argv[1], "PROF", &chain) != TCL_OK)
	return (TCL_ERROR);

/* Get the amplitude and fwhm */

   if (Tcl_GetDouble (interp, argv[2], &amp)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &fwhm)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &fract)  != TCL_OK) return(TCL_ERROR);

/* Done with the handles, etc. Now the action begins */

   if (imgFuncFit (chain, &amp, &fwhm, &fract) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in FuncFit", TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Return coefficients as the result */
   sprintf (buf, "%lf %lf", amp, fwhm);
      Tcl_SetResult (interp, buf, TCL_VOLATILE);
   return(TCL_OK);
}



/*============================================================================  
**============================================================================
**
** ROUTINE: tclFunc
**
** DESCRIPTION:
**	Interface to compute sech2 function.
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	imgFunc
**
**
**============================================================================
*/
/*
 * Input is chain, amplitude, fwhm.
 * return is theoretical.
 */

static char *tclFunc_use =
  "USAGE: imgFunc <chain-PROF> amp fwhm index";
static char *tclFunc_hlp =
  "Return a sech2 function";

static int
tclFunc(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   double amp, fwhm;
   int index;
   double th;
   CHAIN *chain;
   char buf[80];

   if(argc != 5) {
      Tcl_SetResult(interp,tclFunc_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get CHAIN pointer */

   if (shTclChainAddrGetFromName(interp, argv[1], "PROF", &chain) != TCL_OK)
	return (TCL_ERROR);

/* Get the amplitude and fwhm */

   if (Tcl_GetDouble (interp, argv[2], &amp)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &fwhm)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[4], &index)  != TCL_OK) return(TCL_ERROR);

/* Done with the handles, etc. Now the action begins */

   th=imgFunc (chain, &amp, &fwhm, &index);
   
/* Return theory as the result */
   sprintf (buf, "%lf", th);
      Tcl_SetResult (interp, buf, TCL_VOLATILE);
   return(TCL_OK);
}


/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclFuncFitDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"imgFuncFit", 
		(Tcl_CmdProc *)tclFuncFit,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclFuncFit_hlp, tclFuncFit_use);

   shTclDeclare(interp,"imgFunc", 
		(Tcl_CmdProc *)tclFunc,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclFunc_hlp, tclFunc_use);

   return;
}

