/*
 * TCL support for image utilities
 *
 * ENTRY POINT		SCOPE
 * imgTclGaussDeclare	public	declare all the verbs defined in this module
 * tclGauss		static	Compute gaussian goodness of fit
 * tclPhoto		static	Box photometry
 *
 * REQUIRED PRODUCTS:
 *	FTCL	 	TCL + Fermi extension
 *      DERVISH
 *
 */
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "shGeneric.h"
#include "img_msg_c.h"
#include "imgProto.h"

static char *module = "img";		/* name of this facility */

/*****************************************************************************/
/*************************** tclGauss     **************************************/
/*****************************************************************************/
/*
 * Input is handle to image, a center, and a radius
 * return counts and area
 */

static char *tclGauss_use =
  "USAGE: imgGauss <region-handle> row col ang fwhm rnoise gain xmax ymax";
static char *tclGauss_hlp =
  "Compute goodness-of-fit to a linear Gaussian (for finding asteroids)";

static int
tclGauss(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   double rowc, colc, ang, fwhm, rnoise, gain;
   int xmax, ymax;
   double amp, erramp, off, erroff;
   double chi2;

   shErrStackClear();

   if(argc != 10) {
      Tcl_SetResult(interp,tclGauss_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclGauss_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the treshold and max */

   if (Tcl_GetDouble (interp, argv[2], &rowc)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &colc)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &ang)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[5], &fwhm)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[6], &rnoise)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[7], &gain)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[8], &xmax)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[9], &ymax)  != TCL_OK) return(TCL_ERROR);


/* Done with the handles, etc. Now the action begins */


   if (imgGauss (reg, rowc, colc, ang, fwhm, rnoise, gain,
	   xmax, ymax, &amp, &erramp, &off, &erroff, &chi2) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in Gauss", TCL_STATIC);
	return(TCL_ERROR);
	}

   sprintf (interp->result, "%.2f %.2f %.2f %.2f %.2f", amp, erramp, off,
	erroff, chi2);

   return(TCL_OK);
}


/*****************************************************************************/
/*************************** tclBoxPhoto *************************************/
/*****************************************************************************/
/*
 * Input is handle to image, a center, and a radius
 * return counts and area
 */

static char *tclBoxPhoto_use =
  "USAGE: imgGauss <region-handle> row col ang xmax ymax";
static char *tclBoxPhoto_hlp =
  "Do photometry in a rectangle";

static int
tclBoxPhoto(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   double rowc, colc, ang, counts, area;
   int xmax, ymax;

   shErrStackClear();

   if(argc != 7) {
      Tcl_SetResult(interp,tclBoxPhoto_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclGauss_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the treshold and max */

   if (Tcl_GetDouble (interp, argv[2], &rowc)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &colc)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &ang)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[5], &xmax)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetInt (interp, argv[6], &ymax)  != TCL_OK) return(TCL_ERROR);


/* Done with the handles, etc. Now the action begins */


   if (imgBoxPhoto (reg, rowc, colc, ang, xmax, ymax, &counts, &area)
	!= IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in Gauss", TCL_STATIC);
	return(TCL_ERROR);
	}

   sprintf (interp->result, "%.2f %.2f", counts, area);

   return(TCL_OK);
}


/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclGaussDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"imgGauss", 
		(Tcl_CmdProc *)tclGauss,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclGauss_hlp, tclGauss_use);

   shTclDeclare(interp,"imgBoxPhoto", 
		(Tcl_CmdProc *)tclBoxPhoto,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclBoxPhoto_hlp, tclBoxPhoto_use);

   return;
}

