/*****************************************************************************
******************************************************************************
**
** FILE:
**      TclHist.c
**
** ABSTRACT:
**      This file contains top level functions to compute ntiles and
**	create histograms of pixel values and other statistical stuff.
**
** ENTRY POINT          SCOPE   DESCRIPTION
** -------------------------------------------------------------------------
** imgTclHistDeclare    public  declare all the verbs defined in this module
** tclStats		static  Compute image statics on a section
** tclHist		static  Compute histogram of pixel values
** tclNtile		static  Compute ntile's using a histogram
**
** ENVIRONMENT:
**      ANSI C.
**
** REQUIRED PRODUCTS:
**      FTCL            TCL + Fermi extension
**      DERVISH
**      IMAGE
**
** AUTHORS:
**      Creation date:  Oct. 1993
**      Steve Kent
**
******************************************************************************
******************************************************************************
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "img_msg_c.h"
#include "imgProto.h"

static char *module="img";	/* Name of this facility */

/***************************************************************************/
/**************************	tclStats	****************************/
/***************************************************************************/

static char *tclStats_use = "imgStats region row1 row2 col1 col2";
static char *tclStats_hlp = "Compute statistics in an image patch";


static int tclStats (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv) {

   REGION *reg = NULL;
/* derived min row and max row */
   int row1, row2, col1, col2;

   char result[160];
/* temporary storage */
   double dtemp;

/* Storage for returned values */
   double mean, stdev;
   int minval, maxval;
   int rowmin, colmin, rowmax, colmax;


/* First check to see that we have right number of arguments. */
   if (argc != 6) {
      Tcl_SetResult(interp, tclStats_use, TCL_STATIC);
      return (TCL_ERROR);
   }

/* Get pointers to rows and columns.  arv[1] points to region identifier  */

/* Use new dervish interface */

   if (shTclAddrGetFromName(interp, argv[1], (void *)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);

   if ((Tcl_GetDouble(interp, argv[2], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   row1 = dtemp;
   if ((Tcl_GetDouble(interp, argv[3], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   row2 = dtemp;
   if ((Tcl_GetDouble(interp, argv[4], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   col1 = dtemp;
   if ((Tcl_GetDouble(interp, argv[5], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   col2 = dtemp;

/* Now do the real work */
   if (imgStats(reg, row1, row2-row1, col1, col2-col1,
	&mean, &stdev, &minval, &rowmin, &colmin, &maxval,
	&rowmax, &colmax)!=0) return (TCL_ERROR);

/* Create the return string as a keyed list */

   sprintf (result,"{mean %10.2f} {stdev %10.2f} {min %d} \
{rowmin %d} {colmin %d} {max %d} {rowmax %d} {colmax %d}",
	mean, stdev, minval, rowmin, colmin, maxval, rowmax, colmax);

      Tcl_SetResult(interp,result,TCL_VOLATILE);
   
   return (TCL_OK);
}


/***************************************************************************/
/**************************	tclHist		****************************/
/***************************************************************************/

/* Compute a histogram of pixel values; return as a TCL list */

static char *tclHist_use =
"USAGE: imgHist region row1 row2 col1 col2 min max";

static char *tclHist_hlp = "Compute histogram of pixel values";

static int tclHist (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv) {

   int pixmin, pixmax;
   REGION *reg = NULL;
   int row1, row2, col1, col2;
/* skip factor */
   int skip=1;

   double dtemp;

/* bin storage */
   int *pixval, *bin;
/* working constants */
   int i, nbin;
/* character strings to format the return list */
   char item[60];

/* First check to see that we have right number of arguments. */
   if (argc != 8) {
      Tcl_SetResult(interp, tclHist_use,TCL_VOLATILE);
      return (TCL_ERROR);
   }

/* Use new dervish interface */

   if (shTclAddrGetFromName(interp, argv[1], (void *)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);

   if ((Tcl_GetDouble(interp, argv[2], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   row1 = dtemp;
   if ((Tcl_GetDouble(interp, argv[3], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   row2 = dtemp;
   if ((Tcl_GetDouble(interp, argv[4], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   col1 = dtemp;
   if ((Tcl_GetDouble(interp, argv[5], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   col2 = dtemp;
   if ((Tcl_GetInt(interp, argv[6], &pixmin)!= TCL_OK)) return (TCL_ERROR);
   if ((Tcl_GetInt(interp, argv[7], &pixmax)!= TCL_OK)) return (TCL_ERROR);

/* allocate storage for the working arrays */
   nbin = pixmax - pixmin + 1;
   pixval = (int *)shMalloc(nbin * sizeof (int) );
   bin = (int *)shMalloc (nbin * sizeof (int) );

/* Now do the real work */

   if (imgHist(reg, row1, col1, row2, col2, pixmin, pixmax,
        skip, pixval, bin) !=0) return (TCL_ERROR);

/* Create the return string as a list of lists */

   for (i=pixmin; i<=pixmax; i++) {
	sprintf (item, " {%d %d} ", pixval[i-pixmin], bin[i-pixmin]);
	Tcl_AppendResult (interp, item, NULL);
	}

   shFree(bin);
   shFree(pixval);   
   return (TCL_OK);
}

/***************************************************************************/
/**************************	tclNtile	****************************/
/***************************************************************************/

/* Compute the histogram of pixel distributions and extract the
** value of the ntile.  This is useful for things like sky compuation,
** flatfield vectors, and setting threshold, saturation for images.
*/

static char *tclNtile_use =
"USAGE: imgNtile region row1 row2 col1 col2 q1 [q2 ...]";

static char *tclNtile_hlp = "Compute percentile points in pixel distribution";

static int tclNtile (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv) {

   int row1, row2, col1, col2;
/* skip factor */
   int skip=1;
   REGION *reg = NULL;

/* bin storage */
   int *pixval, *bin;
/* temporary storage of input arguments and output values*/
   double *ntile, *ntileval;
/* number of ntiles to compute */
   int nntile;
/* working constants */
   int i, n, nbin, ntot;
   double fract, sum;
/* character strings to format the return list */
   char item[60];
   double dtemp;


/* First check to see that we have right number of arguments. */
   if (argc < 7) {
      Tcl_SetResult(interp, tclNtile_use, TCL_VOLATILE);
      return (TCL_ERROR);
   }

/* Use new dervish interface */

   if (shTclAddrGetFromName(interp, argv[1], (void *)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);

   if ((Tcl_GetDouble(interp, argv[2], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   row1 = dtemp;
   if ((Tcl_GetDouble(interp, argv[3], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   row2 = dtemp;
   if ((Tcl_GetDouble(interp, argv[4], &dtemp)!= TCL_OK)) return (TCL_ERROR);
   col1 = dtemp;
   if ((Tcl_GetDouble(interp, argv[5], &dtemp)!= TCL_OK)) return (TCL_ERROR);

   col2 = dtemp;
   nntile = argc - 6;

/* Allocate all sorts of short term storage */
   ntile = (double *) shMalloc (nntile * sizeof (double));
   ntileval = (double *) shMalloc (nntile * sizeof (double));

   for (n=0; n<nntile; n++) {
	if ((Tcl_GetDouble(interp, argv[n+6], &ntile[n])!= TCL_OK))
		return (TCL_ERROR);
	}

/* allocate storage for the working arrays */
   nbin = 65536;
   pixval = (int *)shMalloc(nbin * sizeof (int) );
   bin = (int *)shMalloc (nbin * sizeof (int) );

/* Now do the real work */

   if (imgHist(reg, row1, col1, row2, col2, 0, nbin-1,
        skip, pixval, bin) !=0) return (TCL_ERROR);

/* How many pixels total? */
   ntot = 0;
   for (i=0; i<nbin; i++) {
	ntot = ntot + bin[i];
	}
   if (ntot == 0) return TCL_ERROR;

   Tcl_ResetResult (interp);
/* Now compute the ntile values */
   for (n=0; n<nntile; n++) {
	sum = 0;
	for (i=0; i<nbin; i++) {
	   sum = sum + bin[i];
	   if (sum >= ntot * ntile[n]) {
		fract = 1. + (ntile[n] * ntot - sum) / bin[i];
		ntileval[n] = i - 0.5 + fract;
		break;
		}
	   }
	}

/* Create the return string as a list */

   for (n=0; n<nntile; n++) {
	sprintf (item, "%-10.2f ", ntileval[n]);
	Tcl_AppendResult (interp, item, NULL);
	}

   shFree(bin);
   shFree(pixval);   
   shFree (ntileval);
   shFree (ntile);
   return (TCL_OK);
}

/************ declarations of TCL verbs for this module ***********/

void imgTclHistDeclare (Tcl_Interp *interp) {

   shTclDeclare(interp,"imgStats",
                (Tcl_CmdProc *)tclStats,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclStats_hlp, tclStats_use);

   shTclDeclare(interp,"imgHist",
                (Tcl_CmdProc *)tclHist,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclHist_hlp, tclHist_use);


   shTclDeclare(interp,"imgNtile",
                (Tcl_CmdProc *)tclNtile,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclNtile_hlp, tclNtile_use);

   return;
}
