/******************************************************************************
**
** FILE:
**	imgTclLstsq.c
**
** ABSTRACT:
**	This file contains the tcl interface to LSTSQ routines
**
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "dervish.h"

#include "imgProto.h"
#include "imgLstsq.h"

static char *module = "img";	/* name of this set of code */

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclLstsqNew
**
** DESCRIPTION:
** 	Return a handle to a new LSTSQ
**
**============================================================================
*/

static char *imgTclLstsqNew_use =
  "USAGE: lstsqNew <nparam>";
static char *imgTclLstsqNew_hlp =
  "create a new LSTSQ structure";

static int
imgTclLstsqNew(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   char name[HANDLE_NAMELEN];
   LSTSQ *lstsq = NULL;
   int nparam;
   int j, k, l;

   shErrStackClear();
   
   if(argc != 2) {
      Tcl_SetResult(interp,imgTclLstsqNew_use,TCL_STATIC);
      return(TCL_ERROR);
   }

   if (Tcl_GetInt (interp, argv[1], &nparam) != TCL_OK) {
	return TCL_ERROR;
	}

/*
 * Get a handle for our new LSTSQ
 */


   lstsq = shMalloc(sizeof(LSTSQ));
   if (shTclHandleNew (interp, name, "LSTSQ", (void *)lstsq) != TCL_OK) {
	shFree(lstsq);
        return(TCL_ERROR);
        }

   lstsq->nparam = nparam;
   lstsq->cmat = shMalloc(nparam*(nparam+1)/2*sizeof(double));
   lstsq->rhs = shMalloc(nparam*sizeof(double));
   lstsq->duma = shMalloc(nparam*sizeof(double));
   lstsq->dumb = shMalloc(nparam*sizeof(double));
   lstsq->param = shMalloc(nparam*sizeof(double));
   lstsq->stdev = shMalloc(nparam*sizeof(double));
   lstsq->dumc = shMalloc(nparam*sizeof(int));

   l = 0;
   for (j=0; j<nparam; j++) {
	lstsq->rhs[j] = lstsq->duma[j] = lstsq->dumb[j] = 0.;
	lstsq->dumc[j] = 0;
	lstsq->param[j]=0.;
	lstsq->stdev[j]=0.;
	for (k=0; k<= j; k++) {
	   lstsq->cmat[l] = 0.;
	   l++;
	   }
	}
   Tcl_SetResult(interp,name,TCL_VOLATILE);
   return(TCL_OK);
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclLstsqDel
**
** DESCRIPTION:
**	Delete a LSTSQ and its handle
**
**============================================================================
*/

/*****************************************************************************/

static char *imgTclLstsqDel_use =
  "USAGE: lstsqDel <lstsq-handle>";
static char *imgTclLstsqDel_hlp =
  "Delete a LSTSQ";

static int
imgTclLstsqDel(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   LSTSQ *lstsq;

   shErrStackClear();
   
   if (argc!= 2) {
	Tcl_SetResult (interp, imgTclLstsqDel_use, TCL_STATIC);
	return TCL_ERROR;
	}

   if (shTclAddrGetFromName (interp, argv[1], (void *)&lstsq, "LSTSQ")
                != TCL_OK) return(TCL_ERROR);

   shFree(lstsq->cmat);
   shFree(lstsq->rhs);
   shFree(lstsq->duma);
   shFree(lstsq->dumb);
   shFree(lstsq->param);
   shFree(lstsq->stdev);
   shFree(lstsq->dumc);
   shFree(lstsq);

   p_shTclHandleDel(interp, argv[1]);

   return(TCL_OK);
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclLstsqReset
**
** DESCRIPTION:
**	Reset a LSTSQ (clear out cmat and rhs arrays)
**
**============================================================================
*/

/*****************************************************************************/

static char *imgTclLstsqReset_use =
  "USAGE: lstsqReset <lstsq-handle>";
static char *imgTclLstsqReset_hlp =
  "Reset a LSTSQ";

static int
imgTclLstsqReset(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   LSTSQ *lstsq;
   int j, k, l;

   shErrStackClear();
   
   if (argc!= 2) {
	Tcl_SetResult (interp, imgTclLstsqReset_use, TCL_STATIC);
	return TCL_ERROR;
	}

   if (shTclAddrGetFromName (interp, argv[1], (void *)&lstsq, "LSTSQ")
                != TCL_OK) return(TCL_ERROR);

   l = 0;
   for (j=0; j<lstsq->nparam; j++) {
	lstsq->rhs[j] = 0.;
	for (k=0; k<= j; k++) {
	   lstsq->cmat[l] = 0.;
	   l++;
	   }
	}

   return(TCL_OK);
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclDerivNew
**
** DESCRIPTION:
** 	Return a handle to a new DERIV
**
**============================================================================
*/

static char *imgTclDerivNew_use =
  "USAGE: derivNew <nparam>";
static char *imgTclDerivNew_hlp =
  "create a new DERIV structure";

static int
imgTclDerivNew(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   char name[HANDLE_NAMELEN];
   DERIV *deriv = NULL;
   int nparam;
   int j;

   shErrStackClear();
   
   if(argc != 2) {
      Tcl_SetResult(interp,imgTclDerivNew_use,TCL_STATIC);
      return(TCL_ERROR);
   }

   if (Tcl_GetInt (interp, argv[1], &nparam) != TCL_OK) {
	return TCL_ERROR;
	}

/*
 * Get a handle for our new DERIV
 */


   deriv = shMalloc(sizeof(DERIV));
   if (shTclHandleNew (interp, name, "DERIV", (void *)deriv) != TCL_OK) {
	shFree(deriv);
        return(TCL_ERROR);
        }

   deriv->nparam = nparam;
   deriv->deriv = shMalloc(nparam*sizeof(double));
   deriv->index = shMalloc(nparam*sizeof(int));
   for (j=0; j<nparam; j++) {
	deriv->deriv[j] = 0.;
	deriv->index[j] = j;
	}
   deriv->res = 0.;
   deriv->weight = 1.;

   Tcl_SetResult(interp,name,TCL_VOLATILE);
   return(TCL_OK);
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclDerivDel
**
** DESCRIPTION:
**	Delete a DERIV and its handle
**
**============================================================================
*/

/*****************************************************************************/

static char *imgTclDerivDel_use =
  "USAGE: derivDel <deriv-handle>";
static char *imgTclDerivDel_hlp =
  "Delete a deriv";

static int
imgTclDerivDel(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   DERIV *deriv;

   shErrStackClear();
   
   if (argc!= 2) {
	Tcl_SetResult (interp, imgTclLstsqDel_use, TCL_STATIC);
	return TCL_ERROR;
	}

   if (shTclAddrGetFromName (interp, argv[1], (void *)&deriv, "DERIV")
                != TCL_OK) return(TCL_ERROR);

   shFree(deriv->deriv);
   shFree(deriv->index);
   shFree(deriv);

   p_shTclHandleDel(interp, argv[1]);

   return(TCL_OK);
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclLstsqInc
**
** DESCRIPTION:
** 	Increment a least squares structure with partial derivs and
**		residuals.
**
**============================================================================
*/
/* Inputs are handles to LSTSQ and DERIV structures.
 */

static char *imgTclLstsqInc_use =
  "USAGE: imgLstsqInc <lstsq> <deriv>";
static char *imgTclLstsqInc_hlp =
  "Increment normal equations";

static int 
imgTclLstsqInc(
	   ClientData clientData,
	   Tcl_Interp *interp,
	   int argc,
	   char **argv
	   )

{

   LSTSQ *lstsq = NULL;
   DERIV *deriv = NULL;
   int j, k, l;
   int indexj, indexk;

	if(argc != 3) 
              {
		Tcl_SetResult(interp,imgTclLstsqInc_use,TCL_VOLATILE);
		return(TCL_ERROR);
	      }

/* Decode LSTSQ name */
	if (shTclAddrGetFromName (interp, argv[1], (void **)(&lstsq),
	   "LSTSQ") != TCL_OK) return TCL_ERROR;

/* Decode DERIV name */
	if (shTclAddrGetFromName (interp, argv[2], (void **)(&deriv),
	   "DERIV") != TCL_OK) return TCL_ERROR;

   for (j = 0; j < deriv->nparam; j++) {
	indexj = deriv->index[j];
	lstsq->rhs[indexj] += deriv->weight * deriv->res * deriv->deriv[j];
	for (k = 0; k <= j; k++) {
	   indexk = deriv->index[k];
	   l = indexj*(indexj+1)/2 + indexk;
	   lstsq->cmat[l] += deriv->weight * deriv->deriv[j] * \
		deriv->deriv[k];
	   }
	}

   Tcl_ResetResult (interp);
   return(TCL_OK);
}
/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclLstsqInvert
**
** DESCRIPTION:
** 	Increment a least squares structure with partial derivs and
**		residuals.
**
**============================================================================
*/
/* Inputs are handles to LSTSQ.
 */

static char *imgTclLstsqInvert_use =
  "USAGE: imgLstsqInvert <lstsq> <deriv>";
static char *imgTclLstsqInvert_hlp =
  "Invert normal equations";

static int 
imgTclLstsqInvert(
	   ClientData clientData,
	   Tcl_Interp *interp,
	   int argc,
	   char **argv
	   )

{

   LSTSQ *lstsq = NULL;
   int j, l;
   int ierr;

	if(argc != 2) 
              {
		Tcl_SetResult(interp,imgTclLstsqInvert_use,TCL_VOLATILE);
		return(TCL_ERROR);
	      }

/* Decode LSTSQ name */
	if (shTclAddrGetFromName (interp, argv[1], (void **)(&lstsq),
	   "LSTSQ") != TCL_OK) return TCL_ERROR;

/* Check for zeros on diagonal.  Set to 1 if found */
   for (j=0; j<lstsq->nparam; j++) {
	l = (j+2)*(j+1)/2 - 1;
	if (lstsq->cmat[l] == 0.) lstsq->cmat[l] = 1.;
	}

   ierr = syminv(lstsq->cmat, lstsq->rhs, lstsq->nparam, lstsq->duma, \
		lstsq->dumb,
	lstsq->dumc);

   if (ierr != 1) {
	return (TCL_ERROR);
	}

   for (j=0; j<lstsq->nparam; j++) {
	lstsq->param[j] += lstsq->rhs[j];
	l = (j+2)*(j+1)/2 - 1;
	lstsq->stdev[j] = sqrt(lstsq->cmat[l]);
	}

   return(TCL_OK);
}

/***************************************************************************/

/* TclLstsqDeclare
**
** DESCRIPTION:
** 	Declare new tcl verbs
**
**============================================================================
*/

void
imgTclLstsqDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"lstsqNew", 
		(Tcl_CmdProc *)imgTclLstsqNew, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclLstsqNew_hlp, imgTclLstsqNew_use);

   shTclDeclare(interp,"lstsqDel", 
		(Tcl_CmdProc *)imgTclLstsqDel,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclLstsqDel_hlp, imgTclLstsqDel_use);

   shTclDeclare(interp,"lstsqReset", 
		(Tcl_CmdProc *)imgTclLstsqReset,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclLstsqReset_hlp, imgTclLstsqReset_use);

   shTclDeclare(interp,"derivNew", 
		(Tcl_CmdProc *)imgTclDerivNew, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclDerivNew_hlp, imgTclDerivNew_use);

   shTclDeclare(interp,"derivDel", 
		(Tcl_CmdProc *)imgTclDerivDel,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclDerivDel_hlp, imgTclDerivDel_use);

   shTclDeclare(interp,"lstsqInc", 
		(Tcl_CmdProc *)imgTclLstsqInc, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclLstsqInc_hlp, imgTclLstsqInc_use);

   shTclDeclare(interp,"lstsqInvert", 
		(Tcl_CmdProc *)imgTclLstsqInvert,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclLstsqInvert_hlp, imgTclLstsqInvert_use);

   return;
} 
