/*
 * TCL support for image utilities
 *
 * ENTRY POINT		SCOPE
 * imgTclMbsDeclare	public	declare all the verbs defined in this module
 * tclMbs		static	simulate bright stars
 * tclNoise		static	add noise to a region
 * tclPhoton		static	add photon noise to a region
 *
 * REQUIRED PRODUCTS:
 *      DERVISH
 *
 */
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "shGeneric.h"
#include "imgStar.h"
#include "img_msg_c.h"
#include "imgProto.h"

static char *module = "img";		/* name of this facility */

/*****************************************************************************/
/*************************** tclMbs     **************************************/
/*****************************************************************************/
/*
 * Input is handle to image and a threshold
 */

static char *tclMbs_use =
  "USAGE: imgMbs <region-handle> row col sig ncount [photon/count]";
static char *tclMbs_hlp =
  "Simulate gaussian star profiles including photon statistics";

static int
tclMbs(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   double rowc, colc;
   double sig, count;
   double gain = 1.;

   shErrStackClear();

   if(argc != 6 && argc != 7) {
      Tcl_SetResult(interp,tclMbs_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclMbs_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the star center */

   if (Tcl_GetDouble (interp, argv[2], &rowc)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &colc)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &sig)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[5], &count)  != TCL_OK) return(TCL_ERROR);
   if (argc == 7) {
	if (Tcl_GetDouble (interp, argv[6], &gain)  != TCL_OK) return(TCL_ERROR);
	}

/* Done with the handles, etc. Now the action begins */

   if (imgMbs (reg, rowc, colc, sig, count, gain) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in MBS", TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Null result */
   Tcl_ResetResult (interp);
   return(TCL_OK);
}

/*****************************************************************************/
/*************************** tclNoise     **************************************/
/*****************************************************************************/
/*
 * Input is handle to image and an rms
 */

static char *tclNoise_use =
  "USAGE: imgNoise <region-handle> rms";
static char *tclNoise_hlp =
  "Add gaussian noise to a region";

static int
tclNoise(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   double rms;

   shErrStackClear();

   if(argc != 3) {
      Tcl_SetResult(interp,tclNoise_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclNoise_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the rms */

   if (Tcl_GetDouble (interp, argv[2], &rms)  != TCL_OK) return(TCL_ERROR);

/* Done with the handles, etc. Now the action begins */

   if (imgNoise (reg, rms) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in imgNoise", TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Null result */
   Tcl_ResetResult (interp);
   return(TCL_OK);
}


/****************************************************************************/
/*************************** tclPhoton **************************************/
/****************************************************************************/
/*
 * Input is handle to image and a gain
 */

static char *tclPhoton_use =
  "USAGE: imgPhoton <region-handle> gain";
static char *tclPhoton_hlp =
  "Add photon noise to a region";

static int
tclPhoton(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg = NULL;
   double gain;

   shErrStackClear();

   if(argc != 3) {
      Tcl_SetResult(interp,tclPhoton_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION")
	!= TCL_OK) {
	Tcl_SetResult(interp,tclPhoton_use,TCL_STATIC);
	return (TCL_ERROR);
	}

/* Get the rms */

   if (Tcl_GetDouble (interp, argv[2], &gain)  != TCL_OK) return(TCL_ERROR);

/* Done with the handles, etc. Now the action begins */

   if (imgPhoton (reg, gain) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in imgPhoton", TCL_STATIC);
	return(TCL_ERROR);
	}
   
/* Null result */
   Tcl_ResetResult (interp);
   return(TCL_OK);
}


/*****************************************************************************/
/*************************** tclReseed*********************************/
/*****************************************************************************/
/*
 * Reseed the random number generator.
 */

static char *tclReseed_use =
  "USAGE: imgReseed seed";
static char *tclReseed_hlp =
  "Reseed the random number generator";

static int
tclReseed(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   int seed;

   shErrStackClear();

   if(argc != 2) {
      Tcl_SetResult(interp,tclPhoton_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the seed */

   if (Tcl_GetInt (interp, argv[1], &seed)  != TCL_OK) return(TCL_ERROR);

   reseed (seed);
   
/* Null result */
   Tcl_ResetResult (interp);
   return(TCL_OK);
}


/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclMbsDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"imgMbs", 
		(Tcl_CmdProc *)tclMbs,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclMbs_hlp, tclMbs_use);

   shTclDeclare(interp,"imgNoise", 
		(Tcl_CmdProc *)tclNoise,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclNoise_hlp, tclNoise_use);

   shTclDeclare(interp,"imgPhoton", 
		(Tcl_CmdProc *)tclPhoton,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclPhoton_hlp, tclPhoton_use);

   shTclDeclare(interp,"imgReseed", 
		(Tcl_CmdProc *)tclReseed,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclReseed_hlp, tclReseed_use);

   return;
}

