/*
 * TCL support to make polynomial fit.
 *
 * Commands defined in this file:
 *
 *	polyfit
 *	polycomp
 *
 */

#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "imgProto.h"

static char *module = "tclPolyfit";

/*****************************************************************************/

static char *tclPolyfit_use = "Usage: polyFit {xlist} {ylist} nterm";
static char *tclPolyfit_hlp = "Fit a polynomial of nterm terms";

static int
tclPolyfit(
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   int splitArgc;
   char **splitArgv;
   double *x, *y;
   double *coeff;
   int nelem, ndim;
   int i;
   char buffer[80];

   if(argc != 4) {
	Tcl_SetResult(interp,tclPolyfit_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Input is x-list, y-list, and number of terms.
*/

/* Decode x */
   Tcl_SplitList (interp, argv[1], &splitArgc, &splitArgv);
   if (splitArgc <= 0) {
	Tcl_SetResult (interp, "Zero elements in x list", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   x = (double *)shMalloc (splitArgc * sizeof (double));
   nelem=splitArgc;
   for (i=0; i<nelem; i++) {
	if (sscanf(splitArgv[i], "%lf", &x[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i], NULL);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

/* Decode y */
   Tcl_SplitList (interp, argv[2], &splitArgc, &splitArgv);
   if (splitArgc != nelem) {
	Tcl_SetResult (interp, \
"Number of elements in x and y lists don't match", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   y = (double *)shMalloc (splitArgc * sizeof (double));
   for (i=0; i<nelem; i++) {
	if (sscanf(splitArgv[i], "%lf", &y[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i]);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);
   if (Tcl_GetInt (interp, argv[3], &ndim) != TCL_OK) return TCL_ERROR;
   coeff = (double *)shMalloc (ndim * sizeof(double));
   if (polyfit (x, y, nelem, ndim, coeff) != 1) {
	Tcl_SetResult (interp, "Polyfit failed", TCL_STATIC);
	return TCL_ERROR;
	}
   for (i=0; i<ndim; i++) {
	sprintf (buffer, "%g", coeff[i]);
	Tcl_AppendElement (interp, buffer);
	}
   shFree (x);
   shFree (y);
   shFree (coeff);
   return (TCL_OK);
   }

/*****************************************************************************/

static char *tclPolycomp_use = "Usage: polycomp {xlist} {ylist} {coefflist}";
static char *tclPolycomp_hlp = "Compute residuals of a fit given a coeff list";

static int
tclPolycomp(
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   int splitArgc;
   char **splitArgv;
   double *x, *y;
   double *coeff;
   double *res;
   int nelem, ndim;
   int i;
   char buffer[80];

   if(argc != 4) {
	Tcl_SetResult(interp,tclPolycomp_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Input is x-list, y-list, and coeff-list
*/

/* Decode x */
   Tcl_SplitList (interp, argv[1], &splitArgc, &splitArgv);
   if (splitArgc <= 0) {
	Tcl_SetResult (interp, "Zero elements in x list", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   x = (double *)shMalloc (splitArgc * sizeof (double));
   nelem=splitArgc;
   for (i=0; i<nelem; i++) {
	if (sscanf(splitArgv[i], "%lf", &x[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i], NULL);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

/* Decode y */
   Tcl_SplitList (interp, argv[2], &splitArgc, &splitArgv);
   if (splitArgc != nelem) {
	Tcl_SetResult (interp, \
"Number of elements in x and y lists don't match", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   y = (double *)shMalloc (splitArgc * sizeof (double));
   for (i=0; i<nelem; i++) {
	if (sscanf(splitArgv[i], "%lf", &y[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i]);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

/* Decode coeff */
   Tcl_SplitList (interp, argv[3], &splitArgc, &splitArgv);
   if (splitArgc <= 0) {
	Tcl_SetResult (interp, "Zero elements in coeff list", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   coeff = (double *)shMalloc (splitArgc * sizeof (double));
   ndim = splitArgc;
   for (i=0; i<ndim; i++) {
	if (sscanf(splitArgv[i], "%lf", &coeff[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i], NULL);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

   res = (double *)shMalloc (nelem * sizeof(double));

/* Gak, in case of error, we have memory leak! */
   if (polycomp (x, y, nelem, ndim, coeff, res) != 1) {
	Tcl_SetResult (interp, "Polyfit failed", TCL_STATIC);
	return TCL_ERROR;
	}
   for (i=0; i<nelem; i++) {
	sprintf (buffer, "%g", res[i]);
	Tcl_AppendElement (interp, buffer);
	}
   shFree (x);
   shFree (y);
   shFree (coeff);
   shFree (res);
   return (TCL_OK);
   }

/*****************************************************************************/

static char *tclPolyrms_use = "Usage: polyrms {reslist}";
static char *tclPolyrms_hlp = "Compute residuals from a residual list";

static int
tclPolyrms(
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   int splitArgc;
   char **splitArgv;
   double *res;
   double rms;
   int nelem;
   int i;
   char buffer[80];

   if(argc != 2) {
	Tcl_SetResult(interp,tclPolyrms_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Input is reslist
*/

/* Decode reslist*/
   Tcl_SplitList (interp, argv[1], &splitArgc, &splitArgv);
   if (splitArgc <= 0) {
	Tcl_SetResult (interp, "Zero elements in x list", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   res = (double *)shMalloc (splitArgc * sizeof (double));
   nelem=splitArgc;
   for (i=0; i<nelem; i++) {
	if (sscanf(splitArgv[i], "%lf", &res[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i], NULL);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

   if (polyrms (res, nelem, &rms) != 1) {
	Tcl_SetResult (interp, "Polyrms failed", TCL_STATIC);
	return TCL_ERROR;
	}
   sprintf (buffer, "%f", rms);
   Tcl_AppendElement (interp, buffer);
   shFree (res);
   return (TCL_OK);
   }

/*****************************************************************************/

static char *tclPolyeval_use = "Usage: polyeval {xlist} {coefflist}";
static char *tclPolyeval_hlp = "Evaluate a polynmial at list of points xlist";

static int
tclPolyeval(
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   int splitArgc;
   char **splitArgv;
   double *coeff;
   double *x;
   double val;
   int ndim, nelem;
   int i;
   char buffer[80];
   int errflag = 0;

   if(argc != 3) {
	Tcl_SetResult(interp,tclPolyeval_use,TCL_STATIC);
	return(TCL_ERROR);
	}

/* Input is x-list and coefflist
*/

/* Decode x */
   Tcl_SplitList (interp, argv[1], &splitArgc, &splitArgv);
   if (splitArgc <= 0) {
	Tcl_SetResult (interp, "Zero elements in x list", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   x = (double *)shMalloc (splitArgc * sizeof (double));
   nelem=splitArgc;
   for (i=0; i<nelem; i++) {
	if (sscanf(splitArgv[i], "%lf", &x[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i], NULL);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

/* Decode coefflist*/
   Tcl_SplitList (interp, argv[2], &splitArgc, &splitArgv);
   if (splitArgc <= 0) {
	Tcl_SetResult (interp, "Zero elements in x list", TCL_STATIC);
	free(splitArgv);
	return TCL_ERROR;
	}
   coeff = (double *)shMalloc (splitArgc * sizeof (double));
   ndim=splitArgc;
   for (i=0; i<ndim; i++) {
	if (sscanf(splitArgv[i], "%lf", &coeff[i]) == 0) {
	   Tcl_AppendResult (interp, "Invalid number ",splitArgv[i], NULL);
	   free(splitArgv);
	   return TCL_ERROR;
	   }
	}
   free(splitArgv);

   Tcl_ResetResult(interp);
   for (i=0; i<nelem; i++) {
 	if (polyeval (&val, x[i], coeff, ndim) != 1) {
	   Tcl_SetResult (interp, "Polyeval failed", TCL_STATIC);
	   errflag=1;
	   break;
	   }
	sprintf (buffer, "%f", val);
	Tcl_AppendElement (interp, buffer);
	}
   shFree (x);
   shFree (coeff);
   if (errflag) return (TCL_ERROR);
   return (TCL_OK);
   }

/*****************************************************************************/

/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclPolyfitDeclare (Tcl_Interp *interp) 
{
   shTclDeclare(interp,"polyfit", 
		(Tcl_CmdProc *)tclPolyfit,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclPolyfit_hlp, tclPolyfit_use);

   shTclDeclare(interp,"polycomp", 
		(Tcl_CmdProc *)tclPolycomp,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclPolycomp_hlp, tclPolycomp_use);

   shTclDeclare(interp,"polyrms", 
		(Tcl_CmdProc *)tclPolyrms,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclPolyrms_hlp, tclPolyrms_use);

   shTclDeclare(interp,"polyeval", 
		(Tcl_CmdProc *)tclPolyeval,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclPolyeval_hlp, tclPolyeval_use);

   return;
}

