/******************************************************************************
**
** FILE:
**	imgTclProfile.c
**
** ABSTRACT:
**	This file contains the tcl interface to PROFILE routines
**
** ENTRY POINT		SCOPE	DESCRIPTION
** imgTclProfileDeclare public declare all the verbs defined in this module
** -------------------------------------------------------------------------
**
** ENVIRONMENT:
**	ANSI C.
**
** REQUIRED PRODUCTS:
**	dervish
**	ftcl
**
** AUTHORS:
**	Creation date:  2-Feb_1994
**	Mario Nonino
**
******************************************************************************
******************************************************************************
*/

#include <stdio.h>
#include <string.h>
#include "dervish.h"

#include "imgProto.h"

/*============================================================================  
**============================================================================
**
** LOCAL MACROS, DEFINITIONS, ETC.
**
**============================================================================
*/

/*----------------------------------------------------------------------------
**
** GLOBAL VARIABLES
*/
static char *module = "img";	/* name of this set of code */


/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclProfileNew
**
** DESCRIPTION:
** 	Return a handle to a new PROFILE
**
** RETURN VALUES:
**	TCL_OK    -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	Various tcl and dervish routines
**	imgProfileNew
**
**============================================================================
*/

static char *imgTclProfileNew_use =
  "USAGE: profileNew";
static char *imgTclProfileNew_hlp =
  "create a new PROFILE";

static int
imgTclProfileNew(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   char name[HANDLE_NAMELEN];
   PROFILE *profile;
   
   shErrStackClear();
   
   if(argc != 1) {
      Tcl_SetResult(interp,imgTclProfileNew_use,TCL_STATIC);
      return(TCL_ERROR);
   }
/*
 * Get a handle for our new PROFILE
 */


   if (shTclHandleNew (interp, name, "PROFILE",
        (void *)(profile=imgProfileNew())) != TCL_OK) {
        if (profile != NULL) imgProfileDel(profile);
        return(TCL_ERROR);
        }
   
   Tcl_SetResult(interp,name,TCL_VOLATILE);
   return(TCL_OK);
}


/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclProfileDel
**
** DESCRIPTION:
**	Delete a PROFILE and its handle
**
** RETURN VALUES:
**	TCL_OK    -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	Various tcl and dervish routines
**	imgProfileDel
**
**============================================================================
*/

/*****************************************************************************/

static char *imgTclProfileDel_use =
  "USAGE: profileDel <profile-handle>";
static char *imgTclProfileDel_hlp =
  "Delete a PROFILE";

static int
imgTclProfileDel(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   PROFILE *profile;
   
   shErrStackClear();
   
   if (argc!= 2) {
	Tcl_SetResult (interp, imgTclProfileDel_use, TCL_STATIC);
	return TCL_ERROR;
	}

   if (shTclAddrGetFromName (interp, argv[1], (void *)&profile, "PROFILE")
                != TCL_OK) return(TCL_ERROR);

   imgProfileDel(profile);

/* Modification by S. Kent Aug 2, 1993 delete handle */

   p_shTclHandleDel(interp, argv[1]);

   return(TCL_OK);
}

/*============================================================================  
**============================================================================
**
** ROUTINE: imgTclProfile
**
** DESCRIPTION:
** 	Declare new tcl verbs : interface to routine to find profile
**
** RETURN VALUES:
**	TCL_OK    -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	imgRegProfile
**
**============================================================================
*/
/* Input is handle to region nsteps nderiv handle to centroid handle to sky
 * return handle to: profilelis
 */

static char *imgTclProfile_use =
  "USAGE: imgProfile <region-handle> <axes-handle> start end nderiv sky";
static char *imgTclProfile_hlp =
  "Compute elliptical profile";

static int 
imgTclProfile(
	   ClientData clientData,
	   Tcl_Interp *interp,
	   int argc,
	   char **argv
	   )

{
	char listName[HANDLE_NAMELEN];

	REGION   *reg;
	AXES *centroid;
	double sky;
	CHAIN *profilelist;

	int nderiv=0;
	int start_pixel =0;
	int end_pixel =0;


	if(argc != 7) 
              {
		Tcl_SetResult(interp,imgTclProfile_use,TCL_VOLATILE);
		return(TCL_ERROR);
	      }

/* Decode REGION name */
	if (shTclAddrGetFromName (interp, argv[1], (void **)(&reg), "REGION")
	   != TCL_OK) return TCL_ERROR;

/* Decode AXES name */
	if (shTclAddrGetFromName (interp, argv[2], (void **)(&centroid),
	   "AXES") != TCL_OK) return TCL_ERROR;

/* Get various parameters */
	if (Tcl_GetInt (interp, argv[3], &start_pixel) != TCL_OK)
	   return TCL_ERROR;
	if (Tcl_GetInt (interp, argv[4], &end_pixel) != TCL_OK)
	   return TCL_ERROR;
	if (Tcl_GetInt (interp, argv[5], &nderiv) != TCL_OK)
	   return TCL_ERROR;
	if (Tcl_GetDouble (interp, argv[6], &sky) != TCL_OK)
	   return TCL_ERROR;

/* get a handle for our new PROFILE list 
 */
if(shTclChainHandleNew(interp,listName,"PROFILE",&profilelist) != TCL_OK)
	return(TCL_ERROR);

/* true action ... */
   if (imgRegProfile(reg, profilelist, centroid, start_pixel, end_pixel,
	nderiv, sky) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in Profile routine",TCL_VOLATILE);
	return(TCL_ERROR);
	}

/* return handle as the result */
	Tcl_AppendResult(interp,listName,(char *)NULL);
	return(TCL_OK);
}

/* TclProfileDeclare
**
** DESCRIPTION:
** 	Declare new tcl verbs
**
** RETURN VALUES:
**	TCL_OK    -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	Various tcl and dervish routines
**
**============================================================================
*/

void
imgTclProfileDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"profileNew", 
		(Tcl_CmdProc *)imgTclProfileNew, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclProfileNew_hlp, imgTclProfileNew_use);

   shTclDeclare(interp,"profileDel", 
		(Tcl_CmdProc *)imgTclProfileDel, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclProfileDel_hlp, imgTclProfileDel_use);

   shTclDeclare(interp,"imgProfile", 
		(Tcl_CmdProc *)imgTclProfile, 
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, imgTclProfile_hlp, imgTclProfile_use);
} 
