/*****************************************************************************
******************************************************************************
**
** FILE:
**      TclRowCol.c
**
** ABSTRACT:
**      This file contains top level functions to perform various row and
**	column oriented operations on a region
**
** ENTRY POINT          SCOPE   DESCRIPTION
** -------------------------------------------------------------------------
** imgTclRowColDeclare	public  declare all the verbs defined in this module
** tclRowMed		static  Compute flatfield vector by medianing
** tclColMed		static  Compute flatfield vector by medianing
** tclRowAvg		static  Compute bias vector by averaging
** tclColAvg		static  Compute bias vector by averaging
** tclRowSub		static  Subtract bias vector
** tclColSub		static  Subtract bias vector
** tclConAdd		static  Add a constant
**
** ENVIRONMENT:
**      ANSI C.
**
** REQUIRED PRODUCTS:
**      FTCL            TCL + Fermi extension
**      DERVISH
**      IMAGE
**
** AUTHORS:
**      Creation date:  Oct. 1993
**      Steve Kent
**
******************************************************************************
******************************************************************************
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "dervish.h"
#include "img_msg_c.h"
#include "imgProto.h"

static char *module="img";	/* Name of this facility */

/***************************************************************************/
/**************************	tclRowMed	****************************/
/***************************************************************************/
/* Compute the histogram of pixel distributions and extract the
** value of the median.  Do this row by row.
*/

static char *tclRowMed_use = "USAGE: imgRowMed reg-in reg-out";

static char *tclRowMed_hlp = "Compute median flatfield vector";

static int tclRowMed (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   REGION *regIn = NULL;
   REGION *regOut = NULL;

/* First check to see that we have right number of arguments. */
   if (argc < 3) {
      Tcl_SetResult(interp, tclRowMed_use, TCL_VOLATILE);
      return (TCL_ERROR);
   }

   if (shTclAddrGetFromName(interp, argv[1], (void *)&regIn, "REGION")
	!= TCL_OK) return (TCL_ERROR);
   if (shTclAddrGetFromName(interp, argv[2], (void *)&regOut, "REGION")
	!= TCL_OK) return (TCL_ERROR);

/* Check for dimension compatibility */
   if (regOut->ncol != 1) {
	Tcl_SetResult (interp, "Output row dimension must be 1", TCL_STATIC);
	return TCL_ERROR;
	}
   if (regOut->nrow != regIn->nrow) {
	Tcl_SetResult (interp, "Row dimensions do not match", TCL_STATIC);
	return TCL_ERROR;
	}
   if (imgRowMed (regIn, regOut) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Medianing operation unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclColMed	****************************/
/***************************************************************************/
/* Compute the histogram of pixel distributions and extract the
** value of the median.  Do this column by column.
*/

static char *tclColMed_use = "USAGE: imgColMed reg-in reg-out";

static char *tclColMed_hlp = "Compute median flatfield vector";

static int tclColMed (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   REGION *regIn = NULL;
   REGION *regOut = NULL;

/* First check to see that we have right number of arguments. */
   if (argc < 3) {
      Tcl_SetResult(interp, tclColMed_use, TCL_VOLATILE);
      return (TCL_ERROR);
   }

   if (shTclAddrGetFromName(interp, argv[1], (void *)&regIn, "REGION")
	!= TCL_OK) return (TCL_ERROR);
   if (shTclAddrGetFromName(interp, argv[2], (void *)&regOut, "REGION")
	!= TCL_OK) return (TCL_ERROR);

/* Check for dimension compatibility */
   if (regOut->nrow != 1) {
	Tcl_SetResult (interp, "Output row dimension must be 1", TCL_STATIC);
	return TCL_ERROR;
	}
   if (regOut->ncol != regIn->ncol) {
	Tcl_SetResult (interp, "Column dimensions do not match", TCL_STATIC);
	return TCL_ERROR;
	}
   if (imgColMed (regIn, regOut) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Medianing operation unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclRowAvg	****************************/
/***************************************************************************/
/* Compute the average of pixel distributions row by row. */

static char *tclRowAvg_use = "USAGE: imgRowAvg reg-in reg-out";

static char *tclRowAvg_hlp = "Compute average vector row by row";

static int tclRowAvg (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   REGION *regIn = NULL, *regOut = NULL;

/* First check to see that we have right number of arguments. */
   if (argc < 3) {
      Tcl_SetResult(interp, tclRowAvg_use, TCL_VOLATILE);
      return (TCL_ERROR);
   }

   if (shTclAddrGetFromName(interp, argv[1], (void *)&regIn, "REGION")
	!= TCL_OK) return (TCL_ERROR);
   if (shTclAddrGetFromName(interp, argv[2], (void *)&regOut, "REGION")
	!= TCL_OK) return (TCL_ERROR);

/* Check for dimension compatibility */
   if (regOut->ncol != 1) {
	Tcl_SetResult (interp, "Output col dimension must be 1", TCL_STATIC);
	return TCL_ERROR;
	}
   if (regOut->nrow != regIn->nrow) {
	Tcl_SetResult (interp, "Row dimensions do not match", TCL_STATIC);
	return TCL_ERROR;
	}
   if (imgRowAvg (regIn, regOut) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Row averaging operation unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclColAvg	****************************/
/***************************************************************************/
/* Compute the average of pixel distributions column by column. */

static char *tclColAvg_use = "USAGE: imgColAvg reg-in reg-out";

static char *tclColAvg_hlp = "Compute average vector col by col";

static int tclColAvg (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   REGION *regIn = NULL;
   REGION *regOut = NULL;

/* First check to see that we have right number of arguments. */
   if (argc < 3) {
      Tcl_SetResult(interp, tclColAvg_use, TCL_VOLATILE);
      return (TCL_ERROR);
   }

   if (shTclAddrGetFromName(interp, argv[1], (void *)&regIn, "REGION")
	!= TCL_OK) return (TCL_ERROR);
   if (shTclAddrGetFromName(interp, argv[2], (void *)&regOut, "REGION")
	!= TCL_OK) return (TCL_ERROR);

/* Check for dimension compatibility */
   if (regOut->nrow != 1) {
	Tcl_SetResult (interp, "Output row dimension must be 1", TCL_STATIC);
	return TCL_ERROR;
	}
   if (regOut->ncol != regIn->ncol) {
	Tcl_SetResult (interp, "Column dimensions do not match", TCL_STATIC);
	return TCL_ERROR;
	}
   if (imgColAvg (regIn, regOut) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Row averaging operation unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclRowSub	****************************/
/***************************************************************************/
/* Subtract a vector from an image row by row. */


static char *g_imgRowSub = "imgRowSub";
static ftclArgvInfo g_imgRowSubTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL,
	"Subtract a vector row by row from an image"},
    {"<REGION>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to image"},
    {"<REGROW>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to row vector"},
    {"-dither",   FTCL_ARGV_CONSTANT, (void *)1, NULL,
	"Set if we dither result"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int tclRowSub (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   char *regName = NULL;
   char *regRowName = NULL;
   unsigned int dither = FALSE;

   REGION *reg = NULL;
   REGION *regRow = NULL;
   int retStatus;

   g_imgRowSubTbl[1].dst = &regName;
   g_imgRowSubTbl[2].dst = &regRowName;
   g_imgRowSubTbl[3].dst = &dither;

/* Parse line */
   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_imgRowSubTbl,
                      0, g_imgRowSub)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, regName, (void *)&reg, "REGION")
	!= TCL_OK) return (TCL_ERROR);
   if (shTclAddrGetFromName(interp, regRowName, (void *)&regRow, "REGION")
	!= TCL_OK) return (TCL_ERROR);

/* Check for dimension compatibility */
   if (regRow->ncol != 1) {
	Tcl_SetResult (interp, "regRow col dimension must be 1", TCL_STATIC);
	return TCL_ERROR;
	}
   if (reg->nrow != regRow->nrow) {
	Tcl_SetResult (interp, "Row dimensions do not match", TCL_STATIC);
	return TCL_ERROR;
	}
   if (imgRowSub (reg, regRow, dither) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Row subtraction operation unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclColSub	****************************/
/***************************************************************************/
/* Subtract a vector from an image column by column. */

static char *g_imgColSub = "imgColSub";
static ftclArgvInfo g_imgColSubTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL,
	"Subtract a vector col by col from an image"},
    {"<REGION>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to image"},
    {"<REGCOL>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to col vector"},
    {"-dither",   FTCL_ARGV_CONSTANT, (void *)1, NULL,
	"Set if we dither result"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int tclColSub (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   char *regName = NULL;
   char *regColName = NULL;
   int dither = FALSE;

   REGION *reg = NULL;
   REGION *regCol = NULL;
   int retStatus;

   g_imgColSubTbl[1].dst = &regName;
   g_imgColSubTbl[2].dst = &regColName;
   g_imgColSubTbl[3].dst = &dither;

/* Parse line */
   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_imgColSubTbl,
                      0, g_imgColSub)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, regName, (void *)&reg, "REGION")
	!= TCL_OK) return (TCL_ERROR);
   if (shTclAddrGetFromName(interp, regColName, (void *)&regCol, "REGION")
	!= TCL_OK) return (TCL_ERROR);

/* Check for dimension compatibility */
   if (regCol->nrow != 1) {
	Tcl_SetResult (interp, "regCol row dimension must be 1", TCL_STATIC);
	return TCL_ERROR;
	}
   if (reg->ncol != regCol->ncol) {
	Tcl_SetResult (interp, "Col dimensions do not match", TCL_STATIC);
	return TCL_ERROR;
	}
   if (imgColSub (reg, regCol, dither) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Col subtraction operation unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclConAdd	****************************/
/***************************************************************************/
/* Add a constant to an image */

static char *g_imgConAdd = "imgConAdd";
static ftclArgvInfo g_imgConAddTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL,
	"Add a constant to an image"},
    {"<REGION>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to image"},
    {"<CONSTANT>",  FTCL_ARGV_DOUBLE, NULL, NULL, "Constant value"},
    {"-dither",   FTCL_ARGV_CONSTANT, (void *)1, NULL,
	"Set if we dither result"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int tclConAdd (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   char *regName = NULL;
   double value;
   unsigned int dither = FALSE;

   REGION *reg = NULL;
   REGION *regCol = NULL;
   int retStatus;

   g_imgConAddTbl[1].dst = &regName;
   g_imgConAddTbl[2].dst = &value;
   g_imgConAddTbl[3].dst = &dither;

/* Parse line */
   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_imgConAddTbl,
                      0, g_imgConAdd)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, regName, (void *)&reg, "REGION")
	!= TCL_OK) return (TCL_ERROR);

   if (imgConAdd (reg, value, dither) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Constant addition unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/***************************************************************************/
/**************************	tclCsmooth	****************************/
/***************************************************************************/
/* Add a constant to an image */

static char *g_imgCsmooth = "imgCsmooth";
static ftclArgvInfo g_imgCsmoothTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL,
	"Smooth a region in the column direction"},
    {"<REGION>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to image"},
    {"-dither",   FTCL_ARGV_CONSTANT, (void *)1, NULL,
	"Set if we dither result"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int tclCsmooth (ClientData clientData, Tcl_Interp *interp,
   int argc, char **argv)
{

   char *regName = NULL;
   unsigned int dither = FALSE;

   REGION *reg = NULL;
   int retStatus;

   g_imgCsmoothTbl[1].dst = &regName;
   g_imgCsmoothTbl[2].dst = &dither;

/* Parse line */
   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_imgCsmoothTbl,
                      0, g_imgCsmooth)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, regName, (void *)&reg, "REGION")
	!= TCL_OK) return (TCL_ERROR);

   if (imgCsmooth (reg, dither) == IMG_SUCCESS) return (TCL_OK);
   Tcl_SetResult (interp, "Constant addition unsuccessful", TCL_STATIC);
   return TCL_ERROR;
}

/************ declarations of TCL verbs for this module ***********/

void imgTclRowColDeclare (Tcl_Interp *interp) {

   shTclDeclare(interp,"imgRowMed",
                (Tcl_CmdProc *)tclRowMed,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclRowMed_hlp, tclRowMed_use);

   shTclDeclare(interp,"imgColMed",
                (Tcl_CmdProc *)tclColMed,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclColMed_hlp, tclColMed_use);

   shTclDeclare(interp,"imgRowAvg",
                (Tcl_CmdProc *)tclRowAvg,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclRowAvg_hlp, tclRowAvg_use);

   shTclDeclare(interp,"imgColAvg",
                (Tcl_CmdProc *)tclColAvg,
                (ClientData) 0,
                (Tcl_CmdDeleteProc *)NULL,
                module, tclColAvg_hlp, tclColAvg_use);

   shTclDeclare(interp, g_imgRowSub, (Tcl_CmdProc *)tclRowSub,
	(ClientData) 0, (Tcl_CmdDeleteProc *)NULL, module,
	shTclGetArgInfo(interp, g_imgRowSubTbl, 0, g_imgRowSub),
	shTclGetUsage(interp, g_imgRowSubTbl, 0, g_imgRowSub));

   shTclDeclare(interp, g_imgColSub, (Tcl_CmdProc *)tclColSub,
	(ClientData) 0, (Tcl_CmdDeleteProc *)NULL, module,
	shTclGetArgInfo(interp, g_imgColSubTbl, 0, g_imgColSub),
	shTclGetUsage(interp, g_imgColSubTbl, 0, g_imgColSub));

   shTclDeclare(interp, g_imgConAdd, (Tcl_CmdProc *)tclConAdd,
	(ClientData) 0, (Tcl_CmdDeleteProc *)NULL, module,
	shTclGetArgInfo(interp, g_imgConAddTbl, 0, g_imgConAdd),
	shTclGetUsage(interp, g_imgConAddTbl, 0, g_imgConAdd));

   shTclDeclare(interp, g_imgCsmooth, (Tcl_CmdProc *)tclCsmooth,
	(ClientData) 0, (Tcl_CmdDeleteProc *)NULL, module,
	shTclGetArgInfo(interp, g_imgCsmoothTbl, 0, g_imgCsmooth),
	shTclGetUsage(interp, g_imgCsmoothTbl, 0, g_imgCsmooth));

   return;
}
