/*****************************************************************************
******************************************************************************
**
** FILE:
**	imgTclSqrt.c
**
** ABSTRACT:
**	This file contains 
**
** ENTRY POINT		SCOPE	DESCRIPTION
** -------------------------------------------------------------------------
** imgTclSqrtDeclare	public	declare all the verbs defined in this module
** tclSqrt		static	Square root unsigned short int image
** tclSqr		static  Reverse above
** tclSim		static	Simulate a noisy image
**
** ENVIRONMENT:
**	ANSI C.
**
** REQUIRED PRODUCTS:
**	FTCL	 	TCL + Fermi extension
**      DERVISH
**	IMAGE
**
** AUTHORS:
**	Creation date:  Oct. 1993
**	Steve Kent
**
******************************************************************************
******************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termio.h>
#include <sys/types.h>
#include <sys/file.h>
#include "dervish.h"
#include "shTclHandle.h"
#include "img_msg_c.h"
#include "imgProto.h"

static char *module = "img";		/* name of this facility */


/*============================================================================  
**============================================================================
**
** ROUTINE: tclSqrt
**
** DESCRIPTION:
**	Call routine to take sqrt of each pixel in an unsigned short int image.
**	This processing is done to enhance compressibility of an image.
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	imgSqrt
**
**============================================================================
*/
/*
 * Input is handle to image, read noise, gain, quantization value
 * No return
 */

static char *tclSqrt_use =
  "USAGE: imgSqrt <region-handle> rnoise gain quant";
static char *tclSqrt_hlp =
  "Take Square root of pixel values";

static int
tclSqrt(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg;
   UDATA **rows;
   int nr, nc;
   double rnoise, gain, quant;

   shErrStackClear();

   if(argc != 5) {
      Tcl_SetResult(interp,tclSqrt_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);


/* Get addr of unsigned int rows */

   if (reg -> type != TYPE_U16) {
        printf ("Must be unsigned 16 bit integers.\n");
        printf ("Type is %d\n",reg->type);
        return (TCL_ERROR);
        }
   rows = reg ->rows_u16;
   nr = reg -> nrow;
   nc = reg -> ncol;

/* Get the remaining parameters */

   if (Tcl_GetDouble (interp, argv[2], &rnoise)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &gain)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &quant)  != TCL_OK) return(TCL_ERROR);


/* Done with the handles, etc. Now the action begins */

   if (imgSqrt (rows, nr,nc, rnoise, gain, quant) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in imgSqrt", TCL_STATIC);
	return(TCL_ERROR);
	}
   
   return(TCL_OK);
}


/*============================================================================  
**============================================================================
**
** ROUTINE: tclSqr
**
** DESCRIPTION:
**	Call routine to take sqr of each pixel in an unsigned short int image.
**	This is restoration of imgSqrt
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	imgSqr
**
**============================================================================
*/
/*
 * Input is handle to image, read noise, gain, quantization value
 * No return
 */

static char *tclSqr_use =
  "USAGE: imgSqr <region-handle> rnoise gain quant";
static char *tclSqr_hlp =
  "Take Square of pixel values";

static int
tclSqr(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{

   REGION *reg;
   UDATA **rows;
   int nr, nc;
   double rnoise, gain, quant;

   shErrStackClear();

   if(argc != 5) {
      Tcl_SetResult(interp,tclSqr_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);


/* Get addr of unsigned int rows */

   if (reg -> type != TYPE_U16) {
        printf ("Must be unsigned 16 bit integers.\n");
        printf ("Type is %d\n",reg->type);
        return (TCL_ERROR);
        }
   rows = reg ->rows_u16;
   nr = reg -> nrow;
   nc = reg -> ncol;

/* Get the remaining parameters */

   if (Tcl_GetDouble (interp, argv[2], &rnoise)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &gain)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &quant)  != TCL_OK) return(TCL_ERROR);


/* Done with the handles, etc. Now the action begins */

   if (imgSqr (rows, nr,nc, rnoise, gain, quant) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in imgSqr", TCL_STATIC);
	return(TCL_ERROR);
	}
   
   return(TCL_OK);
}


/*============================================================================  
**============================================================================
**
** ROUTINE: tclSim
**
** DESCRIPTION:
**	Call routine to simulate a noisy image.
**
** RETURN VALUES:
**	TCL_OK -   Successful completion.
**	TCL_ERROR -   Failed miserably.
**
** CALLS TO:
**	imgSim
**
**============================================================================
*/
/*
 * Input is handle to image, read noise, gain, quantization value
 * No return
 */

static char *tclSim_use =
  "USAGE: imgSim <region-handle> rnoise gain sky";
static char *tclSim_hlp =
  "Simulate a noisy image";

static int
tclSim(
	  ClientData clientData,
	  Tcl_Interp *interp,
	  int argc,
	  char **argv
	  )
{
   REGION *reg;
   UDATA **rows;
   int nr, nc;
   double rnoise, gain, sky;

   shErrStackClear();

   if(argc != 5) {
      Tcl_SetResult(interp,tclSim_use,TCL_STATIC);
      return(TCL_ERROR);
   }

/* Get the region pointer */

   if (shTclAddrGetFromName(interp, argv[1], (void **)&reg, "REGION") != TCL_OK)
	return (TCL_ERROR);


/* Get addr of unsigned int rows */

   if (reg -> type != TYPE_U16) {
        printf ("Must be unsigned 16 bit integers.\n");
        printf ("Type is %d\n",reg->type);
        return (TCL_ERROR);
        }
   rows = reg ->rows_u16;
   nr = reg -> nrow;
   nc = reg -> ncol;

/* Get the remaining parameters */

   if (Tcl_GetDouble (interp, argv[2], &rnoise)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[3], &gain)  != TCL_OK) return(TCL_ERROR);
   if (Tcl_GetDouble (interp, argv[4], &sky)  != TCL_OK) return(TCL_ERROR);


/* Done with the handles, etc. Now the action begins */

   if (imgSim (rows, nr,nc, rnoise, gain, sky) != IMG_SUCCESS) {
	Tcl_SetResult(interp,"Error in imgSim", TCL_STATIC);
	return(TCL_ERROR);
	}
   
   return(TCL_OK);
}


/*****************************************************************************/
/*
 * Declare my new tcl verbs to tcl
 */
void
imgTclSqrtDeclare(Tcl_Interp *interp) 
{
   shTclDeclare(interp,"imgSqrt", 
		(Tcl_CmdProc *)tclSqrt,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclSqrt_hlp, tclSqrt_use);

   shTclDeclare(interp,"imgSqr", 
		(Tcl_CmdProc *)tclSqr,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclSqr_hlp, tclSqr_use);

   shTclDeclare(interp,"imgSim", 
		(Tcl_CmdProc *)tclSim,
		(ClientData) 0,
		(Tcl_CmdDeleteProc *)NULL, 
		module, tclSim_hlp, tclSim_use);

   return;
}

