/*****************************************************************************
******************************************************************************
** <AUTO>
   FILE:
  	shTclTransContrib.c
   <HTML> 
   ABSTRACT:
	Support routines for new chain <==> tbl routines.  Routines are
	supplied to:
	   a. Create and add entries to translation tables.
	   b. Convert from TBLCOLs to CHAINs and back
	The methodology of the old stuff is followed as much as possible,
	but several changes are made:
	   a. A translation table is compiled for a specific TYPE.
	   b. One does not create the destination chain in tblToSchema
		in advance.  This eliminates the -reuse option for now.
	   c. An attempt was made to eliminate unneeded redundant information
		from the translation table specifiers.
	   d. A translation table is intended to be bi-direction capable.
		No need to diddle for reading vs. writing.
	   e. Attributes are specified as per normal expression syntax: e.g.,
		mask->nrow - there is no need to break this into two
		entries in the table.
</HTML>
</AUTO> */

/************************************************************************
**
**	This file contains many routines to convert data from TBLCOL's
**	to CHAINs and back.  It is not recommended that they be used
**	publicly - they work as a set and the job of initiallizing data
**	various data structures is split among the routines in ways that
**	may not be apparent.
**
** ENTRY POINT		SCOPE	DESCRIPTION
** -------------------------------------------------------------------------
** strtolower		Local	Convert a string to lower case
** fieldNameGet		Local	Strip off dimension info from an attribute
** tclTransTableNew	Local	Create a new translation table
** transEntryDel	Local	Delete a single translation entry
** tclTransEntryAdd	Local	Add a new entry to a translation table
** tclTransEntryDel	Local	Delete an entry from a translation table
** tclTransTableDel	Local	Delete a translation table
** shTblColNewFromTrans	Local
** chainNewFromTrans	Local
** shTclTransTableCompile Local
** tableNewFromTable 	Local  Create a new TBLCOL given a compiles trans.tbl.
** chainCopyToTbl	Local
** tblCopyToChain	Local
** tclChainToTbl	Local	TCL interface to chainToTbl
** tclTblToChain	Local	TCL interface to tblToChain
** -------------------------------------------------------------------------
**
**  
** ENVIRONMENT:
**	ANSI C.
**
** REQUIRED PRODUCTS:
**	Shiva
**
** AUTHORS:
**	Creation date:  Mar 1996
**	Steve Kent
**
******************************************************************************
******************************************************************************
*/

#include <math.h>
#include <ctype.h>
#include <string.h>

#include "dervish.h"
#include "shGeneric.h"
#include "shTransContrib.h"

/*============================================================================  
**============================================================================
**
** LOCAL MACROS, DEFINITIONS, ETC.
**
**============================================================================
*/

static char *module="tclTransContrib";

/*****************************************************************************/
/***			strtolower					   ***/
/***			Convert string to lower case			   ***/
/*****************************************************************************/
static int cindex (char *string, char ch) {
   int i;
   for (i=0; i<strlen(string); i++) {
	if (string[i] == ch) return i;
	}
   return -1;
   }

/*****************************************************************************/

static void strtolower(char *str) {
   char upper[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   char lower[] = "abcdefghijklmnopqrstuvwxyz";
   int i;
   int index;

   for (i=0; i<strlen(str); i++) {
	if ((index=cindex(upper, str[i])) >= 0) {
	   str[i]=lower[index];		/* Difference of pointers is
						   the index in the "upper"
						   array */
	   }
	}
   return;
   }

/*****************************************************************************/
/***			fieldNameGet					   ***/
/***	Parse an attribute name and get the leading name before any "<"	   ***/
/*****************************************************************************/

static void fieldNameGet (char *out, char *in) {
   char *ptr;
   ptr = strchr(in, '<');
   if (ptr == NULL) {
	strcpy (out, in);
   } else {
	strncpy (out, in, ptr-in);   out[ptr-in] = '\0';
	strcat (out, "<0>");
   }
   return;
}


/*****************************************************************************/
/***			tclTransTableNew				   ***/
/*****************************************************************************/

static char *g_transTableNew = "transTableNew";
static ftclArgvInfo g_transTableNewTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL, "Create a new TRANSTABLE"},
    {"<TYPE>",  FTCL_ARGV_STRING, NULL, NULL, "Schema TYPE of table"},
    {"-constructor",  FTCL_ARGV_STRING, NULL, NULL, "Constructor proc"},
    {"-destructor",  FTCL_ARGV_STRING, NULL, NULL, "Destructor proc"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int
tclTransTableNew (
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   TRANSTABLE *transTable = NULL;
   char *typeName = NULL;
   char *constrName = NULL;
   char *destrName = NULL;
   char transName[HANDLE_NAMELEN];
   int retStatus;

   g_transTableNewTbl[1].dst = &typeName;
   g_transTableNewTbl[2].dst = &constrName;
   g_transTableNewTbl[3].dst = &destrName;

/* Parse line */
   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_transTableNewTbl,
                      0, g_transTableNew)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   transTable = (TRANSTABLE *)shGenericNew("TRANSTABLE");
   transTable->table = shChainNew("TRANSENTRY");

/* Should be pointer to static stor.*/
   transTable->schemaTypeName = shNameGetFromType(shTypeGetFromName(typeName));

/* Store constructor procedure name if provided */
   if (constrName != NULL) {
	transTable->constructor = (char *)shMalloc(strlen(constrName)+1);
	strcpy (transTable->constructor, constrName);
	}

/* Store destructor name if provided */
   if (destrName != NULL) {
	transTable->destructor = (char *)shMalloc(strlen(destrName)+1);
	strcpy (transTable->destructor, destrName);
	}

/* Make a new handle */
   if (shTclHandleNew (interp, transName, "TRANSTABLE", transTable) != TCL_OK)
	return(TCL_ERROR);

   Tcl_SetResult (interp, transName, TCL_VOLATILE);
   return TCL_OK;
}

/*****************************************************************************/
/***			transEntryDel					   ***/
/*****************************************************************************/
/* This function is used in more than one place */

static void transEntryDel (TRANSENTRY *entry) {

   int i;

   if (entry == NULL) return;
   if (entry->fitsEntry != NULL) shFree(entry->fitsEntry);
   if (entry->schemaEntry != NULL) shFree(entry->schemaEntry);
   if (entry->fitsProgram!= NULL) shProgramDel(entry->fitsProgram);
   if (entry->schemaProgram != NULL) shProgramDel(entry->schemaProgram);
   if (entry->heaplen != NULL) shFree(entry->heaplen);
   if (entry->heaplenProgram != NULL) shProgramDel(entry->heaplenProgram);
   if (entry->dimen != NULL) shFree(entry->dimen);
   if (entry->dimenProgram != NULL) {
	for (i=0; i<entry->schemaProgram->nstar; i++) {
	   shProgramDel(entry->dimenProgram[i]);
	   }
	}
   if (entry->dimenCopy != NULL) shFree(entry->dimenCopy);
   shFree(entry);
   return;
}

/*****************************************************************************/
/***			tclTransEntryAdd				   ***/
/*****************************************************************************/

static char *g_transEntryAdd = "transEntryAdd";
static ftclArgvInfo g_transEntryAddTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL, "Add an entry to a TRANSTABLE"},
    {"<TABLE>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a TRANSTABLE"},
    {"<FITSENTRY>",  FTCL_ARGV_STRING, NULL, NULL, "FITS side field name"},
    {"<SCHEMAENTRY>",  FTCL_ARGV_STRING, NULL, NULL, \
	"Schema side element name"},
    {"-type",     FTCL_ARGV_STRING, NULL, NULL, "FITS side type" },
    {"-heaplen",     FTCL_ARGV_STRING, NULL, NULL, \
	"Schema side heap length info" },
    {"-dimen",     FTCL_ARGV_STRING, NULL, NULL, \
	"Schema side heap dimension info" },
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int
tclTransEntryAdd (
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   char *tableName = NULL;
   char *fitsEntry = NULL;
   char *schemaEntry = NULL;
   char *typeName = NULL;
   char *heaplen = NULL;
   char *dimen = NULL;
   TRANSTABLE *transTable = NULL;
   TRANSENTRY *entry;
   int retStatus;
   char buf[300];
   int i,j;
   char *dptr;
   char *dlast;
   char *test;
   int index;
   int nstar;

   g_transEntryAddTbl[1].dst = &tableName;
   g_transEntryAddTbl[2].dst = &fitsEntry;
   g_transEntryAddTbl[3].dst = &schemaEntry;
   g_transEntryAddTbl[4].dst = &typeName;
   g_transEntryAddTbl[5].dst = &heaplen;
   g_transEntryAddTbl[6].dst = &dimen;

/* Reset the optional parameter */
   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_transEntryAddTbl,
                      0, g_transEntryAdd)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, tableName, (void **)&transTable,
	"TRANSTABLE") != TCL_OK) return TCL_ERROR;

   entry = (TRANSENTRY *)shMalloc(sizeof(TRANSENTRY));

/* Zero out for safety */
   memset(entry, '\0', sizeof(TRANSENTRY));
   entry->fitsEntry = (char *)shMalloc(strlen(fitsEntry)+1);   
   strcpy(entry->fitsEntry, fitsEntry);
   entry->schemaEntry = (char *)shMalloc(strlen(schemaEntry)+1);
   strcpy(entry->schemaEntry, schemaEntry);

/* Compile into a program */
   if ((entry->schemaProgram = shProgramCreate (transTable->schemaTypeName,
	entry->schemaEntry)) == NULL) goto error;

/* Allocate space for the dimension info programs and copy functions */
   if (entry->schemaProgram->nstar > 0) {
	nstar = entry->schemaProgram->nstar;
	entry->dimenProgram = (PROGRAM **)shMalloc(nstar*sizeof(PROGRAM *));
	memset (entry->dimenProgram, '\0', nstar*sizeof(PROGRAM *));
	entry->dimenCopy =
	   (COPYPROGRAM **)shMalloc(nstar*sizeof(COPYPROGRAM *));
	memset (entry->dimenCopy, '\0', nstar*sizeof(COPYPROGRAM *));
	}

/* FITS side type is taken from schema side unless overridden */
   if (typeName != NULL) entry->fitsType = shTypeGetFromName(typeName);
      else entry->fitsType = entry->schemaProgram->type;

/* If heaplen argument is provided, decode into a program and a copy
 * function.  heaplen must be a schema side variable.
*/

   if (heaplen != NULL) {
	if (entry->schemaProgram->nstar == 0) {
	   Tcl_SetResult (interp, "heaplen is inappropriate for this arg.",
		TCL_STATIC);
	   goto error;
	   }
	entry->heaplen = (char *)shMalloc(strlen(heaplen)+1);
	strcpy (entry->heaplen, heaplen);
	if ((entry->heaplenProgram = shProgramCreate
		(transTable->schemaTypeName, heaplen)) == NULL) {
	   Tcl_AppendResult (interp, "Unable to compile program for ",
		   heaplen, NULL);
	   goto error;
	   }
	if (entry->heaplenProgram->nstar != 0 || entry->heaplenProgram->ndim
		!= 0) {
	   Tcl_SetResult(interp, "heaplen must point to a single item",
			TCL_STATIC);
	   goto error;
	   }
	if ((entry->heaplenCopy = p_shCopyFuncGet (shTypeGetFromName("INT"),
		entry->heaplenProgram->type)) == NULL) {
	   Tcl_SetResult(interp, "heaplen must point to a numerical type",
		   TCL_STATIC);
	   goto error;
	   }
	}

/* If dimen argument is provided, decode into a program and a copy
 * function.  dimen comes in as a list of either schema side variable
 * names or integers.
*/
   if (dimen != NULL) {
	if (entry->schemaProgram->nstar == 0) {
	   Tcl_SetResult (interp, "dimen is inappropriate for this arg.",
		TCL_STATIC);
	   goto error;
	   }

/* There must be no more than nstar dimensions.  If the schema datatype is
 * a string, then there can be one less, in which case the string is assumed
 * to be null terminated.  For now, only nstar=1 is allowed for strings.
*/
	nstar = entry->schemaProgram->nstar;
	entry->dimen = (char *)shMalloc(strlen(dimen)+1);
	strcpy (entry->dimen, dimen);

/* dimen comes in the form of a TCL list.  Split into individual arguments */
        dptr = entry->dimen - 1;
        dlast = entry->dimen + strlen(entry->dimen) - 1;
	for (j=0; j<nstar; j++) {
	   dptr++;
	   if (dptr > dlast) break;
	   if (*dptr == ' ') continue;
	   i=0; buf[0]='\0';
	   while (*dptr != ' ') {
		if (dptr > dlast) break;
		buf[i++] = *dptr;
		buf[i] = '\0';
		dptr ++;
		}
/* Convert to integer; check for error */
	   if (i == 0) goto error;

/* We can input either a number or a schema side variable.  Test the first
 * character and see if it is a digit.
*/
	   if (isdigit(buf[0])) {
              index = strtol(buf, &test, 10);
              if (buf+strlen(buf) - test != 0) {
		Tcl_AppendResult (interp, "Unable to convert to int: ",
		   buf, NULL);
		goto error;
		}
	      entry->schemaProgram->sdim[j] = index;
	   } else {
	      if ((entry->dimenProgram[j] = shProgramCreate
		(transTable->schemaTypeName, buf)) == NULL) {
		Tcl_AppendResult (interp, "Unable to compile program for ",
		   buf, NULL);
		goto error;
		}
	      if (entry->dimenProgram[j]->nstar != 0 ||
	        entry->dimenProgram[j]->ndim != 0) {
		Tcl_SetResult(interp, "dimen must point to a single item",
			TCL_STATIC);
		goto error;
		}
	      if ((entry->dimenCopy[j] = p_shCopyFuncGet (
	         entry->dimenProgram[j]->type, shTypeGetFromName("INT")))
	         == NULL) {
		Tcl_SetResult(interp, "dimen must point to a numerical type",
		   TCL_STATIC);
		goto error;
		}
	      }
	   }
	}

/* We will defer creating FITS side info.  If we are creating a new table,
 * we could create it now, but if we are reading from a table into a schema,
 * we can permit more flexible functionality if we have table side info first.  
*/
   shChainElementAddByPos (transTable->table, entry, "TRANSENTRY", TAIL, AFTER);
   return TCL_OK;

error:
   transEntryDel(entry);
   return TCL_ERROR;
}


/*****************************************************************************/
/***			tclTransEntryDel				   ***/
/*****************************************************************************/

static char *g_transEntryDel = "transEntryDel";
static ftclArgvInfo g_transEntryDelTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL, "Delete an entry from a TRANSTABLE"},
    {"<TABLE>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a TRANSTABLE"},
    {"<INDEX>",  FTCL_ARGV_INT,    NULL, NULL, "Index of entry to delete"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int
tclTransEntryDel (
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   char *tableName = NULL;
   int index;
   TRANSTABLE *transTable = NULL;
   TRANSENTRY *entry = NULL;
   int retStatus;

   g_transEntryDelTbl[1].dst = &tableName;
   g_transEntryDelTbl[2].dst = &index;

   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_transEntryDelTbl,
                      0, g_transEntryDel)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, tableName, (void **)&transTable,
	"TRANSTABLE") != TCL_OK) return TCL_ERROR;

   if (index >= shChainSize(transTable->table)) {
	Tcl_SetResult (interp, "Index out of bounds", TCL_STATIC);
	return TCL_ERROR;
	}

   entry = shChainElementRemByPos (transTable->table, index);
   if (entry == NULL) return TCL_OK;
   transEntryDel(entry);
   return TCL_OK;
}


/*****************************************************************************/
/***			tclTransTableDel				   ***/
/*****************************************************************************/

static char *g_transTableDel = "transTableDel";
static ftclArgvInfo g_transTableDelTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL, "Delete a TRANSTABLE"},
    {"<TABLE>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a TRANSTABLE"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int
tclTransTableDel (
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   char *tableName = NULL;
   TRANSTABLE *transTable = NULL;
   TRANSENTRY *entry;
   int retStatus;

   g_transTableDelTbl[1].dst = &tableName;

   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_transTableDelTbl,
                      0, g_transTableDel)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, tableName, (void **)&transTable,
	"TRANSTABLE") != TCL_OK) return TCL_ERROR;

   if (transTable->table == NULL) return TCL_OK;
   while (shChainSize(transTable->table) > 0) {
	entry = shChainElementRemByPos (transTable->table, 0);
	transEntryDel(entry);
	}
   shChainDel(transTable->table);
   if (transTable->constructor != NULL) shFree(transTable->constructor);
   if (transTable->destructor != NULL) shFree(transTable->destructor);
   if (transTable->entry != NULL) shFree(transTable->entry);
   shGenericDel (transTable);
   p_shTclHandleDel (interp, tableName);
   return TCL_OK;
}

/*****************************************************************************/
/***			shTblColNewFromTrans				   ***/
/*****************************************************************************/

/* Create and allocate space in a TBLCOL based on a compiled translation table*/
/* We must have number of rows known in advance */
/* We assume that the FITS programs do not yet exist.  All information is
 * taken from the schema side except the field name and (if overridden) the
 * fits type.  This might be a little inflexible since it precludes some
 * more esoteric translation such as heap on one side to inline on the other.
*/

TBLCOL *tblColNewFromTrans (TRANSTABLE *table, int rowCnt) {

   TBLCOL *tblCol;
   TRANSENTRY *entry;
   ARRAY *array;
   TBLFLD *tblfld;
   TYPE type;
   int index;
   int test;
   int i, j;

/* Begin code */
/* Another Nicinski obstinacy: We must specify at least one field when
 * creating the TBLCOL.  We shall do so and then delete it immediately
*/

   if (shTblColCreate(rowCnt, 1, 1, &tblCol) != SH_SUCCESS) return NULL;
   array = shChainElementRemByPos (&tblCol->fld, 0);
   shArrayDel(array);

   for (i=0; i<shChainSize(table->table); i++) {   
	entry = shChainElementGetByPos(table->table, i); /* For convenience */

/* Error checking */
	if (entry == NULL) continue;
/* Check to see if there is a copy function.  If not, skip this entry */
	if (p_shCopyFuncGet(entry->fitsType, entry->schemaProgram->type) == NULL)
	   continue;
/* Passed all tests. Create the field */
	if (shTblFldAdd (tblCol, 1, &array, &tblfld, &index) != SH_SUCCESS)
	   goto error;
	shTblTBLFLDsetWithAscii (array, SH_TBL_TTYPE, entry->fitsEntry, NULL);

/* If fitsType is CHAR or char, convert to type STR.  Life is so much
 * nicer if we do.  TBLCOLs use STR to mean something different from the
 * rest of shiva.  grrrrr.....
*/
	type = entry->fitsType;
	if (type == shTypeGetFromName("CHAR") || type ==
	   shTypeGetFromName("char")) type = shTypeGetFromName("STR");

/* Deal with heap.  We will use heap if the schema side has at least one
 * nstar and if not all schema side heap dimensions are defined in
 * advance.  First, test to see if all dims are supplied.  We check the
 * dimenProgram pointers and the sdim dimensions (this catches the case that
 * we will use strlen later for char * string variables)
*/
	test=0;
	for (j=0; j<entry->schemaProgram->nstar; j++) {
	   if (entry->dimenProgram[j] != NULL ||
		entry->schemaProgram->sdim[j] == 0) test=1;
	   }
	if (test == 0) {

/* Create FITS side inline array.  Add dimension info to array.
 * Remember that 1st dimension in an ARRAY is the table length, which is
 * already filled in.
*/
	   for (j=0; j<entry->schemaProgram->ndim; j++) {
		array->dim[array->dimCnt] = entry->schemaProgram->dim[j];
		array->dimCnt++;
		}
	   for (j=0; j<entry->schemaProgram->nstar; j++) {
		array->dim[array->dimCnt] = entry->schemaProgram->sdim[j];
		array->dimCnt++;
		}

	   if (shArrayDataAlloc (array, shNameGetFromType(type)) != SH_SUCCESS)
		goto error;
	} else {

/* Heap is treated quite differently.  Although FITS allows it, Nicinski
 * does not support arrays of pointers in heap; thus, we flatten all
 * schema side arrays.  The tblcol ARRAY is initialized with space for
 * heap descriptors, but all other initializations (like allocating space
 * for heap) is deferred to run-time.
*/
	   if (shArrayDataAlloc (array, "TBLHEAPDSC") != SH_SUCCESS)
		goto error;

/* The heap info is allocated at run time.  I will need to know the
 * destination data type before then (when compiling the translation table)
 * so insert here.  Nicinski will overwrite later when allocating heap.
 * Should be the same data type then.
*/
	   ((TBLFLD *)(array->info))->heap.schemaType = type;
	   }
	}
   return tblCol;

error:
   shTblcolDel (tblCol);
   return NULL;
   }
   
/*****************************************************************************/
/***			chainNewFromTrans				   ***/
/*****************************************************************************/

/*** Allocate space and all elements for an output chain ***/

static CHAIN *chainNewFromTrans (
	Tcl_Interp *interp,		/* Used in case constructor provided */
	TRANSTABLE *table,		/* IN: Translation table */
	int rowCnt			/* IN: Number of objects to create */
	)
{

   CHAIN *chain = NULL;
   void *entry;
   int i;
   int size;
   const SCHEMA *schema;
   char entryName[HANDLE_NAMELEN];
   char command[300];

   if ((chain = shChainNew(table->schemaTypeName)) == NULL) goto error;

   if ((schema = shSchemaGet(table->schemaTypeName)) == NULL) goto error;
   size = schema->size;

   for (i=0; i<rowCnt; i++) {

/* Constructor, if provided, is a TCL proc. */
	if (table->constructor != NULL) {
	   if (Tcl_Eval(interp, table->constructor) != TCL_OK) goto error;
	   strncpy (entryName, interp->result, HANDLE_NAMELEN-1);
	   entryName[HANDLE_NAMELEN-1] = '\0';
	   if (shTclAddrGetFromName (interp, entryName, &entry,
		table->schemaTypeName) != TCL_OK) goto error;
	   if (p_shTclHandleDel (interp, entryName) != TCL_OK) goto error;
	} else {
/*	   entry = shGenericNew(table->schemaTypeName); */
	   entry = shMalloc(size);
	   if (entry == NULL) goto error;
	   memset (entry, '\0', size);
	   }
	shChainElementAddByPos (chain, entry, table->schemaTypeName,
	   TAIL, AFTER);
	}
   return chain;

/* Delete any partially constructed chain.  Use destructor function if
 * provided; else generic deletes
*/

error:
   if (chain != NULL) {
	for (i=0; i<shChainSize(chain); i++) {
	   entry = shChainElementGetByPos (chain, i);
	   if (table->destructor != NULL) {
	      if (shTclHandleNew (interp, entryName, table->schemaTypeName,
		entry) != TCL_OK) return NULL;
	      strcpy (command, table->destructor);
	      strcat (command, " ");
	      strcat (command, entryName);
	      if (Tcl_Eval (interp, command) != TCL_OK) return NULL;
	      if (p_shTclHandleDel (interp, entryName) != TCL_OK)
		return NULL;
	   } else {
	      shFree(entry);
	      }
	   }
	shChainDel(chain);
	}
   shErrStackPush ("Error while trying to allocate chain in chainNewFromTbl\n");
   return NULL;
   }

/*****************************************************************************/
/***			transTableCompile				   ***/
/*****************************************************************************/
/* This is a C callable routine to compile a translation table.
 * The compiled table is used internally to support conversion from
 * TBLCOLs to Chains and back.
 * Compilations consists of:
 *	a: Converting the translation table pointers from a chain to an array
 *	b: Creating programs for the FITS side.
*/

static int transTableCompile (
	TRANSTABLE *table,	/* IN: Translation table */
	TBLCOL *tblCol)		/* IN: A TBLCOL */
{

   ARRAY *array = NULL;
   TBLFLD *tblfld;

#define MAXBUFLEN 300

   TRANSENTRY *entry;
   SCHEMA *schema;
   SCHEMA_ELEM *schemaElem;	/* Pointer to char string array for compiling
				   schemaElems */
   int i, j;
   char buf[MAXBUFLEN];		/* Print to this, concatenate to schemaBuf */
   char dimbuf[MAXBUFLEN];	/* Really temporary storage */
   TYPE type;			/* Temporary storage */
   int fldindex;
   int nfield;
   int offset;
   int nitem;
   int retstatus;

   if (table == NULL) return TCL_OK;
   table->nEntry = shChainSize(table->table);
   if (table->nEntry == 0) return TCL_OK;
   table->entry = (TRANSENTRY **)shMalloc(table->nEntry * sizeof(TRANSENTRY *));

/* Create array of pointers to elements on chain. */
   for (i=0; i<table->nEntry; i++) {
	table->entry[i] = shChainElementGetByPos (table->table, i);
	}

/* Create a temporary schema analogous to a Shiva TYPE for the FITS side.
 * This schema is a bit of an invention.  Each data element is one field of
 * the FITS table.  All start at offset 0.  Otherwise, it describes one
 * row of the FITS table.
*/

/* Allocate buffers */
   nfield=shChainSize(&tblCol->fld);
   if (nfield == 0) return TCL_OK;
   schema = (SCHEMA *)shMalloc(sizeof(SCHEMA));
   memset (schema, '\0', sizeof(SCHEMA));
   strcpy (schema->name, "FITS");
   schema->type = STRUCT;
   schema->nelem = nfield;
   schema->size = 0;		/* Don't know if size is ever used */
   schema->elems = (SCHEMA_ELEM *)shMalloc(nfield*sizeof (SCHEMA_ELEM));
   memset (schema->elems, '\0', nfield*sizeof(SCHEMA_ELEM));
   for (i=0; i<nfield; i++) {
	schemaElem = &schema->elems[i];
/* Look for the TBLCOL field
 * by index and extract the schema definition from there.
*/
	if (shTblFldLoc (tblCol, i, NULL, 0, SH_CASE_INSENSITIVE,
		&array, &tblfld, &fldindex) != SH_SUCCESS) continue;
	strcpy (schemaElem->name, tblfld->TTYPE);

/* Use lower case */
	strtolower(schemaElem->name);

/* If heap, nstar = 1.  This means that top level pointer for the FITS side
 * will point to the TBLHEAPDSC data structure, not the top of heap itself.
 * I can code this by making the schema element have nstar=1 and an offset
 * corresponding to the data pointer in a TBLHEAPDSC.
*/
	if (array->data.schemaType == shTypeGetFromName("TBLHEAPDSC")) {
	   schemaElem->nstar = 1;

/* If we are going tbl->chain, heap data type is defined.  If we are going
 * chain->tbl, heap data type was filled in by me manually in tblColNewFromTbl.
 * This is one twist that makes these routines not easy to use publically.
*/
	   type = ((TBLFLD *)(array->info))->heap.schemaType;
	   offset = offsetof(TBLHEAPDSC, ptr);
	} else {
	   schemaElem->nstar = 0;
	   type = array->data.schemaType;
	   offset = 0;
	   }

/* If FITS table type is STR, convert to char; alas, tblCols use type
 * STR differently from the compiled in SCHEMA - they use it as a surrogate
 * for char, not char *.
*/
	if (type == shTypeGetFromName("STR")) type = shTypeGetFromName("CHAR");
	strcpy (schemaElem->type, shNameGetFromType (type));

/** Note that left dimension in ARRAY is the table length of a FITS table,
 ** which we shall bypass.  All subsequent dimensions are in normal
 ** C order (and thus last dimension increments first).
*/
	if (array->dimCnt > 1) {
	   buf[0] = '\0';
	   nitem = 0;
	   for (j=1; j<array->dimCnt; j++) {
		if (j>1) strcat (buf, " ");
		sprintf (dimbuf, "%d", array->dim[j]);
		strcat (buf, dimbuf);
		nitem *= array->dim[j];
		}
	   schemaElem->nelem = (char *)shMalloc(strlen(buf)+1);
	   strcpy (schemaElem->nelem, buf);
	   }
	schemaElem->offset = offset;		/* Encode offset = 0 */
	}

/* Now create FITS side programs and identify the proper copy functions */
   for (i=0; i<table->nEntry; i++) {
	entry = table->entry[i];

	strcpy (buf, entry->fitsEntry);

/* Use lower case to achieve case insensitivity */
	strtolower(buf);
	if ((entry->fitsProgram = shProgramCreateBySchema (schema,
	   buf)) == NULL) {
	   goto error;
	   }

/* Need to get an index to which field in the TBLCOL matches the
 * attribute name with dimension info stripped off
*/
	fieldNameGet (buf, entry->fitsEntry);
	if (shTblFldLoc (tblCol, -1, buf, 0, SH_CASE_INSENSITIVE,
		&array, &tblfld, &fldindex) != SH_SUCCESS) continue;
	entry->ifld = fldindex;


/* Set up the copy function, one from FITS to Schema, the other in the
 * other direction.  Do this in separate procedures, because it can get
 * kind of ugly.  If FITS side is STR, the fitsProgram->type should be
 * of type CHAR.
*/
	if ((entry->fits2Schema = p_shCopyFuncGet (entry->fitsProgram->type,
		entry->schemaProgram->type)) == NULL) {
	   shErrStackPush ("Type mismatch for entry %d\n",i);
	   goto error;
	   }
	if ((entry->schema2Fits = p_shCopyFuncGet (entry->schemaProgram->type,
		entry->fitsProgram->type)) == NULL) {
	   shErrStackPush ("Type mismatch for entry %d\n",i);
	   goto error;
	   }
	}
   retstatus = TCL_OK;
   goto done;

error:
   retstatus = TCL_ERROR;

done:
/* Now undo the schema */
   for (i=0; i<nfield; i++) {
	if (schema->elems[i].nelem != NULL) shFree(schema->elems[i].nelem);
	}
   shFree(schema->elems);
   shFree(schema);
   return retstatus;

}

/*****************************************************************************/
/***			transTableUncompile				   ***/
/*****************************************************************************/
/* Uncompile a translation table.
 * Uncompilations consists of:
 *	a: Deleting programs for the FITS side.
 *	b: Deleting the array of translation table pointers
*/

static int transTableUncompile (
	TRANSTABLE *table)	/* IN: Translation table */
{

   TRANSENTRY *entry;
   int i;

   if (table == NULL) return TCL_OK;
   if (table->nEntry == 0) return TCL_OK;

/* Now delete FITS side programs */
   for (i=0; i<table->nEntry; i++) {
	entry = table->entry[i];
	shProgramDel (entry->fitsProgram);
	entry->fitsProgram = NULL;
	entry->ifld = 0;
	}

   if (table->entry != NULL) shFree(table->entry);
   table->entry = NULL;
   return TCL_OK;
}

/*****************************************************************************/
/***			arrayMalloc					   ***/
/*****************************************************************************/
/* Recursive procedure to malloc array of pointers */

static void *arrayMalloc (int *sdim, int nstar, int level, int elemSize)
{
   void **ptr;
   int j;

   if (level < nstar-1) {
	ptr = shMalloc(sdim[level] * sizeof (void *));
	for (j=0; j<sdim[level]; j++) {
	   ptr[j] = arrayMalloc(sdim, nstar, level+1, elemSize);
	   }
	return ptr;
	}
   ptr = (void *)shMalloc(sdim[level] * elemSize);
   memset ((char *)ptr, '\0', sdim[level]*elemSize);
   return ptr;
}

/*****************************************************************************/
/***			arrayLocate					   ***/
/*****************************************************************************/
/* Recursively dereference an array of pointers */

static void *arrayLocate (void *schemaPtr, int *sdim, int *subdim, int nstar,
   int level, int index, int elemSize)
{
   void *ptr;
   int j;

   if (level < nstar-1) {
        j = index/subdim[level];

/* This test is probably not necessary, since we ensure it before calling */
/*	if (j<0 || j>= sdim[level]) return NULL; */
        index = index-j*subdim[level];
        schemaPtr = *(void ***)schemaPtr + j;
        ptr = arrayLocate (schemaPtr, sdim, subdim, nstar, level+1, index,
           elemSize);
        return ptr;
        }
   ptr = *((char **)(schemaPtr)) + index*elemSize;
   return ptr;
}

/*****************************************************************************/
/***			chainCopyToTbl					   ***/
/*****************************************************************************/

static int chainCopyToTbl (
	TBLCOL *tblCol,		/* MODIFIED: Table col */
	CHAIN *chain,		/* IN: Source chain */
	TRANSTABLE *table)	/* IN: Translation table */

{
   ARRAY **array;
   TRANSENTRY *entry = NULL;
   char **fitsData = NULL;
   int *fitsItemSize;
   void *schemaData = NULL;
   void *dataPtr;
   int nfield, nentry;
   int i,j,k, l;
   char *schemaPtr;
   char *fitsPtr;
   int nsch;
   int nfits;
   CURSOR_T cursor;
   TYPE type;
   int size;
   int subdim[30];	/* Will we ever exceed 30 dimensions?! */

/* Quick error checks */
   if (shChainSize(chain) != tblCol->rowCnt) {
	shErrStackPush ("Chain size and tbl size don't match\n");
	return TCL_ERROR;
	}
   nfield = shChainSize(&tblCol->fld);
   nentry = table->nEntry;
   if (nfield == 0) return TCL_OK;
   if (nentry == 0) return TCL_OK;

/* Convert the chain in the tblCol to an array for faster access */
/* Allocate an array of pointers that will point to a row of elements in
 * the data areas of the fits fields
*/
   array = (ARRAY **)shMalloc (nentry * sizeof(ARRAY *));
   fitsData = (char **)shMalloc (nentry * sizeof(char *));
   fitsItemSize = (int *)shMalloc (nentry * sizeof(int));
   for (j=0; j<nentry; j++) {
	entry = table->entry[j];
	if (entry->fitsProgram == NULL) continue;
	array[j] = shChainElementGetByPos (&tblCol->fld, entry->ifld);
	fitsData[j] = (char *)(array[j]->data.dataPtr);
	}

/* Step through the chain.  At each step, call programs to compute pointers
 * to the starting bucket locations on each side.  Then enter a loop
 * to copy the data.
*/

   cursor = shChainCursorNew(chain);
   size = shChainSize(chain);

   for (i=0; i<size; i++) {
	schemaData = shChainWalk (chain, cursor, NEXT);

/* Loop through each attribute */   
/* We process non-heap stuff first */
	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fits2Schema == NULL || entry->schemaProgram == NULL
		|| entry->fitsProgram == NULL) continue;
	   if (entry->schemaProgram->nstar > 0 || entry->fitsProgram->nstar > 0)
		continue;

/* Get pointers to starting values */
	   if ((schemaPtr = shProgramEval (schemaData, entry->schemaProgram))
		== NULL) continue;
	   if ((fitsPtr  = shProgramEval (fitsData[j], entry->fitsProgram))
		== NULL) continue;

	   for (k=0; k<entry->schemaProgram->nitem; k++) {
		entry->schema2Fits (schemaPtr, fitsPtr);
		fitsPtr += array[j]->data.incr;
		schemaPtr += entry->schemaProgram->isize;
		}
	   }

/* Now deal with indirections on the schema side.  We permit the FITS
 * side data to be either inline or in heap storage.
 * First, loop through entries and find those with heap.
*/
	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fits2Schema == NULL || entry->schemaProgram == NULL
		|| entry->fitsProgram == NULL) continue;
	   if (entry->schemaProgram->nstar == 0 && entry->schemaProgram->nstar
		== 0) continue;

/* Fill in dimension info on schema side.  Info is stored in
 * schemaProgram->sdim.  Some entries may have been initialized already.
 * Use programs to fill in the rest.
*/

/* Special case: if FITS side is heap, schema side has nstar = 1, type = char,
 * take schema dimension to be length of FITS side.  This is particularly
 * useful for character strings: the heap side includes the trailing null.
*/
	   if ((entry->schemaProgram->nstar == 1) &&
	       (entry->fitsProgram->nstar == 1) &&
	       (entry->fitsProgram->type == entry->schemaProgram->type))
	 	{
		if (entry->dimenProgram[0] == NULL && entry->fitsProgram->type
			== shTypeGetFromName("CHAR")) {
		   if ((schemaPtr = shProgramEval (schemaData,
			entry->schemaProgram)) != NULL)
	   	     entry->schemaProgram->sdim[0] =
			strlen(*((char **)schemaPtr))+1;
		   }
		}
	   entry->schemaProgram->nout = 1;
	   for (k=0; k<entry->schemaProgram->nstar; k++) {

/* If no program is supplied, schemaPtr should come back as NULL */
		if ((schemaPtr = shProgramEval (schemaData, \
		   entry->dimenProgram[k])) != NULL) entry->dimenCopy[k]
			(schemaPtr, &entry->schemaProgram->sdim[k]);
		entry->schemaProgram->nout *= entry->schemaProgram->sdim[k];
		}

	   nsch = entry->schemaProgram->nout * entry->schemaProgram->nitem;
	   if (nsch == 0) continue;

/* If FITS side is inline, make sure that its size matches schema side */
	   if (entry->fitsProgram->nstar == 0) {
		if (entry->fitsProgram->nitem != nsch) continue;
	   } else {

/* Determine size of heap space needed */
		*((int *)fitsData[j]) = nsch;
		}
	   }
	for (j=0; j<nfield; j++) {
	   entry = table->entry[j];
	   if (entry->fitsProgram == NULL) continue;
	   fitsData[j] += array[j]->data.incr * entry->fitsProgram->nitem;
	   }
	}
   shChainCursorDel (chain, cursor);

/* Now loop through FITS fields again and allocate heap space */
   for (j=0; j<nfield; j++) {
	if (entry->fitsProgram == NULL) continue;

/* Reset pointers to FITS data fields */
	fitsData[j] = (char *)(array[j]->data.dataPtr);
	fitsItemSize[j] = array[j]->data.incr;
	if (array[j]->data.schemaType != shTypeGetFromName("TBLHEAPDSC"))
	   continue;
/* If type is char, convert to str */
	type = entry->fitsProgram->type;
	if (type == shTypeGetFromName("CHAR")) type = shTypeGetFromName("STR");
	if (shTblFldHeapAlloc (array[j],
		   shNameGetFromType(type), 0, NULL)
		   != SH_SUCCESS) continue;
	fitsItemSize[j] = ((TBLFLD *)(array[j]->info))->heap.incr;
	}

/* Step through the chain once again, now copying data to heap. */

   cursor = shChainCursorNew(chain);
   for (i=0; i<size; i++) {
	schemaData = shChainWalk (chain, cursor, NEXT);

/* Loop through each attribute */   
	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fits2Schema == NULL || entry->schemaProgram == NULL
		|| entry->fitsProgram == NULL) continue;
	   if (entry->schemaProgram->nstar == 0 && entry->fitsProgram->nstar
		== 0) continue;

/* Fill in dimension info on schema side.  Need to do this all over again.
 * Info is stored in
 * schemaProgram->sdim.  Some entries may have been initialized already.
 * Use programs to fill in the rest.
*/

/* Special case: if FITS side is heap, schema side has nstar = 1, type = char,
 * take schema dimension to be length of FITS side.  This is particularly
 * useful for character strings: the heap side includes the trailing null.
*/
	   if ((entry->schemaProgram->nstar == 1) &&
	       (entry->fitsProgram->nstar == 1) &&
	       (entry->fitsProgram->type == entry->schemaProgram->type))
	 	{
		if (entry->dimenProgram[0] == NULL && entry->fitsProgram->type
			== shTypeGetFromName("CHAR")) {
		   if ((schemaPtr = shProgramEval (schemaData,
			entry->schemaProgram)) != NULL)
	   	     entry->schemaProgram->sdim[0] =
			strlen(*((char **)schemaPtr))+1;
		   }
		}
	   entry->schemaProgram->nout = 1;
	   for (k=0; k<entry->schemaProgram->nstar; k++) {

/* If no program is supplied, schemaPtr should come back as NULL */
		if ((schemaPtr = shProgramEval (schemaData,
		   entry->dimenProgram[k])) != NULL) entry->dimenCopy[k]
			(schemaPtr, &entry->schemaProgram->sdim[k]);
		entry->schemaProgram->nout *= entry->schemaProgram->sdim[k];
		}

/* Compute total length of FITS and SCHEMA sides and compare */
	   if (entry->fitsProgram->nstar > 0) {
		nfits = *((int *)fitsData[j]);
	   } else {
		nfits = entry->fitsProgram->nitem;
		}
	   nsch = entry->schemaProgram->nout * entry->schemaProgram->nitem;
	   if (nfits != nsch) continue;
	   if (nfits == 0 || nsch == 0) continue;

/* Now do the data copy.  First, reset data pointer */
	   if ((schemaPtr = shProgramEval (schemaData, entry->schemaProgram))
		== NULL) continue;
	   if ((fitsPtr  = shProgramEval (fitsData[j], entry->fitsProgram))
		== NULL) continue;

/* If FITS side is heap, dereference the pointer info matches the schema side
*/
	   if (entry->fitsProgram->nstar > 0) {
		fitsPtr = *((void **)(fitsPtr));
		}

/* Set up auxiliary array dimension info */
	   subdim[entry->schemaProgram->nstar-1] = 1;
	   for (k=1; k<entry->schemaProgram->nstar; k++) {
		subdim[entry->schemaProgram->nstar-k-1] =
		   subdim[entry->schemaProgram->nstar-k] * 
		   entry->schemaProgram->sdim[entry->schemaProgram->nstar-k];
		}

/* Set up a double loop based on the schema side dimension info, since it
 * can be the most complex.
*/
	   for (k=0; k<entry->schemaProgram->nitem; k++) {
		for (l=0; l<entry->schemaProgram->nout; l++) {
		   dataPtr = arrayLocate (schemaPtr,
			entry->schemaProgram->sdim, subdim,
			entry->schemaProgram->nstar, 0, l,
			entry->schemaProgram->size);
		   entry->schema2Fits (dataPtr, fitsPtr);
		   fitsPtr += fitsItemSize[j];
		   }
		schemaPtr += entry->schemaProgram->isize;
		}
	   }

/* Increment starting location in ARRAY data area */
	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fitsProgram == NULL) continue;
	   fitsData[j] += array[j]->data.incr * entry->fitsProgram->nitem;
	   }
	}
   shChainCursorDel (chain, cursor);
   shFree(array);
   shFree(fitsData);
   shFree(fitsItemSize);
   return TCL_OK;
}

/*****************************************************************************/
/***			tblCopyToChain					   ***/
/*****************************************************************************/

/* Run through a translation table and perform copy operation */

static int tblCopyToChain (
	TBLCOL *tblCol,		/* MODIFIED: Table col */
	CHAIN *chain,		/* IN: Source chain */
	TRANSTABLE *table)	/* IN: Translation table */

{
   ARRAY **array;
   TRANSENTRY *entry;
   char **fitsData;
   int *fitsItemSize;
   void *schemaData;
   void *dataPtr;
   int nfield, nentry;
   int i,j,k,l;
   char *schemaPtr = NULL;
   char *fitsPtr = NULL;
   CURSOR_T cursor;
   int size;
   int nsch;
   int nfits;
   int subdim[30];	/* Will we ever exceed 30 dimensions?! */

/* Quick error checks */
   if (shChainSize(chain) != tblCol->rowCnt) {
	shErrStackPush ("Chain size and tbl size don't match\n");
	return TCL_ERROR;
	}
   nfield = shChainSize(&tblCol->fld);
   nentry = table->nEntry;
   if (nfield == 0) return TCL_OK;
   if (nentry == 0) return TCL_OK;

/* Convert the chain in the tblCol to an array for faster access */
/* Allocate an array of pointers that will point to a row of elements in
 * the data areas of the fits fields
*/
   array = (ARRAY **)shMalloc (nentry * sizeof(ARRAY *));
   fitsData = (char **)shMalloc (nentry * sizeof(char *));
   fitsItemSize = (int *)shMalloc (nentry * sizeof(int));
   for (j=0; j<nentry; j++) {
	entry = table->entry[j];
	array[j] = shChainElementGetByPos (&tblCol->fld, entry->ifld);
	fitsData[j] = (char *)(array[j]->data.dataPtr);
	if (entry->fitsProgram == NULL) continue;
	if (entry->fitsProgram->nstar > 0) {
	   fitsItemSize[j] = ((TBLFLD *)(array[j]->info))->heap.incr;
	} else {
	   fitsItemSize[j] = array[j]->data.incr;
	   }
	}

/* Step through the chain.  At each step, call programs to compute pointers
 * to the starting bucket locations on each side.  Then enter a loop
 * to copy the data.
*/
   cursor = shChainCursorNew(chain);
   size = shChainSize(chain);
   for (i=0; i<size; i++) {
	schemaData = shChainWalk (chain, cursor, NEXT);

/* Loop through each attribute */   
/* We process non-heap stuff first */
	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fits2Schema == NULL || entry->schemaProgram == NULL
		|| entry->fitsProgram == NULL) continue;
	   if (entry->schemaProgram->nstar > 0) continue;

/* Get pointers to starting values */
	   if ((schemaPtr = shProgramEval (schemaData, entry->schemaProgram))
		== NULL) continue;
	   if ((fitsPtr  = shProgramEval (fitsData[j], entry->fitsProgram))
		== NULL) continue;

/* If FITS side is heap, dereference the pointer and make sure the length
 * info matches the schema side
*/
	   if (entry->fitsProgram->nstar > 0) {
		if (*(int *)(fitsData[j]) != entry->schemaProgram->nitem)
		   continue;
		fitsPtr = *((void **)(fitsPtr));
		}

   for (k=0; k<entry->schemaProgram->nitem; k++) {
		entry->fits2Schema (fitsPtr, schemaPtr);
		fitsPtr += fitsItemSize[j];
		schemaPtr += entry->schemaProgram->isize;
		}
	   }

/* Now deal with schemas that are out of line.
 * First, loop through entries and find those. */
	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fits2Schema == NULL || entry->schemaProgram == NULL
		|| entry->fitsProgram == NULL) continue;
	   if (entry->schemaProgram->nstar == 0) continue;

/* If heaplen is specified, copy FITS side heap length specifier to it. */
	   if (entry->heaplenProgram != NULL) {
		if ((schemaPtr = shProgramEval (schemaData, \
			entry->heaplenProgram)) == NULL) continue;

/* Heap side info is 1st item pointed to by fitsData[j], so use it */
		if (entry->heaplenCopy != NULL)
		   entry->heaplenCopy (fitsData[j], schemaPtr);
		}

/* Fill in dimension info on schema side.  Info is stored in
 * schemaProgram->sdim.  Some entries may have been initialized already.
 * Use programs to fill in the rest.
*/
	   entry->schemaProgram->nout = 1;

/* Compute total length of FITS and SCHEMA sides and compare */
	   if (entry->fitsProgram->nstar > 0) {
		nfits = *(int *)fitsData[j];
	   } else {
		nfits = entry->fitsProgram->nitem;
		}

/* Special case: if FITS side is heap, schema side has nstar = 1, type = char,
 * take schema dimension to be length of FITS side.  This is particularly
 * useful for character strings: the heap side includes the trailing null.
*/
	   if (entry->schemaProgram->nstar == 1 && entry->fitsProgram->nstar
		== 1 && entry->fitsProgram->type == entry->schemaProgram->type)
	 	{
		if (entry->dimenProgram[0] == NULL && entry->fitsProgram->type
		   == shTypeGetFromName("CHAR"))
		   entry->schemaProgram->sdim[0] = nfits;
		}

	   for (k=0; k<entry->schemaProgram->nstar; k++) {

/* If no program is supplied, schemaPtr should come back as NULL */
		if ((schemaPtr = shProgramEval (schemaData, \
		   entry->dimenProgram[k])) != NULL)
			entry->dimenCopy[k] (schemaPtr,
			   &entry->schemaProgram->sdim[k]);
		entry->schemaProgram->nout *= entry->schemaProgram->sdim[k];
		}

	   nsch = entry->schemaProgram->nout * entry->schemaProgram->nitem;
	   if (nsch == 0) continue;
	   if (nfits != nsch) continue;
	   if (nfits == 0 || nsch == 0) continue;

/* Get pointers to starting values */
	   if ((schemaPtr = shProgramEval (schemaData, entry->schemaProgram))
		== NULL) continue;
/* Construct destination array of pointers if not already in place */
	   for (k=0; k<entry->schemaProgram->nitem; k++) {
		if (*(void **)schemaPtr == NULL) {
		   *(void **)schemaPtr =
			arrayMalloc (entry->schemaProgram->sdim,
			entry->schemaProgram->nstar, 0,
			entry->schemaProgram->size);
		   }
		schemaPtr += entry->schemaProgram->isize;
		}
/* Now do the data copy.  First, reset data pointer */
	   if ((schemaPtr = shProgramEval (schemaData, entry->schemaProgram))
		== NULL) continue;
	   if ((fitsPtr  = shProgramEval (fitsData[j], entry->fitsProgram))
		== NULL) continue;

/* If FITS side is heap, dereference the pointer.  We already check for
 * length compatibility.
*/
	   if (entry->fitsProgram->nstar > 0) {
		fitsPtr = *((void **)(fitsPtr));
		}
/* Set up auxiliary array dimension info */
	   subdim[entry->schemaProgram->nstar-1] = 1;
	   for (k=1; k<entry->schemaProgram->nstar; k++) {
		subdim[entry->schemaProgram->nstar-k-1] =
		   subdim[entry->schemaProgram->nstar-k] * 
		   entry->schemaProgram->sdim[entry->schemaProgram->nstar-k];
		}

/* Set up a double loop based on the schema side dimension info, since it
 * can be the most complex.
*/
	   for (k=0; k<entry->schemaProgram->nitem; k++) {
		for (l=0; l<entry->schemaProgram->nout; l++) {
		   dataPtr = arrayLocate (schemaPtr,
			entry->schemaProgram->sdim, subdim,
			entry->schemaProgram->nstar, 	0, l,
			entry->schemaProgram->size);
		   entry->schema2Fits (fitsPtr, dataPtr);
		   fitsPtr += fitsItemSize[j];
		   }
		schemaPtr += entry->schemaProgram->isize;
		}
	   }
/* Increment starting location in ARRAY data area */

	for (j=0; j<nentry; j++) {
	   entry = table->entry[j];
	   if (entry->fitsProgram == NULL) continue;
	   fitsData[j] += array[j]->data.incr * entry->fitsProgram->nitem;
	   }
	}
   shChainCursorDel (chain, cursor);
   shFree(array);
   shFree(fitsData);		
   shFree(fitsItemSize);
   return TCL_OK;
}

/*****************************************************************************/
/***			tclTblToChain					   ***/
/***	Convert a TBLCOL to a chain based on a translation table	   ***/
/*****************************************************************************/


static char *g_tblToChain = "tblToChain";
static ftclArgvInfo g_tblToChainTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL, "Convert a TBLCOL to a CHAIN"},
    {"<TBLCOL>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a TBLCOL"},
    {"<TRANS>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a TRANSTABLE"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int
tclTblToChain (
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   char *tblColName = NULL;
   char *transTblName = NULL;

   TRANSTABLE *transTable = NULL;
   TBLCOL *tblCol = NULL;
   CHAIN *chain = NULL;
   char chainName[HANDLE_NAMELEN];
   int retStatus;


   g_tblToChainTbl[1].dst = &tblColName;
   g_tblToChainTbl[2].dst = &transTblName;

   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_tblToChainTbl,
                      0, g_tblToChain)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, tblColName, (void **)&tblCol,
	"TBLCOL") != TCL_OK) return TCL_ERROR;

   if (shTclAddrGetFromName(interp, transTblName, (void **)&transTable,
	"TRANSTABLE") != TCL_OK) return TCL_ERROR;

/* Done processing handles. Begin the action */
   if (transTableCompile (transTable, tblCol) != TCL_OK) {
	Tcl_SetResult (interp, "Can't compile table", TCL_STATIC);
	goto error;
	}

   if ((chain = chainNewFromTrans (interp, transTable, tblCol->rowCnt))
	== NULL) goto error;

   if (tblCopyToChain (tblCol, chain, transTable) != TCL_OK) {
	printf ("Can't copy chain\n");
	goto error;
	}
   transTableUncompile (transTable);

   if (shTclHandleNew(interp, chainName, "CHAIN", chain) != TCL_OK)
	return TCL_ERROR;

   Tcl_SetResult (interp, chainName, TCL_VOLATILE);
   return TCL_OK;

error:
   if (chain != NULL) shGenericChainDel(chain);
   transTableUncompile(transTable);
   return TCL_ERROR;

}


/*****************************************************************************/
/***			tclChainToTbl					   ***/
/***	Convert a chain to a TBLCOL based on a translation table	   ***/
/*****************************************************************************/


static char *g_chainToTbl = "chainToTbl";
static ftclArgvInfo g_chainToTblTbl[] = {
    {NULL, FTCL_ARGV_HELP, NULL, NULL, "Convert a CHAIN to a TBLCOL"},
    {"<CHAIN>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a CHAIN"},
    {"<TRANS>",  FTCL_ARGV_STRING, NULL, NULL, "Handle to a TRANSTABLE"},
    {NULL,        FTCL_ARGV_END, NULL, NULL, NULL}
   };

static int
tclChainToTbl (
          ClientData clientData,
          Tcl_Interp *interp,
          int argc,
          char **argv
          )

{
   char *chainName = NULL;
   char *transTblName = NULL;

   TRANSTABLE *transTable = NULL;
   TBLCOL *tblCol = NULL;
   CHAIN *chain;
   char tblColName[HANDLE_NAMELEN];
   int retStatus;


   g_chainToTblTbl[1].dst = &chainName;
   g_chainToTblTbl[2].dst = &transTblName;

   if ((retStatus = shTclParseArgv(interp, &argc, argv, g_chainToTblTbl,
                      0, g_chainToTbl)) != FTCL_ARGV_SUCCESS)
        return retStatus;

   if (shTclAddrGetFromName(interp, transTblName, (void **)&transTable,
	"TRANSTABLE") != TCL_OK) return TCL_ERROR;

   if (shTclChainAddrGetFromName(interp, chainName, transTable->schemaTypeName,
	&chain) != TCL_OK) return TCL_ERROR;

/* Done processing handles. Begin the action */
   if ((tblCol = tblColNewFromTrans (transTable, shChainSize(chain)))
	== NULL) goto error;
   if (transTableCompile (transTable, tblCol) != TCL_OK) goto error;
   if (chainCopyToTbl (tblCol, chain, transTable) != TCL_OK) goto error;
    transTableUncompile (transTable);

   if (shTclHandleNew(interp, tblColName, "TBLCOL", tblCol) != TCL_OK)
	return TCL_ERROR;
   Tcl_SetResult (interp, tblColName, TCL_VOLATILE);

   return TCL_OK;

error:
   if (tblCol != NULL) shTblcolDel(tblCol);
   transTableUncompile (transTable);
   return TCL_ERROR;
}


/*****************************************************************************/

/*
 * Declare my new tcl verbs to tcl
 */
void
shTclTransTableDeclare (Tcl_Interp *interp) 
{
   shTclDeclare(interp, g_transTableNew,
	(Tcl_CmdProc *)tclTransTableNew,
	(ClientData) 0,
	(Tcl_CmdDeleteProc *)NULL, 
	module,
	shTclGetArgInfo(interp, g_transTableNewTbl, 0, g_transTableNew),
	shTclGetUsage(interp, g_transTableNewTbl, 0, g_transTableNew));

   shTclDeclare(interp, g_transEntryAdd,
	(Tcl_CmdProc *)tclTransEntryAdd,
	(ClientData) 0,
	(Tcl_CmdDeleteProc *)NULL, 
	module,
	shTclGetArgInfo(interp, g_transEntryAddTbl, 0, g_transEntryAdd),
	shTclGetUsage(interp, g_transEntryAddTbl, 0, g_transEntryAdd));

   shTclDeclare(interp, g_transEntryDel,
	(Tcl_CmdProc *)tclTransEntryDel,
	(ClientData) 0,
	(Tcl_CmdDeleteProc *)NULL, 
	module,
	shTclGetArgInfo(interp, g_transEntryDelTbl, 0, g_transEntryDel),
	shTclGetUsage(interp, g_transEntryDelTbl, 0, g_transEntryDel));

   shTclDeclare(interp, g_transTableDel,
	(Tcl_CmdProc *)tclTransTableDel,
	(ClientData) 0,
	(Tcl_CmdDeleteProc *)NULL, 
	module,
	shTclGetArgInfo(interp, g_transTableDelTbl, 0, g_transTableDel),
	shTclGetUsage(interp, g_transTableDelTbl, 0, g_transTableDel));

   shTclDeclare(interp, g_tblToChain,
	(Tcl_CmdProc *)tclTblToChain,
	(ClientData) 0,
	(Tcl_CmdDeleteProc *)NULL, 
	module,
	shTclGetArgInfo(interp, g_tblToChainTbl, 0, g_tblToChain),
	shTclGetUsage(interp, g_tblToChainTbl, 0, g_tblToChain));

   shTclDeclare(interp, g_chainToTbl,
	(Tcl_CmdProc *)tclChainToTbl,
	(ClientData) 0,
	(Tcl_CmdDeleteProc *)NULL, 
	module,
	shTclGetArgInfo(interp, g_chainToTblTbl, 0, g_chainToTbl),
	shTclGetUsage(interp, g_chainToTblTbl, 0, g_chainToTbl));

   return;
}

