# Read and verify FITS table files
# Top level routines are as follows:
#	fitsVerify <fits-file>

################################################################################
# Routines for reading a FITS Table file

#Get size of an array.  Return 0 if it does not exist
proc arrsize {var} {
   upvar 1 $var local
   if {[catch {set size [array size local]}]} then {return 0} else {
                return $size}
   }

########################################################################
#Given a header, find the first line no. with a given keyword
proc HdrGetLineNo {inheader keyword} {
   upvar 1 $inheader header
   set len [arrsize header]
   for {set i 1} {$i <= $len} {incr i} {
	if {[HdrKeyGet $header($i)] == $keyword} then {return $i}
	}
   return 0
   }

########################################################################
#Extract the keyword field
proc HdrKeyGet {line} {
   return [lindex [string range $line 0 7] 0]
   }

########################################################################
#Extract the parameter type.  This is REEEALLL simple - N or A
proc HdrParTypeGet {line} {
   set line [string range $line 10 end]
#Get rid of leading and trailing spaces.
   set line [string trim $line " "]
   if {[string index $line 0] != {'} } then {
	return I
   } else {
	return A
	}
   }

########################################################################
#Extract the parameter field
proc HdrParGet {line} {
   set line [string range $line 10 end]
#Get rid of leading and trailing spaces.
   set line [string trim $line " "]
   if {[string index $line 0] != {'} } then {
	return [lindex $line 0]
	}
   set line [string range $line 1 end]
   set len [string length $line]
   set outline {}
   set flag 0
   for {set i 0} {$i < $len} {incr i} {
	set char [string index $line $i]
	if {$char == {"} } then {
	   append outline {\"}
	   continue
	   }
	if {$char == {'}} then {
	   if {$flag == 1} then {
		append outline {'}
		set flag 0
		continue
		}
	   set flag 1
	   } else {
	   if {$flag == 1} then {
		break}
	   append outline $char
	   continue
	   }
	}
   if {$flag == 1} then {return $outline} else  {
# Else error. For now, return the rest of the line as if the last quote were
# present
	return $outline}
   }

########################################################################
#Extract the comment field
proc HdrComGet {line} {
   global VERBOSE
   set line [string range $line 10 end]
#Fits recommends but does not mandate that character strings begin in col.11
#Hence the need to trim blanks
   set line [string trim $line " "]
   if {[string index $line 0] != {'} } then {
	lvarpop line
	set line [string trimleft $line / ]
	return $line
	}
   set line [string range $line 1 end]
   set len [string length $line]
   set flag 0
   for {set i 0} {$i < $len} {incr i} {
	set char [string index $line $i]
	if {$char == {'}} then {
	   if {$flag == 1} then {
		set flag 0
		continue
		}
	   set flag 1
	   } else {
	   if {$flag == 1} then {
		break}
	   continue
	   }
	}
   if {$flag == 1} then {
#Look for leading /
	set line [string range $line $i end]
	if {$line == ""} return
	set size [string length $line]
	for {set i 0} {$i < $size} {incr i} {
	   set char [string index $line $i]
	   if {$char == " "} continue
	   if {$char == "/"} then {
		set retstring [string range $line [expr {$i+1}] end]
		return [string trim $retstring " "]
	   } else {
		set retstring [string range $line $i end]
		return [string trim $retstring " "]
		}
	   }
	return ""
# Else error. For now, return a null string as if the last quote were present
   } else {
	if {[info exists VERBOSE]} then {echo  Error in decoding comment}
	return ""
	}
   }

########################################################################
#Read a header from a FITS file, given a file handle
proc HdrRead {fhandle inheader} {
   upvar 1 $inheader header
   catch {unset header}
   set nline 0
   while {1} {
	for {set i 0} {$i < 36} {incr i} {
	   catch {unset line}
	   set line [read $fhandle 80]
	   if {[eof $fhandle]} {
		close $fhandle
		error "Unexpected eof in HdrRead"
		}
	   if {[HdrKeyGet $line] == "COMMENT"} then {continue}
	   if {[HdrKeyGet $line] == "HISTORY"} then {continue}
	   incr nline
	   set header($nline) $line
	   }
	if {[HdrGetLineNo header END] > 0} then  {return}
	}
   }

########################################################################
#Extract an ascii field from a header, given a keyword
proc HdrGetAsAscii {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	

########################################################################
#Extract an integer field from a header, given a keyword
proc HdrGetAsInt {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	

########################################################################
#Extract a double field from a header, given a keyword
proc HdrGetAsDouble {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	

########################################################################
#Extract a logical field from a header, given a keyword
proc HdrGetAsLogical {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	


########################################################################
#### Read header and return a multitude of decoded header information. ###

proc fitsPduHdrDecode {inheader} {
   global VERBOSE
   upvar 1 $inheader header
   upvar 1 NROW NROW
   upvar 1 NCOL NCOL
   upvar 1 NBYTE NBYTE
   upvar 1 PCOUNT PCOUNT
   upvar 1 EXTEND EXTEND
#Init
   set PCOUNT 0
   set EXTEND 1
   if {[HdrGetAsLogical header SIMPLE] != "T"} then {
	if {[info exists VERBOSE]} then {
	   echo  "Not a SIMPLE=T file; proceeding anyway"
	   }
	}
   if {[HdrGetAsLogical header EXTEND] != "T"} then {
	set EXTEND 0
	}
   set NBYTE [int [fabs [expr [HdrGetAsInt header BITPIX]/8]]]
   set NAXIS [HdrGetAsInt header NAXIS]
   if {  $NAXIS > 3} then {
	error "Too many axes"
	}
   if {$NAXIS >= 0} then {
	set NROW 0
	set NCOL 0
	}
   if {$NAXIS >= 1} then {
	set NROW 1
	set NCOL [HdrGetAsInt header NAXIS1]
	}
   if {$NAXIS >= 2} then {
	set NROW [HdrGetAsInt header NAXIS2]
	}
#For extra axes, scale NBYTE appropriately.  This is a hack, but I am
#not interested in the details of NBTYE, just that the total number of
#data bytes is correct
   if {$NAXIS >= 3} then {
	set extra [HdrGetAsInt header NAXIS3]
	set NBYTE [expr $NBYTE*$extra]
	}
   }

########################################################################
#Decode ascii table header

proc fitsTableHdrDecode {inheader} {
#The following variables are passed back to the calling routine by reference.
#Too many to pass as dummy arguments, so the calling program must be prepared
#to accept their absolute names
   global VERBOSE
   upvar 1 $inheader header
   upvar 1 XTENSION XTENSION
   upvar 1 NROW NROW
   upvar 1 NCOL NCOL
   upvar 1 NBYTE NBYTE
   upvar 1 PCOUNT PCOUNT
#Extract various fields
   set XTENSION [HdrGetAsAscii header XTENSION]
   if {$XTENSION != "BINTABLE" && \
	$XTENSION != "TABLE   " && \
	$XTENSION != "IMAGE   "
   } then {
	error "Not FITS Table extension; found [HdrGetAsAscii hdr XTENSION]"
	}
   set NCOL [HdrGetAsInt header NAXIS1]
   set NROW [HdrGetAsInt header NAXIS2]
#   set NBYTE 1
   set NBYTE [int [fabs [expr [HdrGetAsInt header BITPIX]/8]]]
   set GCOUNT [HdrGetAsInt header GCOUNT]
   set PCOUNT [HdrGetAsInt header PCOUNT]
   if {$GCOUNT != 1} then {
	error "Unclear how to decode files with GCOUNT != 1"
	}
   }

########################################################################
#Print out non-reserved header keywords
proc fitsTableHdrPrint {inheader} {

#The following variables are passed back to the calling routine by reference.
#Too many to pass as dummy arguments, so the calling program must be prepared
#to accept their absolute names
   global VERBOSE
   upvar 1 $inheader header
#Delete all reserved lines, then print the reset
   set nfield [HdrGetAsInt header TFIELDS]
   set keywords {XTENSION BITPIX NAXIS NAXIS1 NAXIS2 PCOUNT GCOUNT TFIELDS}
   set zaplist ""
   foreach keyword $keywords {
	lappend zaplist [HdrGetLineNo header $keyword]
	}
   loop k 1 [expr $nfield+1] {
	lappend zaplist [HdrGetLineNo header TFORM$k]
	lappend zaplist [HdrGetLineNo header TTYPE$k]
	lappend zaplist [HdrGetLineNo header TDIM$k]
	}
   foreach zap $zaplist {
	catch {unset header($zap)}
	}
   set nums [array names header]
   set nums [lsort -integer $nums]
   foreach num $nums {
	if {[string trim $header($num)] == ""} continue
	echo "   [string trim $header($num)]"
	}
   }

########################################################################
proc fitsDataRead {infileid} {
#Total number of bytes of data
   global VERBOSE
   upvar 1 NROW NROW
   upvar 1 NCOL NCOL
   upvar 1 NBYTE NBYTE
   upvar 1 PCOUNT PCOUNT
   set ntot [expr $NROW*$NCOL*$NBYTE+$PCOUNT]
#Padding
   set PAD [expr (int(($ntot-1)/2880)+1)*2880-$ntot]
   if {[info exists VERBOSE]} then {echo "   DATA = $ntot PADDING = $PAD"}
   set sum [expr $ntot+$PAD]
   if {[info exists VERBOSE]} then {echo  "   SUM   " = $sum}
   flush stdout
   set begin [tell $infileid]
#Go to end of file first and get total number of bytes
   seek $infileid 0 end
   set end [tell $infileid]
#Now seek proper position
   seek $infileid $begin start
   seek $infileid $sum current
#Arithmetic
   set actual [expr $end-$begin]
   if {[info exists VERBOSE]} then {echo  "   ACTUAL" = $actual}
   if {$actual < $sum} then {
	if {[info exists VERBOSE]} then {
	   echo  "   Error: Shortfall of [expr $sum-$actual]"
	   }
	return 0
	}
   if {$actual > $sum} then {
	if {[info exists VERBOSE]} then {echo  "   Table extension may follow."}
	return 1
	}
   return 0
   }	

########################################################################
#### Main Routine to read a FITS binary table and verify ###############

proc fitsVerify {infile} {
   global VERBOSE
   set VERBOSE 1
   set infileid [open $infile]
#   fconfigure $infileid -blocking no
   HdrRead $infileid header
   fitsPduHdrDecode header
# Multitude of parameters were returned from fitsTableHdrRead by reference
   if {[info exists VERBOSE]} then {echo  ""}
   if {[info exists VERBOSE]} then {echo  Primary Data Unit}
   if {[info exists VERBOSE]} then {
	echo  "   NROW = $NROW NCOL = $NCOL PCOUNT = $PCOUNT"
	}
   set more [fitsDataRead $infileid]
   if {$more == 0} then {
	close $infileid
	return
	}
   if {$EXTEND == 0} then {
	if {[info exists VERBOSE]} then {echo  No extension present}
	close $infileid
	return
	}
   set i 0
   while {1} {
	incr i
	if {[info exists VERBOSE]} then {echo  ""}
	if {[info exists VERBOSE]} then {echo  Extension hdu $i}
	if {[info exists VERBOSE]} then {
	   echo  "   Beginning position is [tell $infileid]"
	   }
	HdrRead $infileid header
	fitsTableHdrDecode header
#Echo non-reserved FITS keywords
	fitsTableHdrPrint header
# Multitude of parameters were returned from fitsTableHdrRead by reference
	if {[info exists VERBOSE]} then {
	   echo  "   XTENSION = $XTENSION"
	   echo  "   NROW = $NROW NCOL = $NCOL NBYTE = $NBYTE PCOUNT = $PCOUNT"
	   }
	set more [fitsDataRead $infileid]
	if {$more == 0} then {
	   close $infileid
	   return
	   }
	}
   }

########################################################################
proc dummy {} {
# Read the data - see if there is enough
   set schema(1) "FITSROW STRUCT 1 $NCOL"
   set schemaElem(1)(1) "row char 0 $NCOL 0"
   if {$PCOUNT > 0} then {
	set schema(2) "FITSHEAP STRUCT 1 $PCOUNT"
	set schemaElem(2)(1) "heap char 0 $PCOUNT 0"
	}
   if {$PAD > 0} then {
	set schema(3) "FITSPAD STRUCT 1 $PAD"
	set schemaElem(3)(1) "pad char 0 $PAD 0"
	}
   schemaDefine schema schemaElem
   set row [genericNew FITSROW]
   for {set j 1} {$j <= $NROW} {incr j} {
	binRead $infileid $row
	}
   if {$PCOUNT > 0} then {
	set heap [genericNew FITSHEAP]
	binRead $infileid $heap
	}
   if {$PAD > 0} then {
	set pad [genericNew FITSPAD]
	binRead $infileid $pad
	}
   close $infileid
   }
