# Read and Write FITS table files
# Top level routines are as follows:
# Old name:
#   fitsTableOut <format-file> <data-file> <outfile>
# Now renamed to
#   ascii2Fits <format-file> <data-file> <outfile>
# Reverse of above
#   fits2Ascii <fits-file> <ascii-file> [<format-file-out>]
# Template routine for reading a FITS file and storing.
#   fitsTableRead <fits-file> header data
# The last routine will call an action routine to actually do something with
# the data as it is read in. header and data are variables that will be filled
# with data read from the file.  The following global variables are defined
# from the header:
# NROW
# NCOL
# TFIELDS
# TTYPE(1 to TFIELDS)
# TBCOL(1 to TFIELDS)
# TBFORM(1 to TFIELDS)
# TBWIDTH(1 to TFIELDS)
# TBPREC(1 to TFIELDS)
# (Redundant)
# TBCOL1(1 to TFIELDS)
# TBCOL2(1 to TFIELDS)
# The user variable header is filled in as an array with the index going from
# 1 to nline.  The user variable data is filled in with the returned data
# as a keyed list.
#
#First, routines for writing.

# Pad an input string with blanks to size characters.  Truncate if over.

proc pad {instring size} {
   set len [expr {$size - [string length $instring]}]
   for {set i 1} {$i <= $len} {incr i} {
	append instring " "
	}
   return [string range $instring 0 [expr {$size - 1}]]
   }

proc pad80 {instring} {
   return [pad $instring 80]
   }

#Make FITS header lines
proc HdrMakeLineWithAscii {keyword body {comment ""}} {
   set keyword [string range $keyword 0 7]
   set line [format "%-8s= '%-8s'" $keyword $body]
   set line [format "%-30s / %s" $line $comment]
   }

#Make FITS header lines
proc HdrMakeLineWithInt {keyword body {comment ""}} {
   set keyword [string range $keyword 0 7]
   set line [format "%-8s= %20s" $keyword $body]
   set line [format "%-30s / %s" $line $comment]
   }

#Make FITS header lines
proc HdrMakeLineWithDouble {keyword body {comment ""}} {
   set keyword [string range $keyword 0 7]
   set line [format "%-8s= %20s" $keyword $body]
   set line [format "%-30s / %s" $line $comment]
   }

#Make FITS header lines
proc HdrMakeLineWithLogical {keyword body {comment ""}} {
   set keyword [string range $keyword 0 7]
   set line [format "%-8s= %20s" $keyword $body]
   set line [format "%-30s / %s" $line $comment]
   }

#Get size of an array.  Return 0 if it does not exist
proc arrsize {var} {
   upvar 1 $var local
   if {[catch {set size [array size local]}]} then {return 0} else {
		return $size}
   }


#Add a line to the header in ASCII format
# The following procs call the MakeLine procs and append the line to a
# header array

proc HdrAscii {inheader keyword body {comment ""}} {
   upvar 1 $inheader header
   set nline [arrsize header]
   incr nline
   set header($nline) [pad80 [HdrMakeLineWithAscii $keyword $body $comment]]
   }

#Add a line to the header in INT format
proc HdrInt {inheader keyword body {comment ""}} {
   upvar 1 $inheader header
   set nline [arrsize header]
   incr nline
   set header($nline) [pad80 [HdrMakeLineWithInt $keyword $body $comment]]
   }

#Add a line to the header in DOUBLE format
proc HdrDouble {inheader keyword body {comment ""}} {
   upvar 1 $inheader header
   set nline [arrsize header]
   incr nline
   set header($nline) [pad80 [HdrMakeLineWithDouble $keyword $body $comment]]
   }

#Add a line to the header in Logical format
proc HdrLogical {inheader keyword body {comment ""}} {
   upvar 1 $inheader header
   set nline [arrsize header]
   incr nline
   set header($nline) [pad80 [HdrMakeLineWithLogical $keyword $body $comment]]
   }

#Add a comment to the header.
proc HdrComment {inheader {comment ""}} {
   upvar 1 $inheader header
   set nline [arrsize header]
   incr nline
   set header($nline) [pad80 "COMMENT $comment"]
   }

#Add an END line to the header.
proc HdrEnd {inheader} {
   upvar 1 $inheader header
   set nline [arrsize header]
   incr nline
   set header($nline) [pad80 "END"]
   }


#Add blank lines to a header to pad out to a 2880 byte record
proc HdrFill {inheader} {
   upvar 1 $inheader header
   set nline [arrsize header]
   while { [expr {($nline/36)*36 != $nline}]} {
	incr nline
	set header($nline) [pad80 ""]
	}
   }

#Add 3 lines to define a field heading
proc HdrFieldHead {inheader inhead type col format comment} {
   upvar 1 $inheader header
   upvar 1 $inhead nhead
   incr nhead
   HdrAscii header TTYPE$nhead $type $comment
   HdrInt header TBCOL$nhead $col "Start column"
   HdrAscii header TFORM$nhead $format "Fortran format"
   }

#End of generic helper procs

#I assume that I have an ascii data file with the desired data to be
#inserted a FITS Table format file.  I assume that any comments and
#the field heading definitions are in a second file.  The following
#two routines read the second file and insert the definitions into the
#header.

# Read a file with header definition commands.  Extract field headings
proc ReadFieldHeadToHeader {inheader file incount} {
   upvar 1 $inheader header
   upvar 1 $incount nhead
   set nhead 0
   set fileid [open $file]
   set colmax 0
   while { [gets $fileid list] != -1} {
	   set keyword [lvarpop list]
	if { $keyword == "FIELD"} then {
	   set type [lvarpop list]
	   set start [lvarpop list]
	   set format [lvarpop list]
	   set comment $list
	   HdrFieldHead header nhead $type $start $format $comment
#Decode the field width and find the last column
	   set type [string index $format 0]
	   set field [string range $format 1 end]
	   set field [split $field "."]
	   set width [lindex $field 0]
	   set colmax [max $colmax [expr { $start + $width - 1}]]
	   }
	}
   close $fileid
   echo colmax = $colmax
   return $colmax
   }

# Read a file with header definition commands.  Extract comments and insert in
# header

proc ReadCommentToHeader {inheader file} {
   upvar 1 $inheader header
   set fileid [open $file]
   while { [gets $fileid list] != -1} {
# Some comment lines have double quote marks which will baffle any list
# manipulation routine.  Be more cautious and use string extraction stuff
	if { [string range $list 0 7] == "COMMENT "} then {
	   set list [string range $list 8 end]
	   HdrComment header $list
	   }
	}
   close $fileid
   }

#Top level routine to create a FITS Table header.  I assume that I
#read comments and field headings from a separate file.
#Header is built up in memory and written en mass to the output file.
#After data is read and written, we go back and reset the NROW and NCOL
#entries.

proc HdrCreate {inheader file} {
#Read critical parameters from a file
   upvar 1 $inheader header
   upvar 1 saveNAXIS1 saveNAXIS1
   upvar 1 saveNAXIS2 saveNAXIS2
   set nhead 0
   catch {unset header}
#Basic FITS header
   HdrLogical header SIMPLE T
   HdrInt header BITPIX 8
   HdrInt header NAXIS 0
   HdrLogical header EXTEND T
   ReadCommentToHeader header $file
   HdrEnd header
   HdrFill header
   
#Table Extension header
   HdrAscii header XTENSION TABLE "Table extension"
   HdrInt header BITPIX 8 "Character data"
   HdrInt header NAXIS 2 "Simple 2-D matrix"
   HdrInt header NAXIS1  84 "Number of characters per row"
#Save current position for use later in updating NCOL
   set saveNAXIS1 [arrsize header]
   HdrInt header NAXIS2  0 "Number of rows in the table"
#Save current position for use later in updating NROW
   set saveNAXIS2 [arrsize header]
   HdrInt header PCOUNT  0 "No random parameters"
   HdrInt header GCOUNT  1 "Only one group"
   HdrInt header TFIELDS 0 "Number of fields per row"
#Save current position for use later in updating TFIELDS
   set saveTFIELDS [arrsize header]

#Field headings.  Read from a file

   set colmax [ReadFieldHeadToHeader header $file nhead]
#Reset the TFIELDS parameter
   set header($saveTFIELDS) [pad80 [HdrMakeLineWithInt \
		TFIELDS $nhead "Number of fields per row"]]
   HdrEnd header
   HdrFill header
   return $colmax
   }

# Top Level procedure to write a FITS Table format file
proc ascii2Fits {hdrfile datfile outfile} {
   set colmax [HdrCreate header $hdrfile]
   echo Line size is $colmax bytes
   set fileid [open $outfile w]
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	puts $fileid "$header($i)" nonewline
	}
#Read data and write to output file
   set datafile [open $datfile]
   set nrow 0
   set nbyte 0
   while {[gets $datafile line] != -1} {
	set ncol [string length $line]
	set nbyte [expr {$nbyte + $colmax}]
	incr nrow
	puts $fileid [pad $line $colmax] nonewline
	if { ($nrow/1000) * 1000 == $nrow} then {echo Done with $nrow lines}
	}
   close $datafile
#pad to fill out a full record
   while { [expr {($nbyte/2880)*2880 != $nbyte}]} {
	puts $fileid " " nonewline
	incr nbyte
	}
#Reset NROW and NCOL
   seek $fileid [expr { ($saveNAXIS1-1)*80 }]
   puts $fileid [pad80 [HdrMakeLineWithInt NAXIS1 $colmax \
      "Number of characters per row"]] nonewline
   seek $fileid [expr { ($saveNAXIS2-1)*80 }]
   puts $fileid [pad80 [HdrMakeLineWithInt NAXIS2 $nrow \
      "Number of rows"]] nonewline
   close $fileid
   }

################################################################################
# Routines for reading a FITS Table file

#Given a header, find the first line no. with a given keyword
proc HdrGetLineNo {inheader keyword} {
   upvar 1 $inheader header
   set len [arrsize header]
   for {set i 1} {$i <= $len} {incr i} {
	if {[HdrKeyGet $header($i)] == $keyword} then {return $i}
	}
   return 0
   }

#Extract the keyword field
proc HdrKeyGet {line} {
   return [lindex [string range $line 0 7] 0]
   }

#Extract the parameter type.  This is REEEALLL simple - N or A
proc HdrParTypeGet {line} {
   set line [string range $line 10 end]
#Get rid of leading and trailing spaces.
   set line [string trim $line " "]
   if {[string index $line 0] != {'} } then {
	return I
   } else {
	return A
	}
   }

#Extract the parameter field
proc HdrParGet {line} {
   set line [string range $line 10 end]
#Get rid of leading and trailing spaces.
   set line [string trim $line " "]
   if {[string index $line 0] != {'} } then {
	return [lindex $line 0]
	}
   set line [string range $line 1 end]
   set len [string length $line]
   set outline {}
   set flag 0
   for {set i 0} {$i < $len} {incr i} {
	set char [string index $line $i]
	if {$char == {"} } then {
	   append outline {\"}
	   continue
	   }
	if {$char == {'}} then {
	   if {$flag == 1} then {
		append outline {'}
		set flag 0
		continue
		}
	   set flag 1
	   } else {
	   if {$flag == 1} then {
		break}
	   append outline $char
	   continue
	   }
	}
   if {$flag == 1} then {return $outline} else  {
# Else error. For now, return the rest of the line as if the last quote were
# present
	return $outline}
   }

#Extract the comment field
proc HdrComGet {line} {
   set line [string range $line 10 end]
#Fits recommends but does not mandate that character strings begin in col.11
#Hence the need to trim blanks
   set line [string trim $line " "]
   if {[string index $line 0] != {'} } then {
	lvarpop line
	set line [string trimleft $line / ]
	return $line
	}
   set line [string range $line 1 end]
   set len [string length $line]
   set flag 0
   for {set i 0} {$i < $len} {incr i} {
	set char [string index $line $i]
	if {$char == {'}} then {
	   if {$flag == 1} then {
		set flag 0
		continue
		}
	   set flag 1
	   } else {
	   if {$flag == 1} then {
		break}
	   continue
	   }
	}
   if {$flag == 1} then {
#Look for leading /
	set line [string range $line $i end]
	if {$line == ""} return
	set size [string length $line]
	for {set i 0} {$i < $size} {incr i} {
	   set char [string index $line $i]
	   if {$char == " "} continue
	   if {$char == "/"} then {
		set retstring [string range $line [expr {$i+1}] end]
		return [string trim $retstring " "]
	   } else {
		set retstring [string range $line $i end]
		return [string trim $retstring " "]
		}
	   }
	return ""
# Else error. For now, return a null string as if the last quote were present
   } else {
echo Error in decoding comment
	return ""
	}
   }

#Read a header from a FITS file, given a file handle
proc HdrRead {fhandle inheader} {
   upvar 1 $inheader header
   catch {unset header}
   set nline 0
   while {1} {
	for {set i 0} {$i < 36} {incr i} {
	   catch {unset line}
	   set line [read $fhandle 80]
	   if {[HdrKeyGet $line] == "COMMENT"} then {continue}
	   if {[HdrKeyGet $line] == "HISTORY"} then {continue}
	   incr nline
	   set header($nline) $line
	   }
	if {[HdrGetLineNo header END] > 0} then  {return}
	}
   }

#Extract an ascii field from a header, given a keyword
proc HdrGetAsAscii {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	

#Extract an integer field from a header, given a keyword
proc HdrGetAsInt {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	

#Extract a double field from a header, given a keyword
proc HdrGetAsDouble {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	

#Extract a logical field from a header, given a keyword
proc HdrGetAsLogical {inheader keyword} {
   upvar 1 $inheader header
   set nline [arrsize header]
   for {set i 1} {$i <= $nline} {incr i} {
	if {[HdrKeyGet $header($i)] != $keyword} then {continue}
	return [HdrParGet $header($i)]
	}
   return ""
   }	


#### Read header and return a multitude of decoded header information. ###

proc fitsTableHdrRead {fileid inheader} {
   upvar 1 $inheader header
   catch {unset header}
#Read main header
   HdrRead $fileid header
   if {[HdrGetAsLogical header SIMPLE] != "T"} then {
	error "Not a SIMPLE=T file"
	}
   if {[HdrGetAsLogical header EXTEND] != "T"} then {
	error "Not FITS Table format"
	}
   if {[HdrGetAsInt header NAXIS] != 0} then {
	error "Main Data Field not empty; cannot process yet"
	}
#Read extension header
   HdrRead $fileid header
   if {[HdrGetAsAscii header XTENSION] != "TABLE   "} then {
	error "Not FITS Table extension"
	}
   }

#Decode ascii table header
proc fitsTableHdrDecode {inheader} {
#The following variables are passed back to the calling routine by reference.
#Too many to pass as dummy arguments, so the calling program must be prepared
#to accept their absolute names
   upvar 1 $inheader header
   upvar 1 NROW NROW
   upvar 1 NCOL NCOL
   upvar 1 TFIELDS TFIELDS
   upvar 1 TTYPE TTYPE
   upvar 1 TTYPECOM TTYPECOM
   upvar 1 TBCOL TBCOL
   upvar 1 TBFORM TBFORM
   upvar 1 TBWIDTH TBWIDTH
   upvar 1 TBPREC TBPREC
#Redundant versions of above
   upvar 1 TBCOL1 TBCOL1
   upvar 1 TBCOL2 TBCOL2
   catch {unset TTYPE}
   catch {unset TBCOL}
   catch {unset TBFORM}
   catch {unset TBWIDTH}
   catch {unset TBCOL1}
   catch {unset TBCOL2}
   catch {unset TBPREC}
#Extract various fields
   set NCOL [HdrGetAsInt header NAXIS1]
   set NROW [HdrGetAsInt header NAXIS2]
   set TFIELDS [HdrGetAsInt header TFIELDS]
   set GCOUNT [HdrGetAsInt header GCOUNT]
   set PCOUNT [HdrGetAsInt header PCOUNT]
   if {$GCOUNT != 1} then {
	error "Unclear how to decode files with GCOUNT != 1"
	}
   if {$PCOUNT != 0} then {
	error "Unclear how to decode files with PCOUNT != 0"
	}
#Decode parameters of various fields
   for {set i 1} {$i <= $TFIELDS} {incr i} {
	set TTYPE($i) [HdrGetAsAscii header TTYPE$i]
	if {$TTYPE($i) == ""} then {
	   set TTYPE($i) TTYPE$i
	   set TTYPECOM($i) ""
	} else {
	   set lineno [HdrGetLineNo header TTYPE$i]
	   set TTYPECOM($i) [HdrComGet $header($lineno)]
	   }
	set TTYPE($i) [string trim $TTYPE($i)]
	set TBCOL($i) [HdrGetAsInt header TBCOL$i]
	set TFORM [HdrGetAsAscii header TFORM$i]
#Decode the field width
	set type [string index $TFORM 0]
	set TBFORM($i) $type
	set field [string range $TFORM 1 end]
	if {$type == "I" || $type == "A"} then {
	   set TBWIDTH($i) [lindex $field 0]
	   set TBPREC($i) ""
	   } else {
	   set field [split $field "."]
	   set TBWIDTH($i) [lindex $field 0]
	   set TBPREC($i) [lindex $field 1]
	   }
#Useful accessory parameters for decoding data
	set TBCOL1($i) [expr {$TBCOL($i) - 1}]
	set TBCOL2($i) [expr {$TBCOL($i) + $TBWIDTH($i) - 2}]
#	echo TTYPE($i) = \"$TTYPE($i)\"
#	echo TBCOL($i) = \"$TBCOL($i)\"
#	echo TFORM($i) = \"$TFORM($i)\"
#	echo TBFORM($i) = \"$TBFORM($i)\"
#	echo TBWIDTH($i) = \"$TBWIDTH($i)\"
#	echo TBPREC($i) = \"$TBPREC($i)\"
	}
#Don't decode TSCALE, TZERO yet
#   echo NOTE: TSCALE and TZERO decoding not yet implemented if present
   }

####Return next line of data in keyed list ####

proc fitsTableDataRead {fileid} {
   upvar 1 NROW NROW
   upvar 1 NCOL NCOL
   upvar 1 TFIELDS TFIELDS
   upvar 1 TTYPE TTYPE
   upvar 1 TBCOL TBCOL
   upvar 1 TBCOL1 TBCOL1
   upvar 1 TBCOL2 TBCOL2
   upvar 1 TBFORM TBFORM
   upvar 1 TBWIDTH TBWIDTH
   upvar 1 TBPREC TBPREC
   set data ""
   set line [read $fileid $NCOL]
   for {set i 1} {$i <= $TFIELDS} {incr i} {
	set var $TTYPE($i)
	upvar 1 $var $var
	set $var [string range $line $TBCOL1($i) $TBCOL2($i)]
	if {$TBFORM($i) == "I"} then {
	   set $var [string trim [set $var]]
#	   set $var [set $var]
# Blank = 0
	   if {[set $var] == ""} then {set $var 0}
	   lappend data "$var [set $var]"
	   continue
	   }
	if {$TBFORM($i) == "E" || $TBFORM($i) == "F"} then {
	   set $var [string trim [set $var]]
	   if {[set $var] == ""} then {
# Blank = 0
	   	set $var 0
		lappend data "$var [set $var]"
#		set $var [set $var]
		continue
		}
	   set temp [split [set $var] E]
	   set number [lindex $temp 0]
	   set exp [lindex $temp 1]
	   if {[string first . $number] == -1} then {
		set number [expr {$number / [pow 10. $TBPREC($i)]}]
		}
	   if {$exp != {}} then {
		set number [expr {$number * [pow 10. $exp]}]
		}
	   set $var $number
	   lappend data "$var [set $var]"
	   set $var [set $var]
	   }
	if {$TBFORM($i) == "D"} then {
	   set $var [string trim [set $var]]
	   if {[set $var] == ""} then {
# Blank = 0
		set $var 0
		lappend data "$var [set $var]"
#		set $var [set $var]
		continue
		}
	   set temp [split [set $var] D]
	   set number [lindex $temp 0]
	   set exp [lindex $temp 1]
	   if {[string first . $number] == -1} then {
		set number [expr {$number / [pow 10. $TBPREC($i)]}]
		}
	   if {$exp != {}} then {
		set number [expr {$number * [pow 10. $exp]}]
		}
	   set $var $number
	   lappend data "$var [set $var]"
#	   set $var [set $var]
	   }
	if {$TBFORM($i) == "A"} then {
	   lappend data "$var [set $var]"
	   set $var [set $var]
	   }
#End of loop through fields
	}
   return $data
   }

#### Main Routine to convert a FITS table to ASCII: fits2Ascii ###############

proc fits2Ascii {infile outfile {formatfile ""}} {
   set infileid [open $infile]
   set outfileid [open $outfile w]
   fitsTableHdrRead $infileid header
   fitsTableHdrDecode header
#Write out format file if specified
   if {$formatfile != ""} then {
	set formatid [open $formatfile w]
	for {set j 1} {$j <= $TFIELDS} {incr j} {
	   set TFORM $TBFORM($j)$TBWIDTH($j)
	   if {$TBPREC($j) != ""} then {set TFORM "$TFORM.$TBPREC($j)"}
	   puts $formatid [format "FIELD %-20s%-5s%-8s%s" $TTYPE($j) \
		$TBCOL($j) $TFORM $TTYPECOM($j)]
	   }
#No comments here
	close $formatid
	}
# Multitude of parameters were returned from fitsTableHdrRead by reference
   for {set j 1} {$j <= $NROW} {incr j} {
	set line [read $infileid $NCOL]
	puts $outfileid $line
	}
   close $infileid
   close $outfileid
   }

#### Main routine to read a FITS table: fitsTableRead #########################

proc fitsTableRead {file {inheader header} {indata data}} {
   upvar 1 $inheader header
   upvar 1 $indata data
   catch {unset header}
   catch {unset data}
   set fileid [open $file]
   fitsTableHdrRead $fileid header
   fitsTableHdrDecode header
# Multitude of parameters were returned from fitsTableHdrRead by reference
   puts stdout "Input number of objects to read (NROW=$NROW): " nonewline
   gets stdin nprint
   for {set j 1} {$j <= $nprint} {incr j} {
	set data($j) [fitsTableDataRead $fileid]
	if { ($j/100) * 100 == $j} then {echo Done with $j lines}
	echo $data($j)
	}
   close $fileid
   }

